#ifndef APSF_PARAMS_H
#define APSF_PARAMS_H

#include "HPLz.h"
#include "PlusDists.h"
#include "utils.h"


namespace CuTe {
namespace NS_APSF{

using NS_PlusDists::PD_Params;
using std::string;
using std::stringstream;

using namespace NS_Utils::NS_min_max;
using namespace NS_Utils::NS_pow;

enum e_P {
	e_g	// 0=gluon
	,e_q	// 1=any quark or anti-quark
	,e_qb	// 2=the anti-quark to q
	,e_qp	// 3=any (anti-)quark with same sign as q
	,e_qbp	// 4=any (anti-)quark with same sign as qb
//	,e_p	// 5=any (anti-)quark
	,e_qmp	// 6=q  minus qp
	,e_qbmp	// 7=qb minus qbp
};


inline std::ostream& operator<<(std::ostream& os,const e_P& i){
	switch(i){
		case(e_g):
			os<<"g";
			break;
		case(e_q):
			os<<"q";
			break;
		case(e_qb):
			os<<"qb";
			break;
		case(e_qp):
			os<<"qp";
			break;
		case(e_qbp):
			os<<"qbp";
			break;
//		case(e_p):
//			os<<"p";
//			break;
		case(e_qmp):
			os<<"qmp";
			break;
		case(e_qbmp):
			os<<"qbmp";
			break;
		default:
			os<<"WRONG";
			break;
	}
	return os;
}

class APSF_Parameter: public PD_Params {
public:
	HPLz hpl;

	inline void set_weight(const int weight){
		hpl.set_weight(weight);
	}

	inline virtual void update(double z){
//		NS_Utils::DOUT	dout("Ibar_Parameter::update","z");
		PD_Params::update(z);
		hpl.update(z);
	}

	inline APSF_Parameter()
	:PD_Params()
	,hpl(){
		set_weight(4);
	}

	inline APSF_Parameter(const int weight)
	:PD_Params()
	,hpl(){
		set_weight(weight);
	}

	inline virtual ~APSF_Parameter(){

	}

	inline virtual const string to_String(int ntabs=0)const{
		stringstream ss;

		string tabs("");
		for(int i=1;i<=ntabs; ++i) tabs += string("\t");

		ss	<<tabs<<"APSF_Parameter:"<<endl;
		ss	<<PD_Params::to_String(++ntabs);

		return ss.str();
	}

};

}	//NS_APSF

}	//CuTe

#endif

