#include "FO.h"
#include "FO_CS.h"
#include "param.h"

using CuTe::NS_FO::FO_NLO_Kinematic_Parameter;
using CuTe::NS_FO::FO_Kinematic_Paramete_Set;
using CuTe::NS_FO::Quad_Differential_Kernel_FO;
using CuTe::NS_FO::Pol_a;

using CuTe::NS_FO::_FO_CS;
using CuTe::NS_FO::_FO_CS_DY;
using CuTe::NS_FO::_FO_CS_Higgs;


std::ostream& FO_Kinematic_Paramete_Set::properties(std::ostream &os, const int ntabs)const{
	string tab("");
	for(int i=0; i<ntabs; ++i) tab+="\t";
	os	<<tab<<"FO_Kinematic_Paramete_Set"<<endl;
	return _param->properties(os,ntabs+1);
}

void _FO_CS::_update_0(ENUM_LUMI_UPDATE l_update){
	//		NS_Utils::DOUT dout("_FO_CS::_update_0");
	scales_0.update(nnlo_p_0);

	Phixi[0]->update(nnlo_p_0->xi[0],scales_0.muF);
	Phixi[1]->update(nnlo_p_0->xi[1],scales_0.muF);

	lumi0->update(Phixi,l_update);

	//		cout	<<lumi0->to_string(dout.get_ntab());

	Jac_1d	= nnlo_p_0->Jac_1d;
}


void _FO_CS::_update(){
//		NS_Utils::DOUT dout("_FO_CS::_update");
	_update_0(l_update_NLO_0);

	scales.update(nnlo_p);

	Phixi[0]->update(nnlo_p->xi[0],scales.muF);
	Phixi[1]->update(nnlo_p->xi[1],scales.muF);

	lumi->update(Phixi,l_update_NLO_x2);
//		cout	<<lumi->to_string(dout.get_ntab());

	Jac_2d	= nnlo_p->Jac_2d;
}

void _FO_CS::_init_pdfs(){
	Phixi[0]	= new VectorPDF(p->pdfs[0]);
	Phixi[1]	= new VectorPDF(p->pdfs[1]);
}
void _FO_CS::_exit_pdfs(){
	delete Phixi[0];	Phixi[0] = NULL;
	delete Phixi[1];	Phixi[1] = NULL;
}

void _FO_CS::_init(const FO_Kinematic_Paramete_Set *FO_kin,const CS_Parameter_LO *p){
	this->p	=	p;

	nlo_p		= FO_kin->get_nlo();
	nlo_p_tu	= FO_kin->get_nlo_t_u();

	nnlo_p		= FO_kin->get_nnlo();
	nnlo_p_tu	= FO_kin->get_nnlo_t_u();
	nnlo_p_0	= FO_kin->get_nnlo_0();
	nnlo_p_tu_0	= FO_kin->get_nnlo_0_t_u();

	plus_x2_A	= FO_kin->get_plus_x2_A();
	plus_x2_X2	= FO_kin->get_plus_x2_X2();

	scales.init(&(p->mu),&(p->as),&(p->nf));
	scales_0.init(&(p->mu),&(p->as),&(p->nf));

	ord_a	= p->order_as;
	_init_pdfs();
}


void _FO_CS::print_pdf(const string s)const{
	Phixi[0]->print_pdf(s+"0");
	Phixi[1]->print_pdf(s+"1");
}

void _FO_CS::change_PDF(const CS_Parameter_LO *p){
	_exit_pdfs();
	_init_pdfs();
}

void _FO_CS_Higgs::_init_lumis(){
	lumi	= new Lumi_S(&(p->nf));
	lumi0	= new Lumi_S(&(p->nf));
}

void _FO_CS_Higgs::_exit_lumis(){
	delete static_cast<Lumi_S*>(lumi);	lumi = NULL;
	delete static_cast<Lumi_S*>(lumi0);	lumi0 = NULL;
}

void _FO_CS_DY::_init_lumis(){
	lumi	= new Lumi_V(&(p->nf),p->prod,&(p->ckm),p->Theta_W,p->active_flv_charges);
	lumi0	= new Lumi_V(&(p->nf),p->prod,&(p->ckm),p->Theta_W,p->active_flv_charges);
}

void _FO_CS_DY::_exit_lumis(){
	delete static_cast<Lumi_V*>(lumi);	lumi = NULL;
	delete static_cast<Lumi_V*>(lumi0);	lumi0 = NULL;
}


void FO_Kinematic_Paramete_Set::init(const Kinematics *Kin, const int order_a){
//	NS_Utils::DOUT dout("FO_Kinematic_Paramete_Set::init","Kin,ord");
	this->order_a	= order_a;
//	dout()	<<"ord = "<<order_a<<endl;
	this->plus_x2_A	= &(Kin->stux2.plus_x2_A);
	this->plus_x2_X2= &(Kin->stux2.plus_x2_X2);
	_param0->init(Kin);
//		_param.init(Kin);
//		_param0.init(Kin);
//		_param_t_u.init(Kin);
//		_param0_t_u.init(Kin);
}

void FO_Kinematic_Paramete_Set::set_qT(const Kinematics *Kin){
//		NS_Utils::DOUT dout("FO_Kinematic_Paramete_Set::set_qT","Kin");
	_param0->set_qT(Kin);
}

void FO_Kinematic_Paramete_Set::set_y(const Kinematics *Kin){
//		NS_Utils::DOUT dout("FO_Kinematic_Paramete_Set::set_y","Kin");
	_param0->set_y(Kin);
	(*_param) = (*_param0);
}

void FO_Kinematic_Paramete_Set::update_xx(const Kinematics *Kin){
//		NS_Utils::DOUT dout("FO_Kinematic_Paramete_Set::update_xx","Kin");
//		cout	<<Kin->stux2.to_String(dout.get_ntab());
	if(order_a>=2){
//			NS_Utils::DOUT dout("FO_Kinematic_Paramete_Set::update_xx","order>=2");
		_param0->update_0<2>(Kin);
		(*_param0_t_u)	= (*_param0);
		_param0_t_u->swap_t_u();

		_param->update<2>(Kin);
		(*_param_t_u)	= (*_param);
		_param_t_u->swap_t_u();
	}else{
//			NS_Utils::DOUT dout("FO_Kinematic_Paramete_Set::update_xx","order<=1");
		_param0->update_0<1>(Kin);
		(*_param0_t_u)	= (*_param0);
		_param0_t_u->swap_t_u();
	}
}


FO_Kinematic_Paramete_Set::FO_Kinematic_Paramete_Set()
:_param(new FO_NNLO_Kinematic_Parameter())
,_param0(new FO_NNLO_Kinematic_Parameter())
,_param_t_u(new FO_NNLO_Kinematic_Parameter())
,_param0_t_u(new FO_NNLO_Kinematic_Parameter())
,plus_x2_A(NULL),plus_x2_X2(NULL)
,order_a(){

}
FO_Kinematic_Paramete_Set::FO_Kinematic_Paramete_Set(const FO_Kinematic_Paramete_Set &RV)
:_param(new FO_NNLO_Kinematic_Parameter(*(RV._param)))
,_param0(new FO_NNLO_Kinematic_Parameter(*(RV._param0)))
,_param_t_u(new FO_NNLO_Kinematic_Parameter(*(RV._param_t_u)))
,_param0_t_u(new FO_NNLO_Kinematic_Parameter(*(RV._param0_t_u)))
,plus_x2_A(RV.plus_x2_A),plus_x2_X2(RV.plus_x2_X2)
,order_a(RV.order_a){

}
FO_Kinematic_Paramete_Set& FO_Kinematic_Paramete_Set::operator=(const FO_Kinematic_Paramete_Set &RV){
	(*_param)		= *(RV._param);
	(*_param0)		= *(RV._param0);
	(*_param_t_u)	= *(RV._param_t_u);
	(*_param0_t_u)	= *(RV._param0_t_u);
	plus_x2_A		= (RV.plus_x2_A);
	plus_x2_X2		= (RV.plus_x2_X2);
	order_a			= (RV.order_a);
	return *this;
}
FO_Kinematic_Paramete_Set::~FO_Kinematic_Paramete_Set(){
	delete _param;	// No = NULL
	delete _param0;	// No = NULL
	delete _param_t_u;	// No = NULL
	delete _param0_t_u;	// No = NULL
}

const FO_NLO_Kinematic_Parameter* FO_Kinematic_Paramete_Set::get_nlo()const{
	return _param0;
}
const FO_NLO_Kinematic_Parameter* FO_Kinematic_Paramete_Set::get_nlo_t_u()const{
	return _param0_t_u;
}

void Quad_Differential_Kernel_FO::_exit_CS(){
	if(CS!=NULL){
		switch (prod) {
			case PROD_Higgs:
			case PROD_H_RS:
				delete static_cast<_FO_CS_Higgs*>(CS);
				break;
			case PROD_DY:
			case PROD_W:
			case PROD_Z:
				delete static_cast<_FO_CS_DY*>(CS);
				break;
			default:
				std::cerr	<<"*************************************************"	<<std::endl;
				std::cerr	<<"**** FO_CS::_exit_CS() Wrong prod="<<prod<<" ****"	<<std::endl;
				std::cerr	<<"*************************************************"	<<std::endl;
				exit(1);
				break;
		}
		CS = NULL;
	}
}

void Quad_Differential_Kernel_FO::_init_CS(){
//	if(CS!=NULL){	// DEBUG
//		std::cerr	<<"************************************"	<<std::endl;
//		std::cerr	<<"**** Quad_Differential_Kernel_FO::_init_CS() CS!=NULL ****"	<<std::endl;
//		std::cerr	<<"************************************"	<<std::endl;
//		exit(1);
//	}
	switch (prod) {
		case PROD_Higgs:
		case PROD_H_RS:
			CS	= new _FO_CS_Higgs(p,FO_kin);
			break;
		case PROD_DY:
		case PROD_W:
		case PROD_Z:
			CS	= new _FO_CS_DY(p,FO_kin);
			break;
		default:
			std::cerr	<<"*************************************************"	<<std::endl;
			std::cerr	<<"**** Quad_Differential_Kernel_FO::_init_CS() Wrong prod="<<prod<<" ****"	<<std::endl;
			std::cerr	<<"*************************************************"	<<std::endl;
			exit(1);
			break;
	}
}


void Quad_Differential_Kernel_FO::init(const CS_Parameter_LO *p, const FO_Kinematic_Paramete_Set *FO_kin){
	_exit_CS();

	this->p		= p;
	this->FO_kin= FO_kin;
	this->prod	= p->prod;

	order(p->order_as);
	mul_this_by_scalar(0.);

	_init_CS();
}

void Quad_Differential_Kernel_FO::set_qT_mu(const FO_Kinematic_Paramete_Set *FO_kin){
	//only to remind setting FO_kin->set_qT()
	CS->set_qT_mu(FO_kin);
}

void Quad_Differential_Kernel_FO::set_y(const FO_Kinematic_Paramete_Set *FO_kin){
	//only to remind setting FO_kin->set_y()
	CS->set_y(FO_kin);
}

Pol_a& Quad_Differential_Kernel_FO::update_xx(const FO_Kinematic_Paramete_Set *FO_kin){
	//only to remind setting FO_kin->update_xx
//		NS_Utils::DOUT dout("Quad_Differential_Kernel_FO::update_xx");
	return CS->update_xx(*this,FO_kin);
}


void Quad_Differential_Kernel_FO::print_pdf(const std::string s)const{
	CS->print_pdf(s);
}

void Quad_Differential_Kernel_FO::change_PDF(const CS_Parameter_LO *p){
	CS->change_PDF(p);
}
