#ifndef FO_H
#define FO_H

//#include "FO_Lumis.h"
//#include "FO_Higgs.h"
//#include "FO_DY.h"
//#include "param.h"
#include "polynom.h"
#include "enums.h"

namespace CuTe {

namespace NS_Kinematics {
	class Kinematics;
	class Plus_x2;
}  // namespace NS_Kinematics

namespace NS_CS_Params {
	class CS_Parameter_LO;
}  // namespace NS_CS_Params

namespace NS_FO {

class FO_NLO_Kinematic_Parameter;
class FO_NNLO_Kinematic_Parameter;
class _FO_CS;

typedef NS_Polynom::Pol<double> Pol_a;

using NS_Kinematics::Kinematics;
using NS_Kinematics::Plus_x2;
using NS_CS_Params::CS_Parameter_LO;

using namespace NS_enums::NS_PROD;



class FO_Kinematic_Paramete_Set {
protected:
	FO_NNLO_Kinematic_Parameter *_param;
	FO_NNLO_Kinematic_Parameter *_param0;
	FO_NNLO_Kinematic_Parameter *_param_t_u;
	FO_NNLO_Kinematic_Parameter *_param0_t_u;

	const Plus_x2 *plus_x2_A,*plus_x2_X2;

public:
	int order_a;

	void init(const Kinematics *Kin, const int order_a);
	void set_qT(const Kinematics *Kin);
	void set_y(const Kinematics *Kin);
	void update_xx(const Kinematics *Kin);

	const Plus_x2* get_plus_x2_A()const{return plus_x2_A;}
	const Plus_x2* get_plus_x2_X2()const{return plus_x2_X2;}

	const FO_NNLO_Kinematic_Parameter* get_nnlo()const{return _param;}
	const FO_NNLO_Kinematic_Parameter* get_nnlo_0()const{return _param0;}
	const FO_NNLO_Kinematic_Parameter* get_nnlo_t_u()const{return _param_t_u;}
	const FO_NNLO_Kinematic_Parameter* get_nnlo_0_t_u()const{return _param0_t_u;}

	const FO_NLO_Kinematic_Parameter* get_nlo()const;
	const FO_NLO_Kinematic_Parameter* get_nlo_t_u()const;

	FO_Kinematic_Paramete_Set();
	FO_Kinematic_Paramete_Set(const FO_Kinematic_Paramete_Set &RV);
	FO_Kinematic_Paramete_Set& operator=(const FO_Kinematic_Paramete_Set &RV);
	~FO_Kinematic_Paramete_Set();


	std::ostream& properties(std::ostream &os, const int ntabs=0)const;

};

class Quad_Differential_Kernel_FO:public Pol_a{
protected:
	const CS_Parameter_LO *p;
	const FO_Kinematic_Paramete_Set *FO_kin;

	PROD prod;
	_FO_CS *CS;

	void _exit_CS();
	void _init_CS();

public:

	void init(const CS_Parameter_LO *p,const FO_Kinematic_Paramete_Set *FO_kin);

	void set_qT_mu(const FO_Kinematic_Paramete_Set *FO_kin);

	void set_y(const FO_Kinematic_Paramete_Set *FO_kin);

	Pol_a& update_xx(const FO_Kinematic_Paramete_Set *FO_kin);

	Quad_Differential_Kernel_FO()
	:Pol_a(2,0.)
	,p(NULL),FO_kin(NULL)
	,prod(),CS(NULL){
//		NS_Utils::DOUT dout("Quad_Differential_Kernel_FO");
	}

	Quad_Differential_Kernel_FO(const Quad_Differential_Kernel_FO & fo)
	:Pol_a(2,0.)
	,p(fo.p),FO_kin(fo.FO_kin)
	,prod(fo.prod),CS(NULL){
		set_equalt_to(fo);
		if(fo.CS!=NULL)	_init_CS();
	}

	Quad_Differential_Kernel_FO& operator=(const Quad_Differential_Kernel_FO & fo){
//		NS_Utils::DOUT dout("Quad_Differential_Kernel_FO::operator=","fo");
		set_equalt_to(fo);

		 _exit_CS();
		p		= (fo.p);
		FO_kin	= (fo.FO_kin);
		prod	= (fo.prod);
		if(fo.CS!=NULL)	_init_CS();


		return *this;
	}

	~Quad_Differential_Kernel_FO(){
		_exit_CS();
	}

	void print_pdf(const std::string s="")const;

	void change_PDF(const CS_Parameter_LO *p);
};

}  // namespace NS_FO
}  // namespace CuTe

#endif

