#ifndef FO_CS_H_
#define FO_CS_H_

#include "FO_Lumis.h"
#include "FO_DY.h"
#include "FO_Higgs.h"

namespace CuTe {

namespace NS_Kinematics {
	class Kinematics;
	class Plus_x2;
}  // namespace NS_Kinematics

namespace NS_CS_Params {
	class CS_Parameter_LO;
}  // namespace NS_CS_Params


namespace NS_FO {

using NS_Kinematics::Kinematics;
using NS_Kinematics::Plus_x2;
using NS_CS_Params::CS_Parameter_LO;

class _FO_CS{
protected:
	const CS_Parameter_LO *p;
	const FO_NLO_Kinematic_Parameter *nlo_p,*nlo_p_tu;
	const FO_NNLO_Kinematic_Parameter *nnlo_p,*nnlo_p_0,*nnlo_p_tu,*nnlo_p_tu_0;

	FO_Scale_Parameter scales,scales_0;

	const Plus_x2 *plus_x2_A,*plus_x2_X2;

	Lumi *lumi,*lumi0;
	VectorPDF* Phixi[2];

	int ord_a;

	double Jac_1d,Jac_2d;

	void _update_0(ENUM_LUMI_UPDATE l_update);

	void _update();

	inline virtual void _init_lumis(){
		cerr << "************************************************" << endl
			 << "*** _FO_CS::_init_lumis(p) called -> exit(1) ***" << endl
			 << "************************************************" << endl;
		exit(1);
	}

	inline virtual void _exit_lumis(){
		cerr << "**********************************************" << endl
			 << "*** _FO_CS::_set_lumis() called -> exit(1) ***" << endl
			 << "**********************************************" << endl;
		exit(1);
	}

	void _init_pdfs();
	void _exit_pdfs();

	void _init(const FO_Kinematic_Paramete_Set *FO_kin,const CS_Parameter_LO *p);

	inline virtual double _LO()const{
		cerr << "***************************************" << endl
			 << "*** _FO_CS::_LO() called -> exit(1) ***" << endl
			 << "***************************************" << endl;
		exit(1);
		return 0.;
	};
	inline virtual double _NLO()const{
		cerr << "****************************************" << endl
			 << "*** _FO_CS::_NLO() called -> exit(1) ***" << endl
			 << "****************************************" << endl;
		exit(1);
		return 0.;
	};

public:
	inline virtual void init(const CS_Parameter_LO *p, const FO_Kinematic_Paramete_Set *FO_kin){
		cerr << "****************************************" << endl
			 << "*** _FO_CS::init(kin,p) called -> exit(1) ***" << endl
			 << "****************************************" << endl;
		exit(1);
	};



	inline void set_qT_mu(const FO_Kinematic_Paramete_Set *FO_kin){
		//only to remind setting FO_kin->set_qT()
	}

	inline void set_y(const FO_Kinematic_Paramete_Set *FO_kin){
		//only to remind setting FO_kin->set_y()
	}


	inline Pol_a& update_xx(Pol_a &P,const FO_Kinematic_Paramete_Set *FO_kin){
		//only to remind setting FO_kin->update_xx
//		NS_Utils::DOUT dout("_FO_CS::update_xx");

		P.order(ord_a);

		if(ord_a>1){
			_update();
			P[2] = _NLO();
			P[1] = _LO();
		}else if(ord_a>0){
			_update_0(l_update_LO);
			P[1] = _LO();
		}
		P[0] = 0.;

		P.set_of(2.*(scales_0.as));

		for(int i=0; i<=P.order(); ++i)	if(P[i]!=P[i]){
			cerr<<"NAN:	_FO_CS::update_xx["<<i<<"]"<<endl;
			FO_kin->properties(cerr);
			return P;
		}
		return P;
	}

	_FO_CS()
	:p(NULL)
	,nlo_p(NULL),nlo_p_tu(NULL)
	,nnlo_p(NULL),nnlo_p_0(NULL),nnlo_p_tu(NULL),nnlo_p_tu_0(NULL)
	,scales(),scales_0()
	,plus_x2_A(NULL),plus_x2_X2(NULL)
	,lumi(NULL),lumi0(NULL)
	,Phixi()
	,ord_a()
	,Jac_1d(),Jac_2d(){
//		NS_Utils::DOUT dout("_FO_CS");
		Phixi[0]	= NULL;
		Phixi[1]	= NULL;
	}

	virtual ~_FO_CS(){
		_exit_pdfs();
	}

	_FO_CS(const _FO_CS &CS)
	:p(CS.p)
	,nlo_p(CS.nlo_p),nlo_p_tu(CS.nlo_p_tu)
	,nnlo_p(CS.nnlo_p),nnlo_p_0(CS.nnlo_p_0),nnlo_p_tu(CS.nnlo_p_tu),nnlo_p_tu_0(CS.nnlo_p_tu_0)
	,scales(CS.scales),scales_0(CS.scales_0)
	,plus_x2_A(CS.plus_x2_A),plus_x2_X2(CS.plus_x2_X2)
	,lumi(NULL),lumi0(NULL)
	,Phixi()
	,ord_a(CS.ord_a)
	,Jac_1d(CS.Jac_1d),Jac_2d(CS.Jac_2d){
//		NS_Utils::DOUT dout("_FO_CS","CS");
		Phixi[0]	= NULL;
		Phixi[1]	= NULL;
		if(p!=NULL) _init_pdfs();
	}


	_FO_CS& operator=(const _FO_CS &CS){
//		NS_Utils::DOUT dout("_FO_CS::operator=","CS");
		_exit_pdfs();
		p			= (CS.p);
		nlo_p		= (CS.nlo_p);
		nlo_p_tu	= (CS.nlo_p_tu);
		nnlo_p		= (CS.nnlo_p);
		nnlo_p_0	= (CS.nnlo_p_0);
		nnlo_p_tu	= (CS.nnlo_p_tu);
		nnlo_p_tu_0	= (CS.nnlo_p_tu_0);
		scales		= (CS.scales);
		scales_0	= (CS.scales_0);
		plus_x2_A	= (CS.plus_x2_A);
		plus_x2_X2	= (CS.plus_x2_X2);
		ord_a		= (CS.ord_a);
		Jac_1d		= (CS.Jac_1d);
		Jac_2d		= (CS.Jac_2d);

		if(p!=NULL) _init_pdfs();

		return *this;
	}

	void print_pdf(const string s="")const;

	void change_PDF(const CS_Parameter_LO *p);

};


class _FO_CS_Higgs:public _FO_CS{
protected:
	constexpr static const double PrefGG	=C_A;
	constexpr static const double PrefqG	=C_F;
	constexpr static const double PrefGq	=C_F;
	constexpr static const double Prefqp	=2.*C_F*C_F;


	void _init_lumis();
	void _exit_lumis();

	inline void _LO_0(double& GG, double& Gq, double& qG, double& qbq)const{
		using namespace NS_Higgs;
		// qG
		qG +=  (*lumi0)[LS_qG_LO]*qg_Hq_LO(*nlo_p);
		// Gq
		Gq +=  (*lumi0)[LS_Gq_LO]*qg_Hq_LO(*nlo_p_tu);
		// GG
		GG +=  (*lumi0)[LS_GG_LO]*gg_Hg_LO(*nlo_p);
		// qbq
		qbq += (*lumi0)[LS_qbq_LO]*qqbar_Hg_LO(*nlo_p);
	}

	inline double _LO()const{
//		NS_Utils::DOUT dout("_FO_CS_Higgs::_LO");

		double result(0.),GG(0.),qG(0.),Gq(0.),qbq(0.);

		_LO_0(GG,Gq,qG,qbq);

		 GG *=	PrefGG;
		 qG	*=	PrefqG;
		 Gq	*=	PrefGq;
		qbq	*=	Prefqp;


		result	=	GG + qG + Gq + qbq;
		return result*Jac_1d*nlo_p->iM2;
	}



	inline void _NLO_2s(double& GG,double& Gq,double& qG,double& qbq, double& qp)const{
//		NS_Utils::DOUT dout("_FO_CS_Higgs::_NLO_2s");
		using namespace NS_Higgs;
//		double D_jac(param.Ratio_Jac_2d_to_plus);
		double l0,lx;
		double f0d,f00,f01,fx0,fx1,fxf;

		//GG
		l0	= (*lumi0)[LS_GG_LO];
		lx	= (*lumi )[LS_GG_LO];
		f0d	= l0*GG_2s_delta(*nnlo_p_0,scales_0    );
		f00	= l0*GG_2s_plus0(*nnlo_p_0,*nnlo_p_tu_0,scales_0);
		f01	= l0*GG_2s_plus1(*nnlo_p_0,*nnlo_p_tu_0);
		fx0 = lx*GG_2s_plus0(*nnlo_p  ,*nnlo_p_tu  ,scales);
		fx1	= lx*GG_2s_plus1(*nnlo_p  ,*nnlo_p_tu  );
		fxf	= lx*GG_2s_regul(*nnlo_p  ,scales      );

		//DEBUG
//		if(fxf!=fxf)	cerr<<"NAN _NLO_2s fxf GG"<<endl;

	#ifndef PLUS_X2_X2
		qbq	+= (*plus_x2_A)(f0d,fx0,f00,fx1,f01,fxf);
	#else
		qbq	+= (*plus_x2_X2)(f0d,fx0,f00,fx1,f01,fxf);
	#endif


		//qbq
		l0	= (*lumi0)[LS_qbq_LO];
		lx	= (*lumi )[LS_qbq_LO];
		f0d	= l0*qbq_2s_delta(*nnlo_p_0,scales_0);
		f00	= l0*qbq_2s_plus0(*nnlo_p_0,scales_0);
		f01	= l0*qbq_2s_plus1(*nnlo_p_0);
		fx0 = lx*qbq_2s_plus0(*nnlo_p  ,scales);
		fx1	= lx*qbq_2s_plus1(*nnlo_p  );
		fxf	= lx*qbq_2s_regul(*nnlo_p  );

//		//DEBUG
//		if(fxf!=fxf)	cerr<<"NAN _NLO_2s fxf qbq"<<endl;

	#ifndef PLUS_X2_X2
		qbq	+= (*plus_x2_A)(f0d,fx0,f00,fx1,f01,fxf);
	#else
		qbq	+= (*plus_x2_X2)(f0d,fx0,f00,fx1,f01,fxf);
	#endif

		//qp
//		l0	= (*lumi0)[LS_qp_NLO];
		lx	= (*lumi )[LS_qp_NLO];
		f0d	= 0.;
		f00	= 0.;
		f01	= 0.;
		fx0 = 0.;
		fx1	= 0.;
		fxf= lx*qp_2s_regul(*nnlo_p,scales);

		//DEBUG
//		if(fxf!=fxf)	cerr<<"NAN _NLO_2s fxf qp"<<endl;

	#ifndef PLUS_X2_X2
		qbq	+= (*plus_x2_A)(f0d,fx0,f00,fx1,f01,fxf);
	#else
		qbq	+= (*plus_x2_X2)(f0d,fx0,f00,fx1,f01,fxf);
	#endif

		//qG
		l0	= (*lumi0)[LS_qG_LO];
		lx	= (*lumi )[LS_qG_LO];
		f0d	= l0*qG_2s_delta(*nnlo_p_0,scales_0);
		f00	= l0*qG_2s_plus0(*nnlo_p_0,scales_0);
		f01	= l0*qG_2s_plus1(*nnlo_p_0);
		fx0 = lx*qG_2s_plus0(*nnlo_p  ,scales);
		fx1	= lx*qG_2s_plus1(*nnlo_p  );
		fxf	= lx*qG_2s_regul(*nnlo_p  ,scales);

//		//DEBUG
//		if(fxf!=fxf)	cerr<<"NAN _NLO_2s fxf qG"<<endl;

		qG	+= (*plus_x2_A)(f0d,fx0,f00,fx1,f01,fxf);


		//Gq
		l0	= (*lumi0)[LS_Gq_LO];
		lx	= (*lumi )[LS_Gq_LO];
		f0d	= l0*qG_2s_delta(*nnlo_p_tu_0,scales_0);
		f00	= l0*qG_2s_plus0(*nnlo_p_tu_0,scales_0);
		f01	= l0*qG_2s_plus1(*nnlo_p_tu_0);
		fx0 = lx*qG_2s_plus0(*nnlo_p_tu  ,scales);
		fx1	= lx*qG_2s_plus1(*nnlo_p_tu  );
		fxf	= lx*qG_2s_regul(*nnlo_p_tu  ,scales);

		//DEBUG
//		if(fxf!=fxf)	cerr<<"NAN _NLO_2s fxf Gq"<<endl;

		Gq	+= (*plus_x2_A)(f0d,fx0,f00,fx1,f01,fxf);
	}

//	inline void _NLO_2R_ns(double& qq, double& qp)const{
////		NS_Utils::DOUT dout("H_NLO::Crosssection::_NLO_2R_ns","qq,qp");
//		using namespace Coefficients;
//		//qq
//		qq	+= (*lumi)[qq_NLO]*qq_2R_ns(param);
//
//		//qp
//		qp	+= (*lumi)[qp_NLO]*qp_finite(param);
////		dout()	<<"l_qp = "<<(*lumi)[qp_NLO]	<<"\t"<<"qp_finite = "<<qp/(*lumi)[qp_NLO]	<<endl;
//	}

	inline void _NLO_2R_ns(double& GG,double& Gq,double& qG,double& qbq,double& qq, double& qp)const{
//		NS_Utils::DOUT dout("_FO_CS_Higgs::_NLO_2R_ns");
//		NS_Utils::DOUT dout("H_NLO::Crosssection::_NLO_2R_ns","qq,qp");
		using namespace NS_Higgs;

		GG	+= (*lumi)[ LS_GG_LO]	* GG_2R_ns(*nnlo_p   ,*nnlo_p_tu,scales);//GG_finite(*nnlo_p,*nnlo_p_tu,*nf);
		qG	+= (*lumi)[ LS_qG_LO]	* qG_2R_ns(*nnlo_p   ,*nnlo_p_tu);//qG_finite(*nnlo_p,*nnlo_p_tu,*nf);
		Gq	+= (*lumi)[ LS_Gq_LO]	* qG_2R_ns(*nnlo_p_tu,*nnlo_p   );//Gq_finite(*nnlo_p,*nnlo_p_tu,*nf);
		qbq	+= (*lumi)[LS_qbq_LO]	*qbq_2R_ns(*nnlo_p   ,*nnlo_p_tu,scales);//qbq_finite(*nnlo_p,*nnlo_p_tu,*nf);
		qq	+= (*lumi)[ LS_qq_NLO]* qq_2R_ns(*nnlo_p);
		qp	+= (*lumi)[ LS_qp_NLO]* qp_2R_ns(*nnlo_p);//qp_finite(*nnlo_p);
//		dout()	<<"l_qp = "<<(*lumi)[qp_NLO]	<<"\t"<<"qp_finite = "<<qp/(*lumi)[qp_NLO]	<<endl;
	}

	inline double _NLO()const{
//		NS_Utils::DOUT dout("_FO_CS_Higgs::_NLO");
		double result(0.),GG(0.),Gq(0.),qG(0.),qbq(0.),qq(0.),qp(0.);

#ifndef ONLY_NON_SINGULAR
//		_NLO_0(GG,Gq,qG/*,qbq*/);
//	 	 GG	*= param.Ratio_Jac_2d_to_delta;
//	 	 Gq	*= param.Ratio_Jac_2d_to_delta;
//		 qG	*= param.Ratio_Jac_2d_to_delta;
//		qbq	*= param.Ratio_Jac_2d_to_delta;

		_NLO_2s(GG,Gq,qG,qbq,qp);
#endif
#ifndef ONLY_SINGULAR
		_NLO_2R_ns(GG,Gq,qG,qbq,qq,qp);
#endif

//		dout()	<<" GG = "<<GG*param.Jac_2d	<<	"\t"<<"L[ GG] = "<<(*lumi)[GG_LO]	<<"\t"<<"L0[ GG] = "<<(*lumi0)[GG_LO]<<endl;
//		dout()	<<"qGq = "<<(qG+Gq)*param.Jac_2d	<<endl;
//		dout()	<<"qbq = "<<qbq*param.Jac_2d	<<endl;
//		dout()	<<"   += qp = "<<(qbq+qp)*param.Jac_2d	<<endl;
//		dout()	<<" qq = "<<qq*param.Jac_2d	<<endl;
//		dout()	<<"   += qp = "<<(qq+qp)*param.Jac_2d	<<endl;
//		dout()	<<" qp = "<<qp*param.Jac_2d	<<endl;
//		dout()	<<"  q = "<<(qbq + qq + qp)*param.Jac_2d	<<endl;

		result	=	GG + Gq + qG + qbq + qq + qp;
		result	*=	Jac_2d;

		return result*nlo_p->iM2;
	}


public:

	inline void init(const CS_Parameter_LO *p,const FO_Kinematic_Paramete_Set *FO_kin){
		_init(FO_kin,p);
		_init_lumis();
	}


	_FO_CS_Higgs()
	:_FO_CS(){
//		NS_Utils::DOUT dout("_FO_CS_Higgs");
//		cout << "DY_NLO::Crosssection()" << endl;
	}

	_FO_CS_Higgs(const CS_Parameter_LO *p,const FO_Kinematic_Paramete_Set *FO_kin)
	:_FO_CS(){
//		NS_Utils::DOUT dout("_FO_CS_Higgs","p,kin");
		init(p,FO_kin);
	}

	_FO_CS_Higgs(const _FO_CS_Higgs &CS)
	:_FO_CS(CS){
//		NS_Utils::DOUT dout("_FO_CS_Higgs","CS");
		_init_lumis();
	}

	_FO_CS_Higgs& operator=(const _FO_CS_Higgs &CS){
//		NS_Utils::DOUT dout("_FO_CS_Higgs::operator=","CS");
		_exit_lumis();
		_FO_CS::operator =(CS);
		if(p!=NULL) _init_lumis();

		return *this;
	}

	~_FO_CS_Higgs(){
		_exit_lumis();
	}

};


class _FO_CS_DY:public _FO_CS{
protected:

	constexpr static const double PrefqG	=0.5;
	constexpr static const double PrefGq	=0.5;
	constexpr static const double Prefqbq	=C_F;
	constexpr static const double PrefGG	=1./(4.*C_F);
	constexpr static const double Prefqq	=0.5*C_F;

	void _init_lumis();
	void _exit_lumis();

	inline void _LO_0(double& qG, double& Gq, double& qbq)const{
		using namespace NS_DY;

		double A_qG0 =  A_qG(*nlo_p);
		double A_Gq0 =  A_qG(*nlo_p_tu);
		double A_qbq0= A_qbq(*nlo_p);
		// qG
		qG +=  (*lumi0)[LV_qG_LO]*A_qG0;
		// Gq
		Gq +=  (*lumi0)[LV_Gq_LO]*A_Gq0;
		// qbq
		qbq += (*lumi0)[LV_qbq_LO]*A_qbq0;
	}

	inline double _LO()const{
		double result(0.),qG(0.),Gq(0.),qbq(0.);

		_LO_0(qG,Gq,qbq);

		 qG	*=	PrefqG;
		 Gq	*=	PrefGq;
		qbq	*=	Prefqbq;

//		cout	<<"LO:"	<<endl
//				<<"\t"	<<"\tqG="<<qG	<<"\t= "<<A_qG0<<" * "<<(*lumi0)[ qG_LO]	<<endl
//				<<"\t"	<<"\tGq="<<Gq	<<"\t= "<<A_Gq0<<" * "<<(*lumi0)[ Gq_LO]	<<endl	<<endl
//				<<"\t"	<<"\tqGq="<<qG+Gq	<<endl
//				<<"\t"	<<"\tqbq="<<qbq	<<"\t= "<<A_qbq0<<" * "<<(*lumi0)[ qbq_LO]	<<endl	<<endl
//				;

//		//DEBUG
//		if(qG!=qG)	cerr<<"NAN\t_FO_CS_DY::_LO() qg"<<endl;
//		if(Gq!=Gq)	cerr<<"NAN\t_FO_CS_DY::_LO() Gq"<<endl;
//		if(qbq!=qbq)	cerr<<"NAN\t_FO_CS_DY::_LO() qbq"<<endl;
//		if(Jac_1d!=Jac_1d)	cerr<<"NAN\t_FO_CS_DY::_LO() Jac_1d"<<endl;

		result	=	qG + Gq + qbq;
		return result*Jac_1d*0.5;
	}



//	inline void _NLO_0(double& qG, double& Gq, double& qbq)const{
////		cout << "-> Crosssection::_NLO_0()" << endl;
//		using namespace NS_DY;
//
////		// qG
////		double B1_qG	= B1_qG_finite(*nnlo_p_0,scales_0,A_qG0);
////		double C1_qG	= C1_qG_finite(*nnlo_p_0,scales_0,A_qG0);
////		qG += (*lumi0)[LV_qG_LO ]*(B1_qG + C1_qG);
//
//		double B3_qG	= B3_qG(*nnlo_p_0);
//		qG += (*lumi0)[LV_qG_NLO]*(B3_qG);
//
//////		// Gq
////		double B1_Gq	= B1_qG_finite(*nnlo_p_tu_0,scales_0,A_Gq0);
////		double C1_Gq	= C1_qG_finite(*nnlo_p_tu_0,scales_0,A_Gq0);
////		Gq += (*lumi0)[LV_Gq_LO ]*(B1_Gq + C1_Gq);
//
//		double B3_Gq	= B3_qG(*nnlo_p_tu_0);
//		Gq	+= (*lumi0)[LV_Gq_NLO]*(B3_Gq);
//
//#ifndef ONLY_MIXED
////		// qbq
////		double B1_qbq	= B1_qbq_finite(*nnlo_p_0,scales_0,A_qbq0);
////		double B2_qbq	= B2_qbq_finite(scales_0,A_qbq0);
////		double C1_qbq	= C1_qbq_finite(*nnlo_p_0,scales_0,A_qbq0);
////		double D_0_aa	= D_0_aa_finite(*nnlo_p_0,A_qbq0);
////		qbq	+= (*lumi0)[LV_qbq_LO]*(B1_qbq+C1_qbq+(B2_qbq+D_0_aa)*(scales_0.nf));
//
//		double B3_qbq	= B3_qbq(*nnlo_p_0);
//		qbq	+= (*lumi0)[LV_qbq_NLO_ab]*B3_qbq;
//#endif
//	}


	inline void _NLO_Plus(double& qG,double& Gq,double& qbq)const{
		using namespace NS_DY;

		double A_qG0 =  A_qG(*nlo_p);
		double A_Gq0 =  A_qG(*nlo_p_tu);
		double A_qbq0= A_qbq(*nlo_p);

		double l0,lx;
		double f0d,f00,f01,fx0,fx1,fxf;

		//**** qG
		//---- qG_NLO
		l0	= (*lumi0)[LV_qG_LO];
		lx	=  (*lumi)[LV_qG_LO];
		f0d	= l0*nnlo_qg_nlo_delta(*nnlo_p_0,*nnlo_p_tu_0,scales_0,A_qG0);
		f00	= l0*nnlo_qg_nlo_plus0(*nnlo_p_0,*nnlo_p_tu_0,scales_0,A_qG0);
		f01	= l0*nnlo_qg_nlo_plus1(*nnlo_p_0,*nnlo_p_tu_0,scales_0,A_qG0);
		fx0	= lx*nnlo_qg_nlo_plus0(*nnlo_p  ,*nnlo_p_tu  ,scales  ,A_qG0);
		fx1	= lx*nnlo_qg_nlo_plus1(*nnlo_p  ,*nnlo_p_tu  ,scales  ,A_qG0);
		fxf	= lx*nnlo_qg_nlo_finit(*nnlo_p  ,*nnlo_p_tu  ,scales  ,A_qG0);

		qG	+= (*plus_x2_A)(f0d,fx0,f00,fx1,f01,fxf);


		//---- qG_NNLO
		l0	= (*lumi0)[LV_qG_NLO];
		f0d	= l0*nnlo_qg_nnlo_delta(*nnlo_p_0,*nnlo_p_tu_0,scales_0,A_qG0);

		qG	+= plus_x2_A->delta(f0d);

		//**** Gq
		//---- Gq_NLO
		l0	= (*lumi0)[LV_Gq_LO];
		lx	=  (*lumi)[LV_Gq_LO];
		f0d	= l0*nnlo_qg_nlo_delta(*nnlo_p_tu_0,*nnlo_p_0,scales_0,A_Gq0);
		f00	= l0*nnlo_qg_nlo_plus0(*nnlo_p_tu_0,*nnlo_p_0,scales_0,A_Gq0);
		f01	= l0*nnlo_qg_nlo_plus1(*nnlo_p_tu_0,*nnlo_p_0,scales_0,A_Gq0);
		fx0	= lx*nnlo_qg_nlo_plus0(*nnlo_p_tu  ,*nnlo_p  ,scales  ,A_Gq0);
		fx1	= lx*nnlo_qg_nlo_plus1(*nnlo_p_tu  ,*nnlo_p  ,scales  ,A_Gq0);
		fxf	= lx*nnlo_qg_nlo_finit(*nnlo_p_tu  ,*nnlo_p  ,scales  ,A_Gq0);

		Gq	+= (*plus_x2_A)(f0d,fx0,f00,fx1,f01,fxf);

		//---- Gq_NNLO
		l0	= (*lumi0)[LV_Gq_NLO];
		f0d	= l0*nnlo_qg_nnlo_delta(*nnlo_p_tu_0,*nnlo_p_0,scales_0,A_Gq0);

		Gq	+= plus_x2_A->delta(f0d);


		//**** qbq
		//---- qbq_NLO
		l0	= (*lumi0)[LV_qbq_LO];
		lx	=  (*lumi)[LV_qbq_LO];
		f0d	= l0*nnlo_qbq_nlo_delta(*nnlo_p_tu_0,*nnlo_p_0,scales_0,A_qbq0);
		f00	= l0*nnlo_qbq_nlo_plus0(*nnlo_p_tu_0,*nnlo_p_0,scales_0,A_qbq0);
		f01	= l0*nnlo_qbq_nlo_plus1(*nnlo_p_tu_0,*nnlo_p_0,scales_0,A_qbq0);
		fx0	= lx*nnlo_qbq_nlo_plus0(*nnlo_p_tu  ,*nnlo_p  ,scales  ,A_qbq0);
		fx1	= lx*nnlo_qbq_nlo_plus1(*nnlo_p_tu  ,*nnlo_p  ,scales  ,A_qbq0);
		fxf	= lx*nnlo_qbq_nlo_finit(*nnlo_p_tu  ,*nnlo_p  ,scales  ,A_qbq0);

		qbq	+= (*plus_x2_A)(f0d,fx0,f00,fx1,f01,fxf);

		//---- qbq_NNLO_ab
		l0	= (*lumi0)[LV_qbq_NLO_ab];
		f0d	= l0*nnlo_qbq_nnlo_ab_delta(*nnlo_p_tu_0,*nnlo_p_0,scales_0,A_qbq0);
		fxf	= l0*nnlo_qbq_nnlo_ab_finit(*nnlo_p_tu  ,*nnlo_p  ,scales  ,A_qbq0);

		qbq	+= plus_x2_A->delta(f0d,fxf);


	}

	inline void _NLO_x2(double& qG, double& Gq, double& GG, double& qbq, double& qq)const{
		using namespace NS_DY;

		// GG
		GG += (*lumi)[LV_GG_NLO]*nnlo_GG_nnlo_x2(*nnlo_p,*nnlo_p_tu,scales);

		//qbq

		qbq	+=	(*lumi)[LV_qbq_NLO_bb]*nnlo_qbq_nnlo_bb_x2(*nnlo_p,*nnlo_p_tu,scales);

		qbq +=	(*lumi)[LV_qbq_NLO_bc]*nnlo_qbq_nnlo_bc_x2(*nnlo_p,*nnlo_p_tu,scales);
		qbq +=	(*lumi)[LV_qbq_NLO_bd]*nnlo_qbq_nnlo_bd_x2(*nnlo_p,*nnlo_p_tu,scales);

		double d_cc	= nnlo_qbq_nnlo_cc_x2(*nnlo_p,*nnlo_p_tu,scales);
		qbq +=	(*lumi)[LV_qbq_NLO_cc]*d_cc;

		double d_dd	= nnlo_qbq_nnlo_dd_x2(*nnlo_p,*nnlo_p_tu,scales);
		qbq +=	(*lumi)[LV_qbq_NLO_dd]*d_dd;

		double d_LL	= nnlo_qbq_nnlo_LL_x2(*nnlo_p,*nnlo_p_tu,scales);
		qbq +=	(*lumi)[LV_qbq_NLO_LL]*d_LL;

		double d_LR	= nnlo_qbq_nnlo_LR_x2(*nnlo_p,*nnlo_p_tu,scales);
		qbq +=	(*lumi)[LV_qbq_NLO_LR]*d_LR;

//		// q_q
		double E_aa		=d_cc;						//(A28)
		double E_cc		=d_cc;						//(A28)
		qq +=	(*lumi)[LV_qq_NLO_aa]*(E_aa+E_cc);

		double E_bb		=d_dd;						//(A29)
		double E_dd		=d_dd;						//(A29)
		qq +=	(*lumi)[LV_qq_NLO_bb]*(E_bb+E_dd);

		qq +=	(*lumi)[LV_qq_NLO_ac]*nnlo_qq_nnlo_ac_x2(*nnlo_p,*nnlo_p_tu,scales);

		qq +=	(*lumi)[LV_qq_NLO_bd]*nnlo_qq_nnlo_bd_x2(*nnlo_p,*nnlo_p_tu,scales);

		qq +=	(*lumi)[LV_qq_NLO_ad]*nnlo_qq_nnlo_ad_x2(*nnlo_p,*nnlo_p_tu,scales);

		double  E_LL_ab	=-d_LR;		//(A30)
		double  E_LL_cd	=-d_LR;		//(A30)
		qq +=	(*lumi)[LV_qq_NLO_LL]*(E_LL_ab+E_LL_cd);

		double  E_LR_ab	=-d_LL;		//(A31)
		double  E_LR_cd	=-d_LL;		//(A31)
		qq +=	(*lumi)[LV_qq_NLO_LR]*(E_LR_ab+E_LR_cd);

	}

	inline double _NLO()const{
//		cout << "-> Crosssection::_NLO(xx)" << endl;
		double result(0.),qG(0.),Gq(0.),GG(0.),qbq(0.),qq(0.);

		_NLO_Plus(qG,Gq,qbq);
		 if(qG!=qG)	cerr<<"NAN\t_FO_CS_DY::_NLO() qG PLUS"<<endl;
		 if(Gq!=Gq)	cerr<<"NAN\t_FO_CS_DY::_NLO() Gq PLUS"<<endl;
		 if(qbq!=qbq)	cerr<<"NAN\t_FO_CS_DY::_NLO() qbq PLUS"<<endl;
		_NLO_x2(qG,Gq,GG,qbq,qq);

		 qG	*=	PrefqG;
		 Gq	*=	PrefGq;
		 GG	*=	PrefGG;
		qbq	*=	Prefqbq;
		 qq	*=	Prefqq;


		 if(qG!=qG)	cerr<<"NAN\t_FO_CS_DY::_NLO() qG X2"<<endl;
		 if(Gq!=Gq)	cerr<<"NAN\t_FO_CS_DY::_NLO() Gq X2"<<endl;
		 if(GG!=GG)	cerr<<"NAN\t_FO_CS_DY::_NLO() GG X2"<<endl;
		 if(qbq!=qbq)	cerr<<"NAN\t_FO_CS_DY::_NLO() qbq X2"<<endl;
		 if(qq!=qq)	cerr<<"NAN\t_FO_CS_DY::_NLO() qq X2"<<endl;


//			cout	<<"NLO:"	<<endl
////					<<"\t"	<<"\tqG="<<qG	<<endl
////					<<"\t"	<<"\tGq="<<Gq	<<endl
//					<<"\t"	<<"\tqGq="<<qG+Gq	<<endl
//					<<"\t"	<<"\tGG="<<GG	<<endl
//					<<"\t"	<<"\tqbq="<<qbq	<<endl
//					<<"\t"	<<"\tqq="<<qq	<<endl
//					;

		result	=	qG + Gq + GG + qbq + qq;
		result	*=	Jac_2d;

		return result*0.5;
	}


public:

	void init(const CS_Parameter_LO *p,const FO_Kinematic_Paramete_Set *FO_kin){
		_init(FO_kin,p);//deletes lumis
		_init_lumis();
	}



	_FO_CS_DY()
	:_FO_CS(){
//		NS_Utils::DOUT dout("_FO_CS_DY");
	}

	_FO_CS_DY(const CS_Parameter_LO *p,const FO_Kinematic_Paramete_Set *FO_kin)
	:_FO_CS(){
//		NS_Utils::DOUT dout("_FO_CS_DY","p,kin");
		init(p,FO_kin);
	}

	_FO_CS_DY(const _FO_CS_DY &CS)
	:_FO_CS(CS){
//		NS_Utils::DOUT dout("_FO_CS_DY","CS");
		if(p!=NULL) _init_lumis();
	}

	_FO_CS_DY& operator=(const _FO_CS_DY &CS){
//		NS_Utils::DOUT dout("_FO_CS_DY::operator=","CS");
		_exit_lumis();
		_FO_CS::operator =(CS);
		if(p!=NULL) _init_lumis();

		return *this;
	}

	~_FO_CS_DY(){
		_exit_lumis();
	}

};

}  // namespace NS_FO
}  // namespace CuTe


#endif //FO_CS_H_
