#ifndef FO_DY_H
#define FO_DY_H

//#include "konst.h"
#include "utils.h"
#include "FO_Parameter.h"

namespace CuTe {
namespace NS_FO{
namespace NS_DY {

using namespace NS_Utils::NS_pow;


inline double A_qG(const FO_NLO_Kinematic_Parameter& p){
//	Pol<double> erg(2,pow_2(p.s+p.t));
//	erg[1]	+= -p.s*p.t+p.M2*p.u;
//	erg[1]	*= 2.;
//	erg[0]	+= 2.*(-p.s*p.t+p.M2*p.u);
//	erg		*= -p.is*p.it;
//	erg.mul_this_by_Pol(p.Pref_qG,2);
	return -(p.s2+p.t2+2.*p.u*p.M2)*p.it*p.is;
}

inline double A_qbq(const FO_NLO_Kinematic_Parameter& p){
	return +(p.u2+p.t2+2.*p.s*p.M2)*p.it*p.iu;
}
///////////////////////////////////////////////////////
//                      B_qG						//
///////////////////////////////////////////////////////

// 1/w cancels in B1+C1
// Typos corrected
inline double B1_qG_finite(const FO_NNLO_Kinematic_Parameter&p, const FO_Scale_Parameter ps, double A_qG){
	double B1_qG(0.);
	double sPt = p.s +p.t;
	double isPt = 1./sPt;
	double sPu = p.s +p.u;
	double isPu = 1./sPu;
	double tPu = p.t +p.u;
	double itPu = 1./tPu;


//	B1_qG	+=	-A_qG[2]*(2.*C_F+C_A)
//				+A_qG[1]*(
//					3.*C_F
//					-2.*C_F*p.fu
//					+11./6.*C_A
//					+C_A*(p.fu-p.fs-p.ft)
//				);
	B1_qG	+=	A_qG*(
					-8.*C_F
					-C_F*pow_2(p.fu)
					-2.*zeta_2*(C_F-C_A)
					+0.5*C_A*(pow_2(p.fu)-pow_2(p.fs)-pow_2(p.ft))
					+C_A*(11./6.*ps.fmuR+p.f1t-p.f2t)
				);
	B1_qG	-=	C_F*(
					p.u*itPu	//	/(p.t+p.u)
					+p.u*isPu	//	/(p.s+p.u)
					+tPu*p.is
					+sPu*p.it
				)
				+(
					C_F*(4.*p.u2+2.*p.u*p.t+4.*p.s*p.u+p.s*p.t)*pow_2(isPu)
					+C_A*p.t*isPu
				)*p.ft;
	B1_qG	-=	(
					 C_F*(4.*p.u2+2.*p.s*p.u+4.*p.u*p.t+p.t*p.s)*itPu
					+C_A*p.s
				)*itPu*p.fs;
	B1_qG	-=	(2.*C_F-C_A)*(
					2.*(p.u*isPt+2.)*p.u*isPt*p.fu
					-p.M2*p.is*p.it*(p.s2+p.t2)*isPt
					+(p.u2+pow_2(sPu))*p.is*p.it*(p.f1t+p.f1u-p.ft*p.fu+3.*zeta_2)
					+(p.u2+pow_2(tPu))*p.is*p.it*(p.f1u-p.f2u)
				);
	return B1_qG;
}

// 1/w cancels in B2+C2
inline double B2_qG_finite(const FO_Scale_Parameter ps, double A_qG){
	return -A_qG*ps.fmuR/3.;
}

inline double B3_qG(const FO_NNLO_Kinematic_Parameter&p){
	double  u(p.u);
	double Q2(p.M2);
	double fu(p.fu);
//	cout << "B3_qG(p)" << "\tu="<<u << "\tQ2="<<Q2 << "\tfu="<<fu << endl;
	return (u+Q2)*(1.-Q2*fu/(u-Q2))/(u-Q2);
}

///////////////////////////////////////////////////////
//                      B_q_q						//
///////////////////////////////////////////////////////

//	1/w cancels in the sum B1+C1
inline double B1_qbq_finite(const FO_NNLO_Kinematic_Parameter&p, const FO_Scale_Parameter ps,double A_qbq){
	double B1_qbq(0.);
	double sPt = p.s +p.t;
	double sPu = p.s +p.u;
	double uPt = p.u +p.t;

	double isPt = 1./sPt;
	double isPu = 1./sPu;
	double iuPt = 1./uPt;
	//		double tPu = p.t +p.u;
	//		double itPu = 1./tPu;
//	B1_qbq	+=	-A_qbq[2]*(2.*C_F+C_A)
//							+A_qbq[1]*(
//									3.*C_F
//									-2.*C_F*p.fs
//									+11./6.*C_A
//									+C_A*(p.fs-p.fu-p.ft)
//							);
	B1_qbq	+=	A_qbq*(
					-8.*C_F
					-C_F*pow_2(p.fs)
					+zeta_2*(4.*C_F-C_A)
					+0.5*C_A*(pow_2(p.fs)-pow_2(p.ft+p.fu))
					+C_A*(11./6.*ps.fmuR+p.f1t+p.f1u)
				);
	B1_qbq	+=	C_F*(
					 p.s*isPt
					+p.s*isPu
					+sPt*p.iu
					+sPu*p.it
				)
				+(
					 C_F*(4.*p.s2+2.*p.s*p.t+4.*p.s*p.u+p.u*p.t)*pow_2(isPu)
					+C_A*p.t*isPu
				)*p.ft;
	B1_qbq	+=	(
					C_F*(4.*p.s2+2.*p.s*p.u+4.*p.s*p.t+p.t*p.u)*pow_2(isPt)
					+C_A*p.u*isPt
				)*p.fu;
	B1_qbq	+=	(2.*C_F-C_A)*(
					 2.*(p.s*iuPt+2.)*p.s*iuPt*p.fs
					-p.M2*p.iu*p.it*(p.u2+p.t2)*iuPt
					+(p.s2+pow_2(sPu))*p.iu*p.it*(p.f1t-p.f2t)
					+(p.s2+pow_2(sPt))*p.iu*p.it*(p.f1u-p.f2u)
				);
	return B1_qbq;
}

//	1/w cancels in the sum B2+D0
inline double B2_qbq_finite(const FO_Scale_Parameter ps, double A_qbq){
	return -A_qbq*ps.fmuR/3.;
}

inline double B3_qbq(const FO_NNLO_Kinematic_Parameter&p){
	return -(p.s+p.M2)*(1.-p.M2*p.fs/(p.s-p.M2))/(p.s-p.M2);
}

///////////////////////////////////////////////////////
//                     C_qG						//
///////////////////////////////////////////////////////
// 1/w cancels in B1+C1
inline double C1_qG_finite(const FO_NNLO_Kinematic_Parameter&p, const FO_Scale_Parameter ps, double A_qG){
	double C1_qG(0.);
//	C1_qG	+=	 A_qG[2]*(2.*C_F+C_A)
//				-A_qG[1]*(
//						3.*C_F
//						-2.*C_F*p.fu
//						+11./6.*C_A
//						+C_F*(p.fu-p.fs-p.ft)
//				);
	C1_qG	+=	A_qG*(
					+C_F*(
						3.5
						+2.*ps.fmuF*(p.fu-p.fA)
						+pow_2(p.fA)
						-1.5*(ps.fmuF+p.fA)
					)
					+C_A*(
						zeta_2
						+0.5*pow_2(p.fs-p.ft-p.fu)
						+2.*p.ft*(ps.fmuF-p.fA)
						+2.*p.fA*(p.fs-p.fu-ps.fmuF+p.fA)
						-11./6.*ps.fmuF
					)
				);
	return C1_qG;
}

// 1/w cancels in B2+C2
inline double C2_qG_finite(const FO_Scale_Parameter ps,double A_qG){
	return	A_qG*ps.fmuF/3.;

}

inline double _C3_qG_Plus_0(const FO_NNLO_Kinematic_Parameter&p, const FO_Scale_Parameter ps){
	double C3_qG_p(0.);
	C3_qG_p	+=	C_F*(-1.5+p.fsu+2.*p.ftu-2.*ps.fmuF+(p.t+p.u)*p.il*p.fl);
	C3_qG_p	+=	C_A*(2.*p.fstu+0.5*(p.fst-p.fsu)-p.ftu-2.*ps.fmuF);
	C3_qG_p	*=	(-p.s*p.it-p.t*p.is-2.*p.u*p.M2*p.is*p.it);
	return C3_qG_p;
}

inline double _C3_qG_Plus_1(const FO_NNLO_Kinematic_Parameter&p){
	return (-p.s*p.it-p.t*p.is-2.*p.u*p.M2*p.is*p.it)*(2.*C_F+4.*C_A);
}

inline double _C3_qG_finite(const FO_NNLO_Kinematic_Parameter&p, const FO_Scale_Parameter ps){
	const double Q2(p.M2);
	const double s(p.s)			,is(p.is),is2(p.is2);
	const double t(p.t),t2(p.t2),it(p.it),it2(p.it2);
	const double u(p.u),u2(p.u2),iu(p.iu),iu2(p.iu2);
	const double x2(p.x2),x4(p.x4);
	const double du(p.du),dt(p.dt);
	const double dst(p.dst),dtu(p.dtu);
	const double fM2(ps.fmuF),fx2(p.fx2);
	const double fst(p.fst),fsu(p.fsu),ftu(p.ftu);
	const double fstu(p.fstu);
	const double il(p.il),il2(p.il2),il3(p.il3);
	const double fl(p.fl),flt(p.flt);

	 return du*(2*C_F*(fM2 - fx2 - it*s) + C_A*(-1 + it*s)) +
	   2*pow_3(dt)*(-3*C_A + 4*C_F - (-C_A + C_F)*(-fM2 + fx2))*s*t +
	   pow_2(dt)*(C_F*(3*s - 8*t - 4*u) - C_A*(9*s - 4*t + 4*u) +
	      2*(-fM2 + fx2)*(C_F*(t + u) + C_A*(3*s + 2*u))) +
	   dt*(-(C_F*(-fM2 + fx2)*it*(s + 2*u)) + C_A*(7 - it*(3*s + 2*u)) +
	      (C_F*(-fM2 + fx2) - C_A*(-fM2 + (flt + fst)/4. + fx2 + (fstu - ftu + fx2)/2.))*
	       (-2*(it*s + is*t) + 4*(1 + is*u)*(1 - it*u)) +
	      C_A*((2*flt + 2*fst)*it*(s + u) + 2*(-fM2 + fx2)*(-2 + it*(5*s + 4*u))) -
	      C_F*(4 + it*(s - 2*u) + is*(-3*t + 2*u) - is*it*u2)) +
	   ((-C_A/2. + C_F)*(ftu*is*(2*t + 4*u) + it*(s + 2*u)*(fsu + fl*il*(t + u))))/x2 +
	   C_F*dtu*(-t + 4*u*(it*Q2 + is*(-Q2 + t)*(1 - it*u)) +
	      2*(fM2 - ftu - fx2)*(t + 2*it*s*(Q2 + 2*u) + 2*is*(t - 2*u + 2*it*u2)*(u - x2) +
	         4*it*u*(2*u - x2) - x2) + x2) + il2*
	    (C_A*(-1 + it*u)*(s + x2) + C_F*((it*s*(-3*t + u))/2. + (3*il2*s*pow_2(t - u)*(1 + it*u))/2. +
	         (1 - it*u)*((-7*u)/4. + x2) + (t - u)*(2.75 + (is*(3*t + 3*u))/2. - 2*is*x2))) +
	   fl*il3*(C_A*(1 - it*u)*(s*(s - x2) + ((t + u)*(Q2 + s - x2))/4.) +
	      C_F*(5*s*u - 2*s*(t + it*u2) - (s*(-3 + it*u)*(s - x2))/2. +
	         (3*il2*s*(1 - it*u)*(-t2 + u2)*(Q2 + s - x2))/4. + 4*(t - u)*x2 +
	         (-t2 + u2)*(2.75 - (it*u)/2. + (is*(Q2 - x2))/4. + it*x2))) +
	   C_F*(-2*is + (3*it)/4. - iu + 3*is*it*u - 2*ftu*(-2*is + it + 4*is*it*u) +
	      flt*is*it*(5*Q2 + u - 3*x2) - is*iu*(t + x2) + Q2*(-it2/2. + is*iu2*x2) +
	      fsu*(5*is - is*iu2*Q2*x2 + iu*(1 + 2*is*t - 3*is*x2)) +
	      (-fM2 + fx2)*(8*is + 3*it + iu - 4*is*it*u + is*iu*(2*t - 3*x2) - Q2*(it2 + is*iu2*x2)) +
	      (-flt + fsu - 2*ftu)*is*it*(-x2 + is*(2*u*(-u + x2) - x4))) +
	   C_F*dst*(-flt + fsu - 2*ftu)*is2*(1 - it*x2)*(-2*u*(-u + x2) + x4) +
	   fl*il*(C_F*((it*(11*s - 2*u - 4*x2))/4. + (is*(t - u - 2*x2))/4. - 16*is*it*Q2*x2 +
	         2*is*(1 + it*u)*(5*Q2 + x2)) - (C_A*
	         ((it*(t + u))/2. + 3*is*(t - 3*x2) - 7*it*(s + is*u*x2) + 5*(is*u - it*(u + x2)) +
	           2*is*it*(u2 + 3*x4)))/2.) + C_A*
	    (-is - it + 2*is*it*Q2*(1 - it*Q2) + fsu*is*(1 - it*u) + ftu*is*(-1 + 2*it*u) -
	      2*flt*(is + 3*it + 2*is*it*(u - x2)) + 2*is*it*x2 - 12*is*it2*Q2*(1 - it*Q2)*x2*(1 - it*x2) +
	      (fstu + (-fst - fsu)/2. + fx2)*is*(4 - 2*it*u - it*x2) +
	      fst*((15*is)/2. - 2*it2*Q2 - (9*is*it*x2)/2.) +
	      2*(-fM2 + fx2)*(6*is - 2*it - 2*is*it*Q2 - 3*is*it*x2) + it2*(3*Q2 + 4*(1 - 3*it*Q2)*x2) -
	      2*it2*(s + is*x4) + 2*(-fM2 + fst + fx2)*
	       (-4*is*it*Q2*(1 - it*Q2) + (2*Q2*x2)*it*it2 + 2*is*it2*Q2*(1 - it*Q2)*x2*(1 - it*x2) +
	         it2*(s + is*x4)));
//	C3_qG_f	+=	p.is2*(C_F-0.5*C_A)*(
//			(p.s+2.*p.u)*p.it*((p.t+p.u)*p.il*p.fl+p.fsu)
//			+(2.*p.t+4.*p.u)*p.is*p.ftu
//	);
//	C3_qG_f	+=	1./p.x2*(C_F-0.5*C_A)*(
//			(p.s+2.*p.u)*p.it*((p.t+p.u)*p.il*p.fl+p.fsu)
//			+(2.*p.t+4.*p.u)*p.is*p.ftu
//	);
//	C3_qG_f	+=	p.fl*p.il3*(
//			C_F*(
//					0.75*p.s*p.il2*(p.s+p.M2-p.x2)*(p.u2-p.t2)*(1.-p.u*p.it)
//					+(p.u2-p.t2)*(
//							0.25*(p.M2-p.x2)*p.is
//							+0.25*11.
//							+p.x2*p.it
//							-0.5*p.u*p.it
//					)
//					-2.*p.s*(p.t+p.u2*p.it)
//					-0.5*p.s*(p.s-p.x2)*(p.u*p.it-3.)
//					+4.*p.x2*(p.t-p.u)
//					+5.*p.s*p.u
//			)
//			+C_A*(1.-p.u*p.it)*(
//					0.25*(p.t+p.u)*(p.s+p.M2-p.x2)
//					+p.s*(p.s-p.x2)
//			)
//	);
//	C3_qG_f	+=	p.fl*p.il*(
//			C_F*(
//					0.25*(p.t-p.u-2.*p.x2)*p.is
//					+0.25*(11.*p.s-2.*p.u-4.*p.x2)*p.it
//					+2.*p.is*(1.+p.u*p.it)*(5.*p.M2+p.x2)
//					-16.*p.M2*p.x2*p.is*p.it
//			)
//			-0.5*C_A*(
//					0.5*(p.t+p.u)*p.it
//					+5.*(p.u*p.is-(p.u+p.x2)*p.it)
//					-7.*p.it*(p.s+p.u*p.x2*p.is)
//					+3.*p.is*(p.t-3.*p.x2)
//					+2.*p.is*p.it*(p.u2+3.*p.x4)
//			)
//	);
//	C3_qG_f	+=	p.il2*(
//			C_F*(
//					1.5*p.s*p.il2*pow_2(p.t-p.u)*(1.+p.u*p.it)
//					+0.5*p.s*p.it*(p.u-3.*p.t)
//					+(1.-p.u*p.it)*(p.x2-1.75*p.u)
//					+(p.t-p.u)*(2.75+1.5*(p.t+p.u)-2.*p.x2*p.is)
//			)
//			+C_A*(p.u*p.it-1.)*(p.s+p.x2)
//	);
//	C3_qG_f	+=	2.*pow_3(p.dt)*p.s*p.t*(
//			4.*C_F
//			-3.*C_A
//			-(C_F-C_A)*(p.fx2-ps.fmu)
//	);
//	C3_qG_f	+=	pow_2(p.dt)*(
//			C_F*(3.*p.s-8.*p.t-4.*p.u)
//			-C_A*(9.*p.s-4.*p.t+4.*p.u)
//			+2.*(p.fx2-ps.fmu)*(
//					C_F*(p.t+p.u)
//					+C_A*(3.*p.s+2.*p.u)
//			)
//	);
//	C3_qG_f	+=	p.dt*(
//			(
//					C_F*(p.fx2-ps.fmu)
//					-C_A*(
//							p.fx2
//							-ps.fmu
//							+0.5*(p.fstu+p.fx2-p.ftu)
//							+0.25*(p.fst+p.flt)
//					)
//			)*(
//					4.*(1.+p.u*p.is)*(1.-p.u*p.it)
//					-2.*(p.s*p.it+p.t*p.is)
//			)
//			-C_F*(p.fx2-ps.fmu)*(p.s+2.*p.u)*p.it
//			+C_A*(
//					(2.*p.flt+2.*p.fst)*(p.u+p.s)*p.it
//					+2.*(p.fx2-ps.fmu)*((5.*p.s+4.*p.u)*p.it-2.)
//			)
//			-C_F*(
//					4.
//					+(p.s-2.*p.u)*p.it
//					+(2.*p.u-3.*p.t)*p.is
//					-p.u2*p.is*p.it
//			)
//			+C_A*(7.-(3.*p.s+2.*p.u)*p.it)
//	);
//	C3_qG_f	+=	p.du*(
//			2.*C_F*(ps.fmu-p.fx2-p.s*p.it)
//			+C_A*(p.s*p.it-1.)
//	);
//	C3_qG_f	+=	p.dst*C_F*(p.fsu-2.*p.ftu-p.flt)
//								*(1.-p.x2*p.it)
//								*(p.x4-2.*p.u*(p.x2-p.u))*p.is2;
//	C3_qG_f	+=	p.dtu*C_F*(
//			2.*(ps.fmu-p.fx2-p.ftu)*(
//					2.*p.s*p.it*(p.M2+2.*p.u)
//					+p.t
//					-p.x2
//					+4.*p.u*p.it*(2.*p.u-p.x2)
//					+2.*p.is*(p.u-p.x2)*(p.t-2.*p.u+2.*p.u2*p.it)
//			)
//			+p.x2
//			-p.t
//			+4.*p.u*(p.M2*p.it+(1.-p.u*p.it)*(p.t-p.M2)*p.is)
//	);
//	C3_qG_f	+=	C_F*(
//			p.is*p.it*((2.*p.u*(p.x2-p.u)-p.x4)*p.is-p.x2)
//			*(p.fsu-2.*p.ftu-p.flt)
//			-2.*(p.it-2.*p.is+4.*p.u*p.is*p.it)*p.ftu
//			+(p.u-3.*p.x2+5.*p.M2)*p.is*p.it*p.flt
//			+(5.*p.is+p.iu*(1.+2.*p.t*p.is-3.*p.x2*p.is)-p.x2*p.M2*p.is*p.iu2)*p.fsu
//			+(
//					8.*p.is
//					+3.*p.it
//					+p.iu
//					-4.*p.u*p.is*p.it
//					+(2.*p.t-3.*p.x2)*p.is*p.iu
//					-p.M2*(p.it2+p.x2*p.is*p.iu2)
//			)*(p.fx2-ps.fmu)
//			-2.*p.is
//			+0.75*p.it
//			-p.iu
//			+3.*p.u*p.is*p.it
//			-(p.t+p.x2)*p.is*p.iu
//			+p.M2*(p.x2*p.is*p.iu2-0.5*p.it2)
//	);
//	C3_qG_f	+=	C_A*(
//			2.*(p.fx2-ps.fmu)*(6.*p.is-2.*p.it-2.*p.M2*p.is*p.it-3.*p.x2*p.is*p.it)
//			+(7.5*p.is-4.5*p.x2*p.is*p.it-2.*p.M2*p.it2)*p.fst
//			+2.*(
//					p.it2*(p.s+p.x4*p.is)
//					+2.*p.x2*p.M2*p.it2*p.it
//					-4.*p.M2*p.is*p.it*(1.-p.M2*p.it)
//					+2.*p.x2*p.M2*p.is*p.it2*(1.-p.x2*p.it)*(1.-p.M2*p.it)
//			)*(p.fx2-ps.fmu+p.fst)
//			+p.is*(4.-2.*p.u*p.it-p.x2*p.it)*(p.fx2+p.fstu-0.5*(p.fst+p.fsu))
//			+p.is*(1.-p.u*p.it)*p.fsu
//			+p.is*(2.*p.u*p.it-1.)*p.ftu
//			-2.*(3.*p.it+p.is+2.*p.is*p.it*(p.u-p.x2))*p.flt
//			-p.is
//			-p.it
//			+2.*p.x2*p.is*p.it
//			+p.it2*(3.*p.M2+4.*p.x2*(1.-3.*p.M2*p.it))
//			-2.*p.it2*(p.s+p.x4*p.is)
//			+2.*p.M2*p.is*p.it*(1.-p.M2*p.it)
//			-12.*p.x2*p.M2*p.is*p.it2*(1-p.x2*p.it)*(1.-p.M2*p.it)
//	);
//	return C3_qG_f;
}

///////////////////////////////////////////////////////
//                     C_G_G						//
///////////////////////////////////////////////////////
inline double C_G_G(const FO_NNLO_Kinematic_Parameter& p, const FO_Scale_Parameter ps){
	double C_G_G(0.);
	const double s(p.s),is(p.is),is2(p.is2);
	const double t(p.t),t2(p.t2),it(p.it);
	const double u(p.u),u2(p.u2);
	const double Q2(p.M2),x2(p.x2);
	const double ds(p.ds),dt(p.dt),du(p.du);
	const double dst(p.dst),dtu(p.dtu);
	const double il(p.il),il2(p.il2);
	const double fl(p.fl),flt(p.flt);
	const double fst(p.fst),fsu(p.fsu),ftu(p.ftu),fstu(p.fstu);
	const double fx2(p.fx2),fmu(ps.fmuF);

	C_G_G += C_A*ds*fl*il*(
			t2*il2*(
					1.5*il2*( 2.*s-t-u)*(u2-t2)
					+2.*s
					-3.*t
					+u
					+(t2-u2)*is
			)
			-1.5*s
			-t*( 4.5 + 5.*t*is +3.*u*is)
	);
	C_G_G += C_A*fl*il*(
			t2*il2*( 1.5*il2*(u2-t2)+1.)*(Q2-x2)*is
			-11.*0.25*is*(Q2-x2)
			-4.*(1.+t*is)
	);
	C_G_G += C_F*fl*il*( 2.*ds*( t+u-2.*s )+6. )
						+C_A*t*il2*(
								3.*pow_2(t-u)*il2*(
										1.
										+0.5*(t+u)*is
										-2.*s*ds
								)
								+ds*( 2.*s - 3.*t +3.*u)
								-1.
								-u*is
								+x2*(t-u)*is2
								+(ds-0.5*is)*pow_2(t-u)*is
						);
	C_G_G += 4.*C_F*dtu*(
			(fx2-fmu+ftu)*( 0.5*s+ 2.*t*(1.+t*is))
			-t*( 1. + (t+u)*is)
	);
	C_G_G += 2.*ds*(
			(C_A-2.*C_F)*dt*t2*is*(fst+flt)
			+C_F*dst*(fsu-flt-2.*ftu)*(s+2.*u*(1.+u*is))
	);
	C_G_G += ds*(
			2.*C_F*(
					fst
					-(3.+4.*(t+u)*is)*flt
					-(4.+8.*t*is)*ftu
			)
			+0.5*C_A*(
					(fst+flt)*(2.+(u+3.*t)*is)
					+1.
					+t*(t-u)*is2
			)
	);
	C_G_G += 2.*C_F*dst*(2.*ftu+flt-fsu)*(2.+(t+u)*is)
						+8.*t*pow_2(dt)*(C_F*(1.-fx2+fmu)+C_A);
	C_G_G += (4.*C_F-2.*C_A)*dt*du*t*u*is*(fx2-ftu+fstu);
	C_G_G += dt*(
			(C_A-2.*C_F)*(
					(2.+(t+u)*is)*(fst+flt)
					+2.*is*(u-t)*(fx2+fstu-ftu)
			)
			-8.*C_F*(fx2-fmu-1.)
			+4.*C_A*(2.+(u-t)*is)
	);
	C_G_G += C_F*(
			2.*(fx2-fmu)*(
					2.*is
					-2.*it
					-is*it*(u+x2*Q2*it)
			)
			+4.*is*(fx2+fstu)
			-2.*it*(2.+is*(u+x2*Q2*it))*fst
			+is*it*(4.*Q2*(1.+x2*it)-2.*u)
	)
						-C_A*is*(
								2.*(fx2+fstu)
								+0.5*fst
								+2.5*flt
								+15.*0.25
						);
	return C_G_G;
}

///////////////////////////////////////////////////////
//                     C_qbq						//
///////////////////////////////////////////////////////
//	1/w cancels in the sum B1+C1
inline double C1_qbq_finite(const FO_NNLO_Kinematic_Parameter&p, const FO_Scale_Parameter ps, double A_qbq){
	double C1_qbq(0.);
//	C1_qbq	+=	 A_qbq[2]*(2.*C_F+C_A)
//							-A_qbq[1]*(
//									3.*C_F
//									-2.*C_F*p.fs
//									+11./6.*C_A
//									+C_A*(p.fs-p.fu-p.ft)
//							);
	C1_qbq	+=	A_qbq*(
					+C_A*(
						67./18.
						-11./6.*p.fA
						+pow_2(p.fA)
					)
					+C_F*(
						2.*p.ft
						+2.*p.fu
						-4.*p.fA
						-3.
					)*ps.fmuF
					+(C_F-0.5*C_A)*(
						2.*zeta_2
						+pow_2(2.*p.fA+p.fs-p.ft-p.fu)
					)
				);
	return C1_qbq;
}

inline double _C2_qbq_Plus_0(const FO_NNLO_Kinematic_Parameter&p, const FO_Scale_Parameter ps){
	double C3_qG_p(0.);
	C3_qG_p	+=	-11./6.*C_A
			+2.*C_F*(p.ftu-2.*ps.fmuF)
			+(2.*C_F-C_A)*(2.*p.fstu-p.ftu)
			;
	C3_qG_p	*=	0.5*(p.u*p.it+p.t*p.iu+2.*p.M2*p.s*p.iu*p.it);
	return C3_qG_p;
}

inline double _C2_qbq_Plus_1(const FO_NLO_Kinematic_Parameter&p){
	return 0.5*(p.u*p.it+p.t*p.iu+2.*p.M2*p.s*p.iu*p.it)*(8.*C_F-2.*C_A);
}

inline double _C2_qbq_finite(const FO_NNLO_Kinematic_Parameter&p, const FO_Scale_Parameter ps){
	double C2_qbq(0.);
	C2_qbq	+=	C_F*(
			p.s*pow_2(p.dt)
	+2.*p.dt
	-p.s*p.it*p.iu
	+p.dtu*(p.x2+2.*p.M2*(p.u-p.x2)*p.it)
	-2.*p.dt*p.M2*(p.iu-p.it)
	);
	C2_qbq	+=	C_A*(
			-11./6.*p.s*p.it*p.iu
			+pow_2(p.dt)*p.s2*p.iu*(1.5*p.x2*p.it-2.)
			+2.*p.dt*p.s*p.iu
			+p.M2*p.it2/3.
	);
	C2_qbq	+=	(2.*C_F-C_A)*(
			p.dt*p.iu*pow_2(p.M2-p.u)*(p.du-p.it)*(p.fstu+p.fx2-p.ftu)
			+(p.s-p.M2)*p.it*p.iu*(p.fstu+p.fx2)
	);
	C2_qbq	+=	p.ftu*(
			C_F*p.dtu*(
					4.*p.M2*p.it*p.iu*pow_2(p.M2-p.t)
	+2.*(p.M2+p.s)
	-p.x2
			)
			+C_F*(p.x2-2.*p.s)*p.it*p.iu
			-C_A*p.M2*p.it*p.iu
	);
	C2_qbq	+=	C_F*(ps.fmuF-p.fx2)*(
			p.dtu*(4.*(p.u-p.M2)-p.x2-4.*p.M2*p.it*p.iu*pow_2(p.u-p.M2))
			+pow_2(p.dt)*(p.M2-p.u)
			-p.dt*p.it*(2.*p.M2-p.u)
			-2.*p.it
			+p.M2*p.it2
			+4.*p.M2*p.it*p.iu
	);
	return C2_qbq;
}

///////////////////////////////////////////////////////
//                     D						//
///////////////////////////////////////////////////////

//	1/w cancels in the sum B2+D0
inline double D_0_aa_finite(const FO_NNLO_Kinematic_Parameter& p, double A_qbq){
//	return	(	 A_qbq[1]
//	      	 	        +A_qbq[0]*(p.fA-5./3.)
//	)/3.;
	return A_qbq*(3.*p.fA-5.)/9.;
}

inline double _D_aa_Plus_0(const FO_NLO_Kinematic_Parameter& p){
	return	(p.u*p.it+p.t*p.iu+2.*p.s*p.M2*p.it*p.iu)/6.;
}

inline double _D_aa_finite(const FO_NLO_Kinematic_Parameter& p){
	return	(p.s*p.it*p.iu-p.M2*p.it2)/3.;
}

inline double D_ab(const FO_NNLO_Kinematic_Parameter& p){
	return	0.5*(
			p.fl*p.M2*p.il*p.it*(
					(2.*p.s+p.t+p.u)*p.il2*(
							3.*p.s*pow_2(p.t-p.u)*p.il2
							+p.u
							-p.t
							-p.s
					)
					-1.
			)
			+0.5*(2.*p.s+p.t+p.u)*p.il2*p.it*(
					-3.*p.il2*(p.s+p.M2-p.x2)*pow_2(p.t-p.u)
					+3.*p.t
					-p.u
					+2.*p.s
					-2.*p.x2
			)
	);
}

//inline double D_bb(const FO_NLO_Kinematic_Parameter& p){
//	double D_bb(0.);
//	D_bb	+=	 0.5*p.ds*(p.u*p.is-1.)
//				-1.25*p.is
//				+p.il2*(
//					p.u*p.ds*(
//						-2.*p.s
//						+1.5*p.u*p.is*(p.t-p.u)
//						+4.*p.u
//						-2.*p.t
//					)
//					+0.5*p.u*p.is*(2.*p.s+2.*p.x2+p.t-p.u)
//				)
//				;
//	D_bb	+=	3.*p.u2*p.il4*(p.u-p.t)*(
//					 p.ds*(2.*p.s-p.u-p.t)
//					-(p.s+p.x2)*p.is
//				)
//				+p.fl*p.il*(
//						p.ds*(2.*p.u2*p.is+2.5*p.u+1.5*p.s)
//						+0.75
//						+p.u*p.is
//						-0.5*p.x2*p.is
//				)
//				;
//	D_bb	+=	 p.fl*p.il3*(
//					p.u2*p.ds*(
//						3.*p.u
//						-p.t
//						-p.u2*p.is
//						+p.t2*p.is
//						-2.*p.s
//					)
//					+p.u*p.is*(
//						2.*p.x2*p.t
//						-p.u*p.t
//						-2.*p.x4
//						+4.*p.u*p.x2
//						-3.*p.u2
//						+2.*p.s*p.x2
//						-p.u*p.s
//					)
//				);
//	D_bb	+=	 p.fl*p.il5*(
//					3.*p.ds*p.u2*(p.u2-p.t2)*(p.s-p.M2)
//					+3.*p.u2*p.M2*p.is*(p.u-p.t)*(p.u+p.t-2.*p.x2)
//				);
//	return 0.5*D_bb;
//}

inline double D_bb(const FO_NNLO_Kinematic_Parameter& p){
	double D_bb(0.);
	const double Q2(p.M2);
	const double s(p.s),is(p.is);
	const double t(p.t),t2(p.t2);
	const double u(p.u),u2(p.u2);
	const double x2(p.x2),x4(p.x4);
	const double ds(p.ds);
	const double il(p.il),il2(p.il2),il3(p.il3),il4(p.il4),il5(p.il5);
	const double fl(p.fl);


	D_bb	+=	 0.5*ds*(u*is-1.)
				-1.25*is
				+il2*(
					u*ds*(
						-2.*s
						+1.5*u*is*(t-u)
						+4.*u
						-2.*t
					)
					+0.5*u*is*(2.*s+2.*x2+t-u)
				)
				;
	D_bb	+=	3.*u2*il4*(u-t)*(
					 ds*(2.*s-u-t)
					-(s+x2)*is
				)
				+fl*il*(
						ds*(2.*u2*is+2.5*u+1.5*s)
						+0.75
						+u*is
						-0.5*x2*is
				)
				;
	D_bb	+=	 fl*il3*(
					u2*ds*(
						3.*u
						-t
						-u2*is
						+t2*is
						-2.*s
					)
					+u*is*(
						2.*x2*t
						-u*t
						-2.*x4
						+4.*u*x2
						-3.*u2
						+2.*s*x2
						-u*s
					)
				);
	D_bb	+=	 fl*il5*(
					3.*ds*u2*(u2-t2)*(s-Q2)
					+3.*u2*Q2*is*(u-t)*(u+t-2.*x2)
				);
	return 0.5*D_bb;
}

inline double D_ac(const FO_NNLO_Kinematic_Parameter& p){
	double D_ac(0.);
	D_ac	+=	p.fl*p.il*(
			3.*p.s2*p.M2*p.il4*pow_2(p.t-p.u)*(p.it+p.iu)
			+p.s*p.M2*p.il2*(5.*p.s*p.it-7.*p.s*p.iu+4.*p.u*p.it-4.)
			+1./p.x2*(2.*p.s2*p.iu+(p.M2*(2.*p.s+p.u)-p.u*p.s)*p.it)
			+(2.*p.x2-p.u)*p.it
			-2.
	);
	D_ac	+=	3.*p.s*p.M2*p.il4*p.it*p.iu*(2.*p.x2-p.u-p.t)*pow_2(p.t-p.u)
							+p.il2*(
									p.s*(p.s-p.x2)*(7.*p.iu-5.*p.it)
									+0.5*(p.t*p.iu-p.u*p.it)*(10.*p.s+p.t+3.*p.u-2.*p.x2)
									+2.*p.s
									-2.*p.x2*(1.-p.u*p.it)
							)
							;
	D_ac	+=	p.dt*p.s2*p.iu*(p.dt+3.*p.it)
							+p.fst*p.it*(
									1./p.x2*(2.*p.s2*p.iu+2.*p.s+p.u)
									-2.*p.M2*p.it
							)
							;
	D_ac	+=	-0.5*p.iu+1.5*p.it-3.*p.M2*p.it2;
	return (C_F-0.5*C_A)*D_ac;
}

inline double D_bc(const FO_NNLO_Kinematic_Parameter& p){
	double D_bc(0.);
	D_bc	+=	p.fl*p.il*(
			6.*p.s*p.x2*p.M2*p.il4*p.it*pow_2(p.t-p.u)
	+p.il2*(
			2.*p.s*p.u
			+(p.x2*(1.+p.t*p.is)-0.5*p.t*p.is*(p.t+p.u))*(p.t+p.u)*(1.-p.u*p.it)
			+1.5*p.s*(p.t-p.u)*(1.-p.u*p.it)
			-4.*p.x2*p.M2*(1.+(p.s-p.u)*p.it)
	)
	+p.ds*(6.*p.s+2.*p.u+0.5*(p.t2-p.u2)*p.is)
	-1.
	+0.5*(p.s+4.*p.u-2.*p.x2)*p.it
	+(3.*p.t+2.*p.u-6.*p.x2)*p.is
	+(p.u2-2.*p.u*p.x2+4.*p.x4)*p.is*p.it
	);
	D_bc	+=	 3.*p.x2*p.il4*pow_2(p.t-p.u)*(1.+p.u*p.it-2.*p.M2*p.it)
							+p.il2*(
									-0.5*p.t
									+3.*p.u
									+p.x2
									+(p.t2-p.u2)*p.is
									-0.5*p.u*p.it*(p.u+6.*p.x2)
							)
							;
	D_bc	+=	 2.*p.dt
			+p.ds*(p.fst+p.flt)*(2.*(p.s+p.u)*p.it+0.5*p.is*(p.t+p.u2*p.it))
			;
	D_bc	+=	 0.5*p.fst*(2.*p.it+3.*p.is+(p.u-2.*p.x2)*p.is*p.it)
							+0.5*p.flt*(-2.*p.it+p.is-(p.u+2.*p.x2)*p.is*p.it)
							;
	D_bc	+=	 0.5*p.it
			-p.is
			+p.is*p.it*(2.*p.u-4.*p.x2*p.M2*p.it)
			;
	return (C_F-0.5*C_A)*D_bc;
}

inline double D_cc(const FO_NNLO_Kinematic_Parameter& p, const FO_Scale_Parameter ps){
	double D_cc(0.);
	D_cc	+=	 p.fl*p.il*(p.s*p.il2*p.it*(p.u2-p.t2)+(3.*p.s+2.*p.u)*p.it)
							+(p.t+p.u-2.*p.x2)*p.il2*(1.-p.u*p.it)
							+p.dt*(p.fx2-ps.fmuF)*(p.s*p.dt+4.*p.s*p.it+2.*p.u*p.it)
							;
	D_cc	+=	(p.fst-ps.fmuF+p.fx2)*(
			pow_2(p.s+p.x2-p.M2)*p.is*p.it2
			+p.is*pow_2(p.u*p.it-2.*p.x2*p.M2*p.it2)
	+4.*p.it*(p.M2*p.it-1.)
	)
	+2.*p.fst*p.it*(1.-p.M2*p.it)
	;
	D_cc	+=	-p.dt*(p.s*p.dt-1.-p.u*p.it)
							+2.*p.it*(p.u*p.it+p.x2*(p.it+p.is-p.x2*p.is*p.it))
							+p.it*(p.M2*p.it-1.)
							+12.*p.x2*p.M2*p.is*p.it2*p.it*(p.u-p.x2*p.M2*p.it)
							;
	return 0.5*D_cc;
}

inline double D_LL_cd(const FO_NNLO_Kinematic_Parameter& p){
	double D_cd(0.);
	D_cd	+=	p.fl*p.il*(
					-p.s*p.il2*p.it*pow_2(p.t-p.u)
					+2.*p.ds*(p.s-p.t)
					+(3.*p.s+2.*p.x2)*p.it
					-1.
				);
	D_cd	+=	 (p.u-2.*p.x2)*p.il2*(p.u*p.it-1.)
				+p.ds*p.dst*p.it*(p.fsu-p.flt-2.*p.ftu)*(2.*p.s*(p.s+p.u)+p.u2)
							;
	D_cd	+=	p.ds*(
					(p.fst-p.ftu)*(1.+p.u*p.it+2.*p.t*p.iu+4.*p.s*(p.M2+p.x2)*p.it*p.iu)
					-(p.flt+p.ftu)*(1.+p.u*p.it)
				);
	D_cd	+=	 p.dst*(2.*p.ftu-p.fsu+p.flt)*(1.+(2.*p.s+p.u)*p.it)
							+2.*p.dt*p.s*p.iu
							+p.ftu*p.it*p.iu*(p.s+p.x2)
							-p.it*(p.fst+p.flt+1.-p.s*p.iu)
							;
	return 0.5*D_cd;
}

inline double D_LR_cd(const FO_NNLO_Kinematic_Parameter& p){
	double D_cd(0.);
	D_cd	+=	2.*p.ds*p.s2*p.it*(
					 p.dst*(p.fsu-p.flt-2.*p.ftu)
					+2.*p.iu*(p.fst-p.ftu)
				)
				+p.dst*(2.*p.ftu-p.fsu+p.flt)*(1.+2.*p.s*p.it-p.u*p.it)
				;
	D_cd	+=	 2.*p.it*p.iu*(p.ftu-p.fsu)*(p.s+p.x2-p.u)
				-2.*p.it*(p.ftu+p.flt);
	return 0.5*D_cd;
}

///////////////////////////////////////////////////////
//                     E	 						//
///////////////////////////////////////////////////////
inline double E_ac(const FO_NNLO_Kinematic_Parameter& p){
	return (C_F-0.5*C_A)*(
			(p.fst+p.flt)*(
					4.*p.dt*p.ds*p.s2*p.it
					-2.*p.dt*(1.-p.u*p.it)
					-4.*p.it
			)
			+4.*p.M2*p.it2*p.fst
	);
}

inline double E_ad(const FO_NNLO_Kinematic_Parameter& p){
	return (C_F-0.5*C_A)*(
			p.is*(2.*p.fx2-p.fst-p.fsu+2.*p.fstu)*(
					-2.
					-p.u*p.it
					-p.t*p.iu
					+2.*p.x2*p.it*p.iu*(p.M2-p.s)
			)
			-2.*p.is*(
					2.
					+p.u*p.it
					+p.t*p.iu
					-2.*p.x2*p.M2*(p.it2+p.iu2)
			)
	);
}


//****	qG

//************	NLO_Lumi

inline double nnlo_qg_nlo_delta(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps, double A_LO){
	// B2 + C2 cancel with muR=muF
	return B1_qG_finite(p,ps,A_LO) + C1_qG_finite(p,ps,A_LO);
}

inline double nnlo_qg_nlo_plus0(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps, double A_LO){
	return _C3_qG_Plus_0(p,ps);
}

inline double nnlo_qg_nlo_plus1(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps, double A_LO){
	return _C3_qG_Plus_1(p);
}

inline double nnlo_qg_nlo_finit(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps, double A_LO){
	return _C3_qG_finite(p,ps);
}
//************	NNLO_Lumi

inline double nnlo_qg_nnlo_delta(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps, double A_LO){
	return B3_qG(p);
}




//****	qbq

//************	NLO_Lumi
inline double nnlo_qbq_nlo_delta(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps, double A_LO){
	double nf(ps.nf);
	return 	(
			 B1_qbq_finite(p,ps,A_LO)
			+C1_qbq_finite(p,ps,A_LO)
			+nf*(
				 B2_qbq_finite(ps,A_LO)
				+D_0_aa_finite(p,A_LO)
				)
			);
}

inline double nnlo_qbq_nlo_plus0(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps, double A_LO){
	double nf(ps.nf);
	return 	(
			 _C2_qbq_Plus_0(p,ps) + _C2_qbq_Plus_0(p_tu,ps)
			+nf*(_D_aa_Plus_0(p) + _D_aa_Plus_0(p_tu))
			);
}

inline double nnlo_qbq_nlo_plus1(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps, double A_LO){
	return 	(_C2_qbq_Plus_1(p) + _C2_qbq_Plus_1(p_tu));
}

inline double nnlo_qbq_nlo_finit(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps, double A_LO){
	double nf(ps.nf);
	return 	(
			 _C2_qbq_finite(p,ps) + _C2_qbq_finite(p_tu,ps)
			+nf*(_D_aa_finite(p) + _D_aa_finite(p_tu))
			+D_ac(p) + D_ac(p_tu)	//	D_ad (A22)
			);
}
//************	NNLO_ab_Lumi

inline double nnlo_qbq_nnlo_ab_delta(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps, double A_LO){
	return B3_qbq(p);
}

inline double nnlo_qbq_nnlo_ab_finit(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps, double A_LO){
	return D_ab(p) + D_ab(p_tu);
}

//************	NNLO_bb_Lumi

inline double nnlo_qbq_nnlo_bb_x2(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps){
	return D_bb(p) + D_bb(p_tu);
}

//************	NNLO_bc_Lumi

inline double nnlo_qbq_nnlo_bc_x2(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps){
	return D_bc(p);
}

//************	NNLO_bc_Lumi

inline double nnlo_qbq_nnlo_bd_x2(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps){
	return D_bc(p_tu);//D_bd	<-(A23)
}

//************	NNLO_cc_Lumi

inline double nnlo_qbq_nnlo_cc_x2(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps){
	return D_cc(p,ps);
}
//************	NNLO_dd_Lumi

inline double nnlo_qbq_nnlo_dd_x2(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps){
	return D_cc(p_tu,ps);	//D_dd	<-(A25)
}
//************	NNLO_LL_Lumi

inline double nnlo_qbq_nnlo_LL_x2(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps){
	return D_LL_cd(p) + D_LL_cd(p_tu);
}
//************	NNLO_LR_Lumi

inline double nnlo_qbq_nnlo_LR_x2(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps){
	return D_LR_cd(p) + D_LR_cd(p_tu);
}



//****  qq
//************	NNLO_ac_Lumi

inline double nnlo_qq_nnlo_ac_x2(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps){
	return E_ac(p);
}
//************	NNLO_bd_Lumi

inline double nnlo_qq_nnlo_bd_x2(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps){
	return E_ac(p_tu);	//(A35)
}
//************	NNLO_ad_Lumi

inline double nnlo_qq_nnlo_ad_x2(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps){
	return E_ad(p) + E_ad(p_tu);	//E_bc (A34)
}



//****	GG

//************	NNLO_Lumi

inline double nnlo_GG_nnlo_x2(const FO_NNLO_Kinematic_Parameter& p,const FO_NNLO_Kinematic_Parameter& p_tu,const FO_Scale_Parameter& ps){
	return C_G_G(p,ps) + C_G_G(p_tu,ps);
}


}//NS_DY
}//NS_FO
}//CuTe

#endif	// FO_DY_H


