#ifndef FO_LUMIS_H
#define FO_LUMIS_H

//#include "lhafunc.h"
//#include "konst.h"
#include "vecmath.h"
#include "cstdlib"
#include "enums.h"

namespace CuTe{

namespace NS_CHARGES {
	class Charge;
	class CKM_MATRIX;
}  // namespace NS_CHARGES

namespace NS_LHAFUNC {
	class VectorPDF;
}  // namespace NS_LHAFUNC

namespace NS_FO{

using std::cerr;
using std::endl;
using std::ostream;
using std::string;
using std::stringstream;
using std::vector;

using NS_CHARGES::Charge;
using NS_CHARGES::CKM_MATRIX;
using NS_LHAFUNC::VectorPDF;
//using NS_LHAFUNC::Parton_g;

using namespace NS_enums::NS_Partons;
using namespace NS_enums::NS_PROD;



enum ENUM_LUMI_UPDATE{
	 l_update_LO
	,l_update_NLO_0
	,l_update_NLO_x2
};

class Lumi{
public:
	const int Number_of_Lumis;
protected:
	vector<double> _V;
	const int *_nf;

	inline virtual void _init(const int *_nf){
		this->_nf	= _nf;
		_V.assign(Number_of_Lumis,0.);
//		for(int i=0; i<Number_of_Lumis; ++i) _V[i]=0.;
	}

	Lumi(int Number_of_Lumis)
	:Number_of_Lumis(Number_of_Lumis)
	,_V(Number_of_Lumis,0.)
	,_nf(NULL){
//		NS_Utils::DOUT dout("Lumi","NL");

	}

	Lumi(int Number_of_Lumis, const int *_nf)
	:Number_of_Lumis(Number_of_Lumis)
	,_V(Number_of_Lumis,0.)
	,_nf(_nf){
//		NS_Utils::DOUT dout("Lumi","NL,nf");

	}

	Lumi(const Lumi& L)
	:Number_of_Lumis(L.Number_of_Lumis)
	,_V(L._V)
	,_nf(L._nf){
//		NS_Utils::DOUT dout("Lumi","L");

	}

	Lumi& operator=(const Lumi& L){
//		NS_Utils::DOUT dout("Lumi::operator=","L");
		if(Number_of_Lumis!=L.Number_of_Lumis){
			std::cerr<<"Lumi::operator=(L) Wrong Number_of_Lumis="<<Number_of_Lumis<<" != "<<L.Number_of_Lumis	<<std::endl;
			std::exit(1);
		}
		_V	= L._V;
		_nf	= L._nf;

		return *this;
	}

public:

	inline double& operator[](int i){ return _V[i];}
	inline const double& operator[](int i)const{ return _V[i];}

	inline virtual void update(const VectorPDF* const Phi[2],ENUM_LUMI_UPDATE l_update){
		cerr << "****************************************************" << endl
			 << "*** Lumi::update(Phi,l_update) called -> exit(1) ***" << endl
			 << "****************************************************" << endl;
		exit(1);
	}

	inline void update(const VectorPDF *Phi_0, const VectorPDF *Phi_1,ENUM_LUMI_UPDATE l_update){
		cerr << "****************************************************" << endl
			 << "*** Lumi::update(Phi,l_update) called -> exit(1) ***" << endl
			 << "****************************************************" << endl;
		exit(1);
	}

	virtual ~Lumi(){

	}

	inline virtual const string to_string(int ntabs=0)const{
		cerr << "*****************************************" << endl
			 << "*** Lumi::to_string called -> exit(1) ***" << endl
			 << "*****************************************" << endl;
		exit(1);
	}

private:
	Lumi();

};

ostream& operator<<(ostream& os, const Lumi &lumi);

enum ENUM_LUMI_S {
	 LS_GG_LO
	,LS_Gq_LO
	,LS_qG_LO
	,LS_qbq_LO
	,LS_qq_NLO
	,LS_qp_NLO
};


ostream& operator<<(ostream& os, ENUM_LUMI_S l);


class Lumi_S:public Lumi{
public:
	static const int LS_Number_of_Lumis	= 6;

	void update(const VectorPDF* const Phi[2], ENUM_LUMI_UPDATE l_update);

	Lumi_S()
	:Lumi(LS_Number_of_Lumis){
//		NS_Utils::DOUT dout("Lumi_S");
	}

	Lumi_S(const int *nf)
	:Lumi(LS_Number_of_Lumis){
//		NS_Utils::DOUT dout("Lumi_S","nf");
		_init(nf);
	}

	Lumi_S(const Lumi_S &LS)
	:Lumi(LS){
//		NS_Utils::DOUT dout("Lumi_S","LS");
	}

	Lumi_S& operator=(const Lumi_S &LS){
//		NS_Utils::DOUT dout("Lumi_S::operator=","LS");
		Lumi::operator =(LS);
		return *this;
	}

	inline virtual const string to_string(int ntabs=0)const{
		stringstream ss;
		for(int i=0;i<ntabs; ++i) ss<<"\t";
		string tab(ss.str());
		ss.str(string());


		for(int i=0; i<Number_of_Lumis; ++i){
			ss<<tab<<"Lumi_S["<<static_cast<ENUM_LUMI_S>(i)<<"] = "<<_V[i]<<endl;
		}
		return ss.str();
	}

};

enum ENUM_LUMI_V {
	 LV_qG_LO
	,LV_qG_NLO
	,LV_Gq_LO
	,LV_Gq_NLO
	,LV_GG_NLO
	,LV_qbq_LO
	,LV_qbq_NLO_ab
	,LV_qbq_NLO_bb
	,LV_qbq_NLO_bc
	,LV_qbq_NLO_cc
	,LV_qbq_NLO_bd
	,LV_qbq_NLO_dd
	,LV_qbq_NLO_LL
	,LV_qbq_NLO_LR
	,LV_qq_NLO_aa
	,LV_qq_NLO_bb
	,LV_qq_NLO_ac
	,LV_qq_NLO_bd
	,LV_qq_NLO_ad
	,LV_qq_NLO_LL
	,LV_qq_NLO_LR
};

ostream& operator<<(ostream& os, ENUM_LUMI_V l);

class Lumi_V:public Lumi {
public:
	static const int LV_Number_of_Lumis	= 21;
protected:
//	PROD prod;
//	const CKM_MATRIX *ckm;
//	double Theta_W;
	Charge *C;
	int nf;

	Matrix<N_FMAX,double> L,bL,R,bR,D,bD,QQ;
	double Dpp,bDpp,QQqp;
	Vector<N_FMAX,double> QQ_S,QQ_bS;

//	void _init_C();
//	void _exit_C();

	void set_nf();

	void _init(const int *_nf);
public:



	//update the lumi array
//	void update(const VectorPhi* Phi);


	void update(const VectorPDF* const Phi[2], ENUM_LUMI_UPDATE l_update);

	Lumi_V();
	Lumi_V(const int *_nf, const PROD prod, const CKM_MATRIX *ckm, const double Theta_W, const Vector<N_FMAX+1,bool>& active_flavour);
	Lumi_V(const Lumi_V &LV);
	Lumi_V& operator=(const Lumi_V &LV);
	~Lumi_V();

	inline virtual const string to_string(int ntabs=0)const{
		stringstream ss;
		for(int i=0;i<ntabs; ++i) ss<<"\t";
		string tab(ss.str());
		ss.str(string());

		for(int i=0; i<Number_of_Lumis; ++i){
			ss<<tab<<"Lumi_V["<<static_cast<ENUM_LUMI_V>(i)<<"] = "<<_V[i]<<endl;
		}
		return ss.str();
	}

};

}  // namespace NS_FO
}  // namespace CuTe


#endif
