#ifndef HPLz_h
#define HPLz_h

#include <iostream>
#include "hpl1d.hpp"
using std::cerr;
using std::cout;
using std::endl;
using namespace hpl;

namespace CuTe {


class HPLz{
private:
	double x;
	int nw;
	int n1;
	int n2;
	FA1dC11 Hc1;
	FA2dC11 Hc2;
	FA3dC11 Hc3;
	FA4dC11 Hc4;
	FA1dD11 Hr1;
	FA2dD11 Hr2;
	FA3dD11 Hr3;
	FA4dD11 Hr4;
	FA1dD11 Hi1;
	FA2dD11 Hi2;
	FA3dD11 Hi3;
	FA4dD11 Hi4;


public:

	inline void set_weight(int nw){
		if(nw<1 || nw>4) std::cerr	<<"HPLz weight not in [1:4], nw="<<nw	<<endl;
		this->nw = nw;
	}

	inline void update(double z){
		if(x!=z){
			x=z;
			hplog(x, nw, Hc1, Hc2, Hc3, Hc4, Hr1, Hr2, Hr3, Hr4, Hi1, Hi2, Hi3, Hi4, n1, n2);
		}
	}

	inline void update(double z, int nw_){
		if(x!=z || nw<nw_){
			x=z;
			nw=nw_;
			hplog(x, nw, Hc1, Hc2, Hc3, Hc4, Hr1, Hr2, Hr3, Hr4, Hi1, Hi2, Hi3, Hi4, n1, n2);
		}
	}

	inline void update(double z, int nw_, int n1_, int n2_){
		x	= (z);
		nw	= (nw_);
		n1	= (n1_);
		n2	= (n2_);
		hplog(x, nw, Hc1, Hc2, Hc3, Hc4, Hr1, Hr2, Hr3, Hr4, Hi1, Hi2, Hi3, Hi4, n1, n2);
	}

	inline HPLz():x(-1.),nw(3),n1(-1),n2(1){
	}

	inline HPLz(double z):x(z),nw(3),n1(-1),n2(1){
		hplog(x, nw, Hc1, Hc2, Hc3, Hc4, Hr1, Hr2, Hr3, Hr4, Hi1, Hi2, Hi3, Hi4, n1, n2);
	}

	inline double operator() (int n) const {
		return Hr1(n);
	}
	inline double operator() (int n, int m) const {
		if(nw>=2){
			return Hr2(n,m);
		}else{
			cerr	<<"HPL[{"<<n<<","<<m<<"},"<<x<<"] called, but weight = "<<nw<<" < 2"<<endl;
			return 0.;
		}
	}
	inline double operator() (int n, int m, int o) const{
		if(nw>=3){
			return Hr3(n,m,o);
		}else{
			cerr	<<"HPL[{"<<n<<","<<m<<","<<o<<"},"<<x<<"] called, but weight = "<<nw<<" < 3"<<endl;
			return 0.;
		}
	}
	inline double operator() (int n, int m, int o, int p) const{
		if(nw>=4){
			return Hr4(n,m,o,p);
		}else{
			cerr	<<"HPL[{"<<n<<","<<m<<","<<o<<","<<4<<"},"<<x<<"] called, but weight = "<<nw<<" < 4"<<endl;
			return 0.;
		}
	}

	inline double getz() const {return x;}
	inline int getw() const {return nw;}

};


//std::ostream& operator<<( std::ostream& os, const HPLz &HPL )
//{
//	os << "HPLz(" << HPL.getz() << "," << HPL.getw() << ")" << endl;
//
//	for(int m=-1;m<=1; m++){
//		os << "\tHPL{"<<m<<"}="<<HPL(m);
//	}
//	os << endl;
//
//	for(int m=-1;m<=1; m++){
//		for(int n=-1;n<=1; n++){
//			os << "\tHPL{"<<m<<","<<n<<"}="<<HPL(m,n);
//		}
//		os << endl;
//	}
//	os << endl;
//
//	for(int m=-1;m<=1; m++){
//		for(int n=-1;n<=1; n++){
//			for(int o=-1;o<=1; o++){
//				os << "\tHPL{"<<m<<","<<n<<","<<o<<"}="<<HPL(m,n,o);
//			}
//			os << endl;
//		}
//		os << endl;
//	}
//	os << endl;
//
//	return os;
//}


}  // namespace CuTe


#endif
