#ifndef IBAR_H
#define IBAR_H

#include "Ibar_G_Functions.h"
#include "Ibar_Q_Functions.h"
#include "lhafunc.h"

namespace CuTe {

namespace NS_Ibar {

using namespace NS_APSF;
using NS_LHAFUNC::PDF;
using NS_LHAFUNC::VectorPDF;

using namespace NS_enums::NS_Partons;

using NS_Polynom::Pol;

enum ENUM_IBAR_I { E_Ibar_Q, E_Ibar_G, E_Ibar_G2 };

inline std::ostream& operator<<(std::ostream& os, ENUM_IBAR_I i){
	switch (i) {
	case E_Ibar_Q:
		os	<<"Q_";
		break;
	case E_Ibar_G:
		os	<<"G_";

		break;
	case E_Ibar_G2:
		os	<<"G2";

		break;
	default:
		os	<<"WRONG";
		break;
	}
	return os;
}


class Ibar_ij_conv_f:public Pol_a_l{
public:

	inline void reset_to_zero(){
		order(0);
		_P[0].order(0);
		_P[0][0]	= 0.;
	}

	inline void re_init(int ord_a_or_e){
		order(ord_a_or_e);
		for(int a=0; a<=ord_a_or_e; ++a) _P[a].order(a);
	}

	template< ENUM_IBAR_I i, e_P j >
	Pol_a_l& set(double fz, double f1, const Ibar_Parameter *IP, int nf){
		re_init(IP->order);
			 if(i==E_Ibar_Q)	 Ibar_Q_pol_set<j>(*this,fz,f1,*static_cast<const Ibar_Q_Parameter*>(IP),nf);
		else if(i==E_Ibar_G)	 Ibar_G_pol_set<j>(*this,fz,f1,*static_cast<const Ibar_G_Parameter*>(IP),nf);
		else if(i==E_Ibar_G2)	Ibar_G2_pol_set<j>(*this,fz,f1,*static_cast<const Ibar_G_Parameter*>(IP),nf);
		return *this;
	}

	template< ENUM_IBAR_I i, e_P j >
	Pol_a_l& add(double fz, double f1, const Ibar_Parameter *IP, int nf){
		 if(i==E_Ibar_Q)	 Ibar_Q_pol_add<j>(*this,fz,f1,*static_cast<const Ibar_Q_Parameter*>(IP),nf);
	else if(i==E_Ibar_G)	 Ibar_G_pol_add<j>(*this,fz,f1,*static_cast<const Ibar_G_Parameter*>(IP),nf);
	else if(i==E_Ibar_G2)	Ibar_G2_pol_add<j>(*this,fz,f1,*static_cast<const Ibar_G_Parameter*>(IP),nf);
		return *this;
	}

	inline Ibar_ij_conv_f()
	:Pol_a_l(5,Pol_l(5,0.)){

	}
};

class VectorbarI: public Vector< N_PARTONS, Ibar_ij_conv_f >{
protected:

	VectorPDF phixi,phiz;

	const double*mu;
	const int	*nf;

	double   g_xi;
	double soq_xi;

	double   cg_z;
	double csoq_z;


	Vector<N_FMAX+1,bool>::const_iterator active_flavour_charges_g;


	inline virtual void v_init(){
		cout	<<"******\t*******************\t******" << endl
				<<"******\tVectorbarI::v_init()\t******" << endl
				<<"******\t*******************\t******" << endl;
		exit(1);
	}

	inline virtual void v_set_xi(const Ibar_Parameter *Ip){
		cout	<<"******\t*******************\t******" << endl
				<<"******\tVectorbarI::v_set_xi()\t******" << endl
				<<"******\t*******************\t******" << endl;
		exit(1);
	}

	inline virtual void v_set_z(const Ibar_Parameter *Ip){
		cout	<<"******\t*******************\t******" << endl
				<<"******\tVectorbarI::v_set_z()\t******" << endl
				<<"******\t*******************\t******" << endl;
		exit(1);
	}

	inline virtual void v_update_z(const Ibar_Parameter *Ip){
		cout	<<"******\t*******************\t******" << endl
				<<"******\tVectorbarI::v_update()\t******" << endl
				<<"******\t*******************\t******" << endl;
		exit(1);
	}


	inline void _set_z(const Ibar_Parameter *Ip){
		const double c(Ip->iz);
		const double x(c*(Ip->x));	//= xi/z

		phiz.update(x,*mu);
		cg_z	= c*phiz[Parton_g];
		csoq_z	= c*phiz.soq();

		v_set_z(Ip);
	}

public:


	inline void update_z(const Ibar_Parameter *Ip){
//		NS_Utils::DOUT dout("VectorbarI::update_z");
//		dout()	<<"mu="<<*mu	<<endl;

		_set_z(Ip);

		v_update_z(Ip);
	}


	inline void set_xi(const Ibar_Parameter *Ip){
//		NS_Utils::DOUT dout("VectorbarI::set_xi");
//		dout()	<<"mu="<<*mu<<endl;
//		cout	<<Ip->to_String(dout.get_ntab())	<<endl;

		phixi.update(Ip->x,*mu);
		soq_xi		= phixi.soq();
		g_xi		= phixi[Parton_g];


		v_set_xi(Ip);
	}

	inline void init(const VectorPDF &vpdf, const double *mu, const int *nf, const Vector<N_FMAX+1,bool>::const_iterator& active_flavour_charges_g){
//		NS_Utils::DOUT dout("VectorbarI::init","VectorPDF,mu,nf");
		change_PDF(vpdf);

//		cout<<phixi.properties(dout.get_ntab()+1);
//		cout<<phiz.properties(dout.get_ntab()+1);

		this->mu		= (mu);
		this->nf		= (nf);

		this->active_flavour_charges_g	= active_flavour_charges_g;

		v_init();
	}

	inline void init(const PDF &pdf, const double *mu, const int *nf, const Vector<N_FMAX+1,bool>::const_iterator& active_flavour_charges_g){
//		NS_Utils::DOUT dout("VectorbarI::init","PDF,mu,nf");
		init(VectorPDF(pdf),mu,nf,active_flavour_charges_g);
	}


	inline VectorbarI()
	:Vector< N_PARTONS, Ibar_ij_conv_f >()
	,phixi(),phiz()
	,mu(NULL),nf(NULL)
	,g_xi(),soq_xi()
	,cg_z(),csoq_z()
	,active_flavour_charges_g(){
	}


	inline VectorbarI(const VectorPDF &vpdf, const double *mu, const int *nf, const Vector<N_FMAX+1,bool>::const_iterator& active_flavour_charges_g)
	:Vector< N_PARTONS, Ibar_ij_conv_f >()
	,phixi(vpdf),phiz(vpdf)
	,mu(mu),nf(nf)
	,g_xi(),soq_xi()
	,cg_z(),csoq_z()
	,active_flavour_charges_g(active_flavour_charges_g){
	}

	inline VectorbarI(const PDF *pdf, const double *mu, const int *nf, const Vector<N_FMAX+1,bool>::const_iterator& active_flavour_charges_g)
	:Vector< N_PARTONS, Ibar_ij_conv_f >()
	,phixi(pdf),phiz(pdf)
	,mu(mu),nf(nf)
	,g_xi(),soq_xi()
	,cg_z(),csoq_z()
	,active_flavour_charges_g(active_flavour_charges_g){
	}


	inline virtual ~VectorbarI(){
	}

	inline virtual const string to_String(int ntabs=0)const{
		stringstream ss;
		for(int i=0; i<ntabs; ++i) ss<<"\t";
		string tab(ss.str());

		ss.str(string());

		ss	<<tab<<"VectorbarI"<<endl
			<<tab<<"\t"<<"mu="<<*mu<<"\t"<<"nf="<<*nf<<endl
			<<tab<<"\t"<<"g_xi="<<g_xi<<"\t"<<"soq_xi="<<soq_xi<<endl
			<<tab<<"\t"<<"cg_z="<<cg_z<<"\t"<<"csoq_z="<<csoq_z<<endl;

		Vector<N_FMAX+1,bool>::const_iterator itr(active_flavour_charges_g);

		ss	<<tab;
		for(Parton p(1); p<=Parton_t; ++p){
			ss	<<"\t"<<p<<" = "<<*(++itr);
		}
		ss	<<endl;

		return ss.str();
	}


	inline void print_pdf(const string s="")const{
		phixi.print_pdf(s+"xi");
		phiz.print_pdf(s+"z");
	}

	inline virtual double get_B_LO_by_index(int i)const{
		cout	<<"******\t*******************\t******" << endl
				<<"******\tVectorbarI::get_LO(i)\t******" << endl
				<<"******\t*******************\t******" << endl;
		exit(1);
	}

	inline void change_PDF(const VectorPDF& vpdf){
		phixi	= vpdf;
		phiz	= vpdf;
	}

	inline void change_PDF(const PDF& pdf){
		change_PDF(VectorPDF(pdf));
	}

	inline const Ibar_ij_conv_f& operator()(const Parton& p)const{
		return _V[p.i];
	}
};

class VectorbarI_H: public VectorbarI{
protected:

	inline virtual void v_init(){
		for(int i=2; i<N_PARTONS; ++i) _V[i].reset_to_zero();
	}

	inline virtual void v_set_xi(const Ibar_Parameter *Ip){
//		NS_Utils::DOUT dout("VectorbarI_H::v_set_xi","Ip");
	}

	inline virtual void v_set_z(const Ibar_Parameter *Ip){
	}

	inline void v_update_z(const Ibar_Parameter *IG){
		_V[0].set<E_Ibar_G ,e_g>(  cg_z,  g_xi,IG,*nf);
		_V[0].add<E_Ibar_G ,e_q>(csoq_z,soq_xi,IG,*nf);

		_V[1].set<E_Ibar_G2,e_g>(  cg_z,  g_xi,IG,*nf);
		_V[1].add<E_Ibar_G2,e_q>(csoq_z,soq_xi,IG,*nf);

		_V[1].reduce_order();
	}

public:
	inline VectorbarI_H()
	:VectorbarI(){
	}

	inline VectorbarI_H(const VectorPDF &vpdf, const double *mu, const int *nf, const Vector<N_FMAX+1,bool>::const_iterator& active_flavour_charges_g)
	:VectorbarI(vpdf,mu,nf,active_flavour_charges_g){
		v_init();
	}

	inline VectorbarI_H(const PDF &pdf, const double *mu, const int *nf, const Vector<N_FMAX+1,bool>::const_iterator& active_flavour_charges_g)
	:VectorbarI(pdf,mu,nf,active_flavour_charges_g){
		v_init();
	}

	inline ~VectorbarI_H(){
	}

	inline virtual const string to_String(int ntabs=0)const{
		stringstream ss;
		for(int i=0; i<ntabs; ++i) ss<<"\t";
		string tab(ss.str());

		ss.str(string());

		ss	<<tab<<"VectorbarI_H"<<endl
			<<VectorbarI::to_String(ntabs+1)
			<<tab<<"\t"<<"B_G \t="<<_V[0]<<endl
			<<tab<<"\t"<<"B_G2\t="<<_V[1]<<endl;

		return ss.str();
	}

	inline virtual double get_B_LO_by_index(int i)const{
		return g_xi;
	}
};


class VectorbarI_DY: public VectorbarI{
protected:

	Parton  q;
	Parton bq;
//	int  iq;
//	int ibq;
//	double  q_xi;
//	double bq_xi;
//	double  cq_z;
//	double cbq_z;

	double  soqp_xi;
	double soqpb_xi;
	double  csoqp_z;
	double csoqpb_z;

	Ibar_ij_conv_f Ibar_q_g_conv_Phi,Ibar_q_qp_conv_Phi,Ibar_q_qbp_conv_Phi;

	inline virtual void v_init(){
		_V[Parton_g.i].reset_to_zero();
	}

	inline virtual void v_set_xi(const Ibar_Parameter *Ip){
//		NS_Utils::DOUT dout("VectorbarI_DY::v_set_xi","Ip");
		soqp_xi		= phixi.soqp(*nf);
		soqpb_xi	= phixi.sobp(*nf);

		Vector<N_FMAX+1,bool>::const_iterator itr(active_flavour_charges_g);

		for(q=Parton(1); q<=Parton(*nf); ++q){
			bq	= q.bar();
			if(!(*(++itr))){
				_V[ q.i].reset_to_zero();
				_V[bq.i].reset_to_zero();
			}
		}

		for(; q<=NS_LHAFUNC::Parton_t; ++q){
			bq	= q.bar();
			_V[ q.i].reset_to_zero();
			_V[bq.i].reset_to_zero();
		}

	}

	inline virtual void v_set_z(const Ibar_Parameter *Ip){
		const double c(Ip->iz);

		csoqp_z		= c*phiz.soqp(*nf);
		csoqpb_z	= c*phiz.sobp(*nf);
	}

	inline void v_update_z(const Ibar_Parameter *_IQ){
		Ibar_q_g_conv_Phi.set<E_Ibar_Q,e_g>(cg_z,g_xi,_IQ,*nf);
		Ibar_q_g_conv_Phi.reduce_order();


		const double c(_IQ->iz);
		const int ord(_IQ->order);

		 Ibar_q_qp_conv_Phi.set<E_Ibar_Q, e_qp>( csoqp_z, soqp_xi,_IQ,*nf);
		 Ibar_q_qp_conv_Phi.add<E_Ibar_Q,e_qbp>(csoqpb_z,soqpb_xi,_IQ,*nf);
		 Ibar_q_qp_conv_Phi.reduce_order();

		Ibar_q_qbp_conv_Phi.set<E_Ibar_Q, e_qp>(csoqpb_z,soqpb_xi,_IQ,*nf);
		Ibar_q_qbp_conv_Phi.add<E_Ibar_Q,e_qbp>( csoqp_z, soqp_xi,_IQ,*nf);
		Ibar_q_qbp_conv_Phi.reduce_order();

		Vector<N_FMAX+1,bool>::const_iterator itr(active_flavour_charges_g);
		for(q=Parton(1); q<=Parton(*nf); ++q){
			if(*(++itr)){
//		for(q=Parton(*nf); q> Parton_g; --q){
				bq 		= (-q);
				const double  q_xi	=  phixi[ q];
				const double bq_xi	=  phixi[bq];
				const double  cq_z	= c*phiz[ q];
				const double cbq_z	= c*phiz[bq];
				const int	  iq	=  q.i;
				const int	  ibq	= bq.i;

				_V[ iq].set<E_Ibar_Q,e_qmp>( cq_z, q_xi,_IQ,*nf);
				_V[ibq].set<E_Ibar_Q,e_qmp>(cbq_z,bq_xi,_IQ,*nf);

				_V[ iq].add_this_Pol(Ibar_q_g_conv_Phi,ord);
				_V[ibq].add_this_Pol(Ibar_q_g_conv_Phi,ord);

				_V[ iq].add<E_Ibar_Q,e_qbmp>(cbq_z,bq_xi,_IQ,*nf);
				_V[ibq].add<E_Ibar_Q,e_qbmp>( cq_z, q_xi,_IQ,*nf);

				_V[ iq].add_this_Pol(Ibar_q_qp_conv_Phi ,ord);
				_V[ibq].add_this_Pol(Ibar_q_qbp_conv_Phi,ord);
			}
		}
	}

public:
	inline VectorbarI_DY()
	:VectorbarI()
	,q(),bq()
//	,iq(),ibq()
//	,q_xi(),bq_xi()
//	,cq_z(),cbq_z()
	,soqp_xi(),soqpb_xi()
	,csoqp_z(),csoqpb_z(){
	}

	inline VectorbarI_DY(const VectorPDF &vpdf, const double *mu, const int *nf, const Vector<N_FMAX+1,bool>::const_iterator& active_flavour_charges_g)
	:VectorbarI(vpdf,mu,nf,active_flavour_charges_g)
	,q(),bq()
//	,iq(),ibq()
//	,q_xi(),bq_xi()
//	,cq_z(),cbq_z()
	,soqp_xi(),soqpb_xi()
	,csoqp_z(),csoqpb_z(){
		v_init();
	}

	inline VectorbarI_DY(const PDF &pdf, const double *mu, const int *nf, const Vector<N_FMAX+1,bool>::const_iterator& active_flavour_charges_g)
	:VectorbarI(pdf,mu,nf,active_flavour_charges_g)
	,q(),bq()
//	,iq(),ibq()
//	,q_xi(),bq_xi()
//	,cq_z(),cbq_z()
	,soqp_xi(),soqpb_xi()
	,csoqp_z(),csoqpb_z(){
		v_init();
	}

	inline ~VectorbarI_DY(){
	}

	inline virtual const string to_String(int ntabs=0)const{
		stringstream ss;
		for(int i=0; i<ntabs; ++i) ss<<"\t";
		string tab(ss.str());

		ss.str(string());

		ss	<<tab<<"VectorbarI_DY"<<endl
			<<VectorbarI::to_String(ntabs+1);

		for(Parton q=Parton(-(*nf)); q< Parton_g; ++q){
			ss<<tab<<"\t"<<"B_["<<q<<"]\t="<<_V[q.i]<<endl;
		}

		for(Parton q=Parton(1); q<=Parton((*nf)); ++q){
			ss<<tab<<"\t"<<"B_["<<q<<"]\t="<<_V[q.i]<<endl;
		}

		return ss.str();
	}

	inline virtual double get_B_LO_by_index(int i)const{
		return phixi.get_element_(i);
	}
};


}  // namespace Ibar

}  // namespace CuTe

#endif	//IBAR_H
