#ifndef IH_PARAMS_H
#define IH_PARAMS_H

//#include "HPLz.h"
//#include "PlusDists.h"
//#include "utils.h"
#include "APSF_Parameter.h"


namespace CuTe {
namespace NS_Ibar{

using NS_APSF::APSF_Parameter;
using NS_APSF::e_P;
using std::string;
using std::stringstream;

using namespace NS_Utils::NS_min_max;
using namespace NS_Utils::NS_pow;


class Ibar_Parameter: public APSF_Parameter {
protected:

	virtual void _func_update(){
		std::cerr	<<"********************************************************"<<endl
					<<"*** Ibar_Parameter::_func_update() called -> exit(1) ***"<<endl
					<<"********************************************************"<<endl;
		exit(1);
	}

public:
	int order;
	bool improved;
	int order_a;

	double LNz,LNz2,LNz3,LNz4,LNz5;

	inline void init(int order, bool improved, int order_a=0){
//		NS_Utils::DOUT dout("Ibar_Parameter::init");
//		dout()<<"in:"<<"\to="<<order<<"\ti="<<improved<<"\to_a="<<order_a<<endl;
		int max_weight;

		if(improved && (order_a>=order)){ improved = false;}

		if(!improved){
			order	= max_of(order,order_a);
			order_a	= order;
			max_weight	= max_of(1 ,2*order_a-1);
		}else{
			order_a	= max_of(order_a,order/2);

			max_weight	= max_of(max_of(1 ,2*order_a-1),order-1);
		}
		this->improved	= improved;
		this->order		= order;
		this->order_a	= order_a;
//		dout()<<"set:"<<"\to="<<order<<"\ti="<<improved<<"\to_a="<<order_a<<endl;

		hpl.set_weight(max_weight);
	}

	inline virtual void update(double z){
//		NS_Utils::DOUT	dout("Ibar_Parameter::update","z");
		APSF_Parameter::update(z);
		hpl.update(z);

		LNz		= hpl(0);
		LNz2	= LNz*LNz;
		LNz3	= LNz*LNz2;
		LNz4	= LNz2*LNz2;
		LNz5	= LNz2*LNz3;
		_func_update();
	}

	inline Ibar_Parameter()
	:APSF_Parameter()
	,order(0),improved(false),order_a(0)
	,LNz(),LNz2(),LNz3(),LNz4(),LNz5(){

	}

	inline virtual ~Ibar_Parameter(){

	}

	inline virtual const string to_String(int ntabs=0)const{
		stringstream ss;

		string tabs("");
		for(int i=1;i<=ntabs; ++i) tabs += string("\t");

		ss	<<tabs<<"Ibar_Parameter:"<<endl;
		ss	<<tabs<<"\t"<<"ord="<<order<<"\t"<<"imp="<<improved<<"\t"<<"ord_a="<<order_a<<endl;
		ss	<<APSF_Parameter::to_String(++ntabs);
		ss	<<tabs<<"\t"<<"LNz="<<LNz<<"\t"<<"LNz2="<<LNz2<<"\t"<<"LNz3="<<LNz3<<"\t"<<"LNz4="<<LNz4<<"\t"<<"LNz5="<<LNz5<<endl;

		return ss.str();
	}

};

}	//NS_Ibar

}	//CuTe

#endif

