#ifndef KINEMATICS_OPTIONS_H_
#define KINEMATICS_OPTIONS_H_

#include "options.h"

namespace CuTe {
namespace NS_Kinematics {

using NS_OPTIONS::Option;
using NS_OPTIONS::VOption;
using std::string;

class OPS_Cuts: public NS_OPTIONS::_Options_Set{
public:
	Option<double> 		abs_y_max;
	VOption<double,2>	abs_y_exclude;
	Option<double> 		kT_min;
	Option<double> 		abs_eta_max;
	VOption<double,2>	abs_eta_exclude;


protected:

	void v_init_ops(){
//		std::cout<<"\t"<<"OPS_Cuts::v_init_ops()"<<std::endl;
		abs_y_max		= Option<double>("abs_y_max","[double]");
		abs_y_max.set_def_exp("No Cut");

		abs_y_exclude	= VOption<double,2>("abs_y_exclude","[Min Max]");
		abs_y_exclude.set_def_exp("No Cut");

		kT_min			= Option<double>("kT_min","[double]");
		kT_min.set_def_exp("No Cut");

		abs_eta_max		= Option<double>("abs_eta_max","[double]");
		abs_eta_max.set_def_exp("No Cut");

		abs_eta_exclude	= VOption<double,2>("abs_eta_exclude","[Min Max]");
		abs_eta_exclude.set_def_exp("No Cut");
	}

	void v_init_vec(){
//		std::cout<<"\t"<<"OPS_Cuts::v_init_vec()"<<std::endl;
		options.push_back(&abs_y_max);
		options.push_back(&abs_y_exclude);
		options.push_back(&kT_min);
		options.push_back(&abs_eta_max);
		options.push_back(&abs_eta_exclude);
	}

public:

	OPS_Cuts():_Options_Set("Cut Options"){
//		std::cout<<"\t"<<"OPS_Cuts()"<<std::endl;
		init();
	}

	OPS_Cuts(const string& filename):_Options_Set("Cut Options"){
//		std::cout<<"\t"<<"OPS_Cuts("<<filename<<")"<<std::endl;
		init(filename);
	}

	OPS_Cuts(const OPS_Cuts& OS)
	:_Options_Set(OS)
	,abs_y_max(OS.abs_y_max)
	,abs_y_exclude(OS.abs_y_exclude)
	,kT_min(OS.kT_min)
	,abs_eta_max(OS.abs_eta_max)
	,abs_eta_exclude(OS.abs_eta_exclude){
		init_vec();
	}

};

}  // namespace NS_Kinematics
}  // namespace CuTe
#endif //KINEMATICS_OPTIONS_H_
