#include <Kinematics_Two_P_PS.h>

using namespace CuTe::NS_Kinematics;

const double Two_P_PS::acc_abs = 1e-5;
const double Two_P_PS::acc_rel = 0.;

Two_P_PS::Two_P_PS()
:gsl_int(intMem)
,M(),qT(),y()
,lM2(),lT(),ly()
,L_nu()//,Jac_nu()
,nu(),Snu()
,L_u()
,cuts()
,a(),aoE()
,b(),thb()
,ex()
,thb0(),thb1()
,E_nu_p(),E_nu_m()
,E_u_m(),E_u_p()
,integration_bins(12,0.){
	gsl_int.function	= &_gsl_integrand;
	gsl_int.params		= this;
}

Two_P_PS::Two_P_PS(const Two_P_PS &RV)
:gsl_int(RV.gsl_int)
,M(RV.M),qT(RV.qT),y(RV.y)
,lM2(RV.lM2),lT(RV.lT),ly(RV.ly)
,L_nu(RV.L_nu)//,Jac_nu(RV.Jac_nu)
,nu(RV.nu),Snu(RV.Snu)
,L_u(RV.L_u)
,cuts(RV.cuts)
,a(RV.a),aoE(RV.aoE)
,b(RV.b),thb(RV.thb)
,ex(RV.ex)
,thb0(RV.thb0),thb1(RV.thb1)
,E_nu_p(RV.E_nu_p),E_nu_m(RV.E_nu_m)
,E_u_m(RV.E_u_m),E_u_p(RV.E_u_p)
,integration_bins(12,0.){
	gsl_int.function	= &_gsl_integrand;
	gsl_int.params		= this;
}

Two_P_PS& Two_P_PS::operator=(const Two_P_PS &RV){
	gsl_int.result	= (RV.gsl_int.result);
	gsl_int.error	= (RV.gsl_int.error);

	M		= (RV.M);
	qT		= (RV.qT);
	y		= (RV.y);
	lM2		= (RV.lM2);
	lT		= (RV.lT);
	ly		= (RV.ly);
	L_nu	= (RV.L_nu);
	//Jac_nu	= (RV.Jac_nu);
	nu		= (RV.nu);
	Snu		= (RV.Snu);
	L_u		= (RV.L_u);
	cuts	= (RV.cuts);
	a		= (RV.a);
	aoE		= (RV.aoE);
	b		= (RV.b);
	thb		= (RV.thb);
	ex		= (RV.ex);
	thb0	= (RV.thb0);
	thb1	= (RV.thb1);
	E_nu_p	= (RV.E_nu_p);
	E_nu_m	= (RV.E_nu_m);
	E_u_m	= (RV.E_u_m);
	E_u_p	= (RV.E_u_p);
	return *this;
}

double Two_P_PS::init_and_return_min_mp(const double M){
	this->M	= M;

	cuts= false;
	a	= -1.;
	b	= -1.;
	thb	= +1.;
	gsl_int.result	= 1.;
	gsl_int.error	= 0.;

	ex	= false;

	return M;
}

double Two_P_PS::init_and_return_min_mp(const double M, const double a, const double b){
	double mp_min(init_and_return_min_mp(M));

	if(0.<a){
		cuts	= true;
		this->a	= a;
		mp_min = max_of(2.*a,mp_min);
	}

	if(0.<b){
		cuts	= true;
		this->b	= b;
		thb		= tanh(b);
	}

	return mp_min;
}

double Two_P_PS::init_and_return_min_mp(const double M, const double a, const double b, const double b0, const double b1){
	const double min_mp(init_and_return_min_mp(M,a,b));

	if( (0.<=b0) && (b0<b1) ){
		ex		= true;
		cuts	= true;

		thb0	= tanh(b0);
		thb1	= tanh(b1);
	}

	return min_mp;
}

double Two_P_PS::init_and_return_min_mp(const double M, const OPS_Cuts &op){
	double a(-1.);
	double b(-1.);
	if(op.kT_min.is_set)		a=op.kT_min();
	if(op.abs_eta_max.is_set)	b=op.abs_eta_max();

	if(op.abs_eta_exclude.is_set){
		return init_and_return_min_mp(M,a,b,op.abs_eta_exclude[0],op.abs_eta_exclude[1]);
	}else{
		return init_and_return_min_mp(M,a,b);
	}
}

double Two_P_PS::update_qT_y(const double qT, const double y){
	this->qT	= qT;
	this->y		= y;
	if(cuts){
//		NS_Utils::DOUT dout("Two_P_PS::update_qT_y","with cuts");
		const double iE(1./(sqrt(pow_2(M)+pow_2(qT))*cosh(y)));

//		dout()	<<"qT = "<<qT	<<"\t"<<"y = "<<y	<<"\t"<<"E = "<<1./iE	<<endl;

		integration_bins.clear();

		lM2	= pow_2(M*iE);
		lT	= qT*iE;
		ly	= tanh(y);

//		dout()	<<"lM2 = "<<lM2	<<"\t"<<"lT = "<<lT	<<"\t"<<"ly = "<<ly	<<endl;

		L_nu.init_min_max(-thb,+thb);
		integration_bins.push_back(-thb);
		integration_bins.push_back(+thb);

		if(b>0.){

//			dout()	<<"b = "<<b	<<"\t"<<"thb = "<<thb	<<endl;
			if(L_nu.contains_not(ly)){
//					NS_Utils::DOUT dout("Two_P_PS::update_qT_y","qT,y");
//					dout()	<<"qT = "<<qT	<<"\t"<<"y = "<<y<<endl;
//					dout()	<<"b = "<<b	<<"\t"<<"thb = "<<thb	<<"\t"<<"ly = "<<ly<<endl;
				return _set0();
			}

			const double nub_mm(_fnub_upm(-thb,-1.));
			const double nub_mp(_fnub_upm(-thb,+1.));
			const double nub_pp(_fnub_upm(+thb,+1.));
			const double nub_pm(_fnub_upm(+thb,-1.));

//			dout()	<<"nu_b:"	<<"\t"<<"(--)"<<nub_mm	<<"\t"<<"(-+)"<<nub_mp	<<"\t"<<"(++)"<<nub_pp	<<"\t"<<"(+-)"<<nub_pm	<<endl;

//			L_nu.add_limit_min_max(nub_mm,nub_pm);	// is only a limit at small qT

			integration_bins.push_back(nub_mm);
			integration_bins.push_back(nub_mp);
			integration_bins.push_back(nub_pp);
			integration_bins.push_back(nub_pm);
		}

		if(a>0.){
			aoE	= a*iE;

			const double lp2(1.-pow_2(ly));

			const double num1(pow_2(2.*aoE)*ly);
			const double num2(sqrt(pow_3(lp2)*(lp2-pow_2(2.*aoE))));
			const double den(1./(pow_2(lp2)+num1*ly));

			const double nua_m((num1-num2)*den);
			const double nua_p((num1+num2)*den);

//			dout()	<<"a = "<<a	<<"\t"<<"aoE = "<<aoE	<<"\t"<<"L_nu_a = ["<<nua_m<<","<<nua_p<<"]"	<<endl;

			L_nu.add_limit_min_max(nua_m,nua_p);

			integration_bins.push_back(nua_m);
			integration_bins.push_back(nua_p);

			if(L_nu.is_unvalid()) return _set0();
		}

		if(ex){
			if(thb0==0.){
				E_nu_p.init_min_max(-thb1,+thb1);
				E_nu_m.set_unvalid();
			}else{
				E_nu_p.init_min_max(+thb0,+thb1);
				E_nu_m.init_min_max(-thb1,-thb0);
				integration_bins.push_back(-thb0);
				integration_bins.push_back(+thb0);
			}
			integration_bins.push_back(-thb1);
			integration_bins.push_back(+thb1);

			E_nu_p.apply_to_limit(L_nu);
			E_nu_m.apply_to_limit(L_nu);

			if(L_nu.is_unvalid()) return _set0();

		}

//		NS_Utils::Timer timer;
//		using NS_Utils::operator<<;

		std::sort(integration_bins.begin(),integration_bins.end());
		vector<double>::const_iterator itr(integration_bins.begin());
		const vector<double>::const_iterator end(integration_bins.end()-1);

//		dout()	<<"bins: "<<integration_bins<<endl;
		Variable_Limit IR;
		double result(0.),error(0.);

//		timer.start();
		while(itr!=end){
			IR.init_min(*itr);
			IR.init_max(*(++itr));
//			dout()	<<	"IR = "<<IR;
			IR.add_limit(L_nu);
//			dout(0)	<<	" -> "<<IR;
			if(IR.is_valid()){
//				dout(1)<<"valid"<<endl;
				if(ex)	result += _calc_one_region(IR,E_nu_m,E_nu_p);
				else	result += _calc_one_region(IR);
			}
//			else{
//				dout(1)<<"!NOT! valid ["<<*(itr-1)<<","<<*itr<<"] ~ "<<gsl_int.integrate_a_to_b(*(itr-1),*itr,acc_rel,acc_abs) <<endl;
//			}
		}
//		timer.stop();
//		cout	<<"result = "<<result<<"\ttime = "<<timer<<endl;



		gsl_int.result = result;
//			char name[96];
//			sprintf(name,"PS(%.0f,%.2f,%.2f,a=%.0f,b=%.2f,b01=[%.2f,%.2f])",M,qT,y,a,b,atanh(thb0),atanh(thb1));
//			gsl_int.status(name);
	}
	return gsl_int.result;
}

ostream& Two_P_PS::into_stream(ostream& os)const{
	os<<"PS("<<M<<","<<qT<<","<<y;
	if(cuts){
		if(0<a)	os	<<",a->"<<a;
		if(0<b)	os	<<",b->"<<b;
		if(ex)	os	<<",b01->["<<atanh(thb0)<<","<<atanh(thb1)<<"]";
	}
	os<<") = ("<<gsl_int.result<<" +- "<<gsl_int.error<<")";
	return os;
}

ostream& CuTe::NS_Kinematics::operator<<(std::ostream& os,const Two_P_PS& ps){
	return ps.into_stream(os);
}


const size_t Two_P_PS_Interpolated::_N_default = 1000;
const GSL_Interpolation_Types Two_P_PS_Interpolated::_Type_default = GSL_Interpolation_Type_Linear;


double Two_P_PS_Interpolated::init_and_return_min_mp(const double M, const size_t N, const GSL_Interpolation_Types Type){
	gsl_interpolator.init(N,Type);
	return Two_P_PS::init_and_return_min_mp(M);
}
double Two_P_PS_Interpolated::init_and_return_min_mp(const double M, double a, const double b, const size_t N, const GSL_Interpolation_Types Type){
	gsl_interpolator.init(N,Type);
	return Two_P_PS::init_and_return_min_mp(M,a,b);
}
double Two_P_PS_Interpolated::init_and_return_min_mp(const double M, const double a, const double b, const double ex_b0, const double ex_b1, const size_t N,  const GSL_Interpolation_Types Type){
	gsl_interpolator.init(N,Type);
	return Two_P_PS::init_and_return_min_mp(M,a,b,ex_b0,ex_b1);
}
double Two_P_PS_Interpolated::init_and_return_min_mp(const double M, const OPS_Cuts &op_c, const size_t N, const GSL_Interpolation_Types Type){
	gsl_interpolator.init(N,Type);
	return Two_P_PS::init_and_return_min_mp(M,op_c);
}

const Variable_Limit& Two_P_PS_Interpolated::update_qT_get_ipol_range(const double qT, const Variable_Limit &L_y){
//	NS_Utils::DOUT dout("Two_P_PS_Interpolated::set_qT","qT,y_min,y_max");
//	dout()	<<"qT = "<<qT	<<"\t"<<"Ly = "<<L_y	<<endl;

	y_interpolation_range = L_y;

	if(cutted_PS()){
		if(b>0.){
			y_interpolation_range.add_limit_min_max(-b,b);
		}

		const int N = gsl_interpolator.itr_init_return_N();
//		const double delta( y_interpolation_range.size()/static_cast<double>(N-2.) );
//
//		gsl_interpolator.itr_step_return_continue(y_interpolation_range.min,0.);
//
//		double y(y_interpolation_range.min+0.5*delta);
//		for(int i=1; i<(N-1); ++i){
//			gsl_interpolator.itr_step_return_continue(y,Two_P_PS::update_qT_y(qT,y));
//			y += delta;
//		}
//
//		gsl_interpolator.itr_step_return_continue(y_interpolation_range.max,0.);

		const double y_max( max_of(fabs(y_interpolation_range.min),y_interpolation_range.max) );
		const double y_delta( y_max/static_cast<double>(N-1.) );

		double y(0.);

		for(int i=0; i<(N-1); ++i){
			gsl_interpolator.itr_step_return_continue(y,Two_P_PS::update_qT_y(qT,y));
			y += y_delta;
		}
		gsl_interpolator.itr_step_return_continue(y_max,0.);

		gsl_interpolator.itr_interpolate_ok();
	}
	return y_interpolation_range;
//		dout()	<<"Max(y) - y_max = "<<y-delta-y_max<<endl;
}


ostream& Two_P_PS_Interpolated::into_stream(ostream& os)const{
	os<<"#\tM="<<M<<"\tqT="<<qT;
	if(cuts){
		if(0<a)	os	<<"\ta->"<<a;
		if(0<b)	os	<<"\tb->"<<b;
		if(ex)	os	<<"\tb01->["<<atanh(thb0)<<","<<atanh(thb1)<<"]";
	}
	os	<<endl
		<<"#\ty\tPS(y)"<<endl;
	gsl_interpolator.into_stream(os);
	return os;
}

ostream& Two_P_PS_Interpolated::into_stream(ostream& os, const size_t N)const{
	os<<"#\tM="<<M<<"\tqT="<<qT;
	if(cuts){
		if(0<a)	os	<<"\ta->"<<a;
		if(0<b)	os	<<"\tb->"<<b;
		if(ex)	os	<<"\tb01->["<<atanh(thb0)<<","<<atanh(thb1)<<"]";
	}
	os	<<endl
		<<"#\ty\tPS(y)"<<endl;
	gsl_interpolator.into_stream(os,N);
	return os;
}
