#ifndef QDCS_H
#define QDCS_H1

#include "QDCS_Hardfunc.h"
#include "QDCS_Sigma0.h"
#include "Res.h"
#include "FO.h"

namespace CuTe {

namespace NS_QDCS {

//using NS_Polynom::Continious_Order_Reduction;

using namespace NS_enums::NS_MS;
using namespace NS_enums::NS_RES;

using NS_CS_Params::SOPS_CS_Params;
using NS_CS_Params::CS_Parameter_LL;
using NS_CS_Params::CS_Parameter_LLO;
using NS_CS_Params::CS_Parameter_LO;
using NS_FO::Quad_Differential_Kernel_FO;
using NS_FO::FO_Kinematic_Paramete_Set;
using NS_Hard::Hardfunc_SQT;
using NS_Hard::Hardfunc_LL;
using NS_Hard::Hardfunc_FO;
//using NS_Hard::Hardfunc_MC;
using NS_Hard::MS_Factor;
using NS_Kinematics::Kinematics;
using NS_LHAFUNC::PDF;
using NS_Resumed::Quad_Differential_Kernel_LL;
using NS_Resumed::Quad_Differential_Kernel_LLO;
using NS_Resumed::Res_Kinematic_Parameters;

using std::vector;
using std::endl;

class Quad_Differential_Crosssection{
protected:
	static const int N_RESULTS = RES_WRONG_STATEMENT;

	//Only for Copy
	const SOPS_CS_Params* _fc_op;
	vector<PDF>* _fc_pdfs;
	const Res_Kinematic_Parameters *_fc_LL_Kin;
	const FO_Kinematic_Paramete_Set *_fc_FO_Kin;
protected:

	vector<double>	results;
	vector<bool> 	to_calc;

//	MS ms;
	vector<int> order;

	CS_Parameter_LL		p_LL;
	CS_Parameter_LLO	p_LLO;
	CS_Parameter_LO		p_LO;

	Sigma_0 A;
	Hardfunc_LL		H_RES;
	MS_Factor		MS_F;
	Hardfunc_SQT	H_SQT;
	Hardfunc_FO		H_FO;


	Quad_Differential_Kernel_LL		QD_LL;
	Quad_Differential_Kernel_LLO	QD_LLO;
	Quad_Differential_Kernel_FO		QD_LO;

	double res_reduction;

	bool not_exponent;
	bool without_combined_result;

protected:

//	inline double _calc_RES(const double times){
//		results[RES_LL]	= ( QD_LL.mul_this_by_Pol(H_RES,order[RES_LL])).of_1()*times;
//		return results[RES_LL];
//	}
//
//	inline double _calc_FO(const double times){
//		results[RES_LO]	= ( QD_LO.mul_this_by_Pol(MS_F,order[RES_LO])).of_1()*times;
//		return results[RES_LO];
//	}
//
//	inline double _calc_SQT(const double times){
//		results[RES_LLO]	= (QD_LLO.mul_this_by_Pol(H_SQT,order[RES_LLO])).of_1()*times;
//		return results[RES_LLO];
//	}

	template< RES thisres>
	inline double _calc(const double times);

//	template< RES thisres>
//	inline double _calc(const double dy=1.){
//		const double times(A()*dy);
//
//		switch (thisres) {
//		case RES_LL:
//			_calc_RES(times);
//			break;
//		case RES_LLO:
//			_calc_SQT(times);
//			break;
//		case RES_LO:
//			_calc_FO(times);
//			break;
//		case RES_MATCHED:
//			_calc_RES(times);
//			_calc_SQT(times);
//			_calc_FO(times);
//
//			results[RES_MATCHING_CORR]	= results[RES_LO] - results[RES_LLO];
//			results[RES_RES_CORR]		= results[RES_LL] - results[RES_LLO];
//			results[thisres]			= results[RES_LL] + results[RES_MATCHING_CORR];
//			break;
//		case RES_MATCHING_CORR:
//			_calc_SQT(times);
//			_calc_FO(times);
//
//			results[thisres]	= results[RES_LO] - results[RES_LLO];
//			break;
//		case RES_RES_CORR:
//			_calc_RES(times);
//			_calc_SQT(times);
//
//			results[thisres]	= results[RES_LL] - results[RES_LLO];
//			break;
//		default:
//			break;
//		}
//		return results[thisres];
//	}

	template< RES thisres >
	inline double* _unset_f(double f[])const{
//		NS_Utils::DOUT dout("QDCS::_unset_f","f");
		*f	= 0.;	// result (thisres)


		switch (thisres) {
			case RES_MATCHED:
				if(without_combined_result)	--f;
				*(++f)	= 0.;	//LL
				*(++f)	= 0.;	//FO
				*(++f)	= 0.;	//SQT
				*(++f)	= 0.;	//MC
				*(++f)	= 0.;	//RC
				break;
			case RES_MATCHING_CORR:
				if(without_combined_result)	--f;
				*(++f)	= 0.;	//FO
				*(++f)	= 0.;	//SQT
				break;
			case RES_RES_CORR:
				if(without_combined_result)	--f;
				*(++f)	= 0.;	//LL
				*(++f)	= 0.;	//SQT
				break;
			default:
				break;
		}
		return f;
	}


public:

	inline void init(const SOPS_CS_Params& op, vector<PDF>& pdfs, const Res_Kinematic_Parameters *LL_Kin, const FO_Kinematic_Paramete_Set *FO_Kin, const bool not_exponent, const bool without_combined_result){
		this->without_combined_result = without_combined_result;
		this->not_exponent	= not_exponent;

		_fc_op		= &(op);
		_fc_pdfs	= &(pdfs);
		_fc_LL_Kin	= (LL_Kin);
		_fc_FO_Kin	= (FO_Kin);
//		NS_Utils::DOUT dout("Quad_Differential_Crosssection::init","op,pdfs");
//
//		dout()	<<"Res(p_LL ) = "<<p_LL.get_res()<<endl;
//		dout()	<<"Res(p_LLO) = "<<p_LLO.get_res()<<endl;
//		dout()	<<"Res(p_LO ) = "<<p_LO.get_res()<<endl;

//		for(int i=0; i<pdfs.size(); ++i){
//			dout()<<"pdfs["<<i<<"].properties:"<<endl;
//			cout<<pdfs[i].properties(dout.get_ntab()+1);
//		}
//		_op		= &op;
//		_pdfs	= &pdfs;

//		dout()	<<"init(results,to_calc)"<<endl;

		results.assign(N_RESULTS,0.);
		to_calc.assign(N_RESULTS,false);


		switch (op.Scheme.resummation()) {
		case RES_LL:
			to_calc[RES_LL				]	= true;
			break;
		case RES_LLO:
			to_calc[RES_LLO				]	= true;
			break;
		case RES_LO:
			to_calc[RES_LO				]	= true;
			break;
		case RES_MATCHING_CORR:
			to_calc[RES_LLO				]	= true;
			to_calc[RES_LO				]	= true;
			to_calc[RES_MATCHING_CORR	]	= true;
			break;
		case RES_RES_CORR:
			to_calc[RES_LL				]	= true;
			to_calc[RES_LLO				]	= true;
			to_calc[RES_RES_CORR		]	= true;
			break;
		case RES_MATCHED:
			to_calc[RES_LL				]	= true;
			to_calc[RES_LLO				]	= true;
			to_calc[RES_LO				]	= true;
			to_calc[RES_MATCHING_CORR	]	= true;
			to_calc[RES_MATCHED			]	= true;
			break;
		default:
			break;
		}


//		dout()	<<"init(p_LL,A)"<<endl;
		p_LL.init(op,pdfs);
		if(not_exponent) p_LL.exponent	= !p_LL.exponent;

		A.init(&p_LL);

		if(to_calc[RES_LL]){
//			dout()	<<"init(RES_LL)"<<endl;
			H_RES.init(&p_LL);
			QD_LL.init(&p_LL,LL_Kin);

			order[RES_LL]	= p_LL.order;
		}

		MS_F.init(&p_LL,op.Scheme.MScheme());

		if(to_calc[RES_LLO]){
//			dout()	<<"init(RES_LLO)"<<endl;
			p_LLO.init(op,pdfs);
			H_SQT.init(&p_LLO,false);
			QD_LLO.init(&p_LLO,LL_Kin);

			order[RES_LLO]	= p_LLO.order;
		}

		if(to_calc[RES_LO]){
//			dout()	<<"init(RES_LO)"<<endl;
			p_LO.init(op,pdfs);
			H_FO.init(&p_LO);
			QD_LO.init(&p_LO,FO_Kin);

			order[RES_LO]	= p_LO.order;
		}

//		dout()	<<"init(order)"<<endl;
		if(to_calc[RES_MATCHING_CORR]) 	order[RES_MATCHING_CORR]	= p_LO.order;
		if(to_calc[RES_RES_CORR]) 		order[RES_RES_CORR]			= p_LL.order;
		if(to_calc[RES_MATCHED]) 		order[RES_MATCHED]			= p_LL.order;

	}

	inline void set_qT_mu(const double qT, const Res_Kinematic_Parameters *LL_Kin, const FO_Kinematic_Paramete_Set *FO_Kin, const double mu, const double mu_star, const double res_reduction){//, const Continious_Order_Reduction *CO_reduction=NULL){
		std::stringstream ss;
		ss<<"qT="<<qT<<",mu="<<mu<<",mu*="<<mu_star<<",f="<<res_reduction;
//		NS_Utils::DOUT dout("QDCS::set_qT_mu",ss.str());
//		dout()	<<"qT="<<qT	<<"\t"<<"mu="<<mu	<<"\t"<<"mu*="<<mu_star	<<endl;

		this->res_reduction	= res_reduction;


		p_LL.update_qT_mu(qT,mu);
		A.update_qT_mu(qT,&p_LL);
//		dout()	<<"A = "<<A()<<endl;

		MS_F.update_mu(qT,mu,mu_star,res_reduction);


		if(to_calc[RES_LLO]){
//			NS_Utils::DOUT dout("QDCS::set_qT_mu","SQT");
			p_LLO.update_qT_mu(qT,mu);
			H_SQT.update_mu(mu);//,H_reduction);
			QD_LLO.set_qT_mu(LL_Kin);//,CO_reduction);

//			dout()	<<"H_SQT:"<<H_SQT<<endl;
		}
		if(to_calc[RES_LO]){
//			NS_Utils::DOUT dout("QDCS::set_qT_mu","FO");
			p_LO.update_qT_mu(qT,mu);
			H_FO.update_mu(mu);
			QD_LO.set_qT_mu(FO_Kin);

//			dout()	<<"H_FO:"<<H_FO<<endl;
		}
		if(to_calc[RES_LL]){
//			NS_Utils::DOUT dout("QDCS::set_qT_mu","RES");
			H_RES.update_mu(mu);//,H_reduction);
			if(p_LL.improved_pc()){
				improve_Pol_a(H_RES,p_LL.order);
			}
			QD_LL.set_qT_mu(LL_Kin);//,CO_reduction);

//			dout()	<<"H_RES:"<<H_RES<<endl;
		}

//		dout()	<<"H_LL = "<<H_LL	<<endl;
//		dout()	<<"H_LLO = "<<H_LLO	<<endl;
//		dout()	<<"H_LO = "<<H_LO	<<endl;
//		dout()	<<"H = "<<H	<<endl;


	}

	template< RES thisres>
	inline void set_y(const Res_Kinematic_Parameters *LL_Kin, const FO_Kinematic_Paramete_Set *FO_Kin){
//		NS_Utils::DOUT dout("Quad_Differential_Crosssection::set_y","y");

		switch (thisres) {
			case RES_LL:
				QD_LL.set_y(LL_Kin);
				break;
			case RES_LLO:
				QD_LLO.set_y(LL_Kin);
				break;
			case RES_LO:
//				p_LO.update_y(y);
				QD_LO.set_y(FO_Kin);
				break;
			case RES_MATCHED:
				QD_LL.set_y(LL_Kin);
				/* no break */
			case RES_MATCHING_CORR:
//				p_LLO.update_y(y);
				QD_LLO.set_y(LL_Kin);
//				p_LO.update_y(y);
				QD_LO.set_y(FO_Kin);
				break;
			case RES_RES_CORR:
				QD_LL.set_y(LL_Kin);
				QD_LLO.set_y(LL_Kin);
				break;
			default:
				break;
		}
	}


	template< RES thisres>
	inline double update_xx(const Res_Kinematic_Parameters *LL_Kin, const FO_Kinematic_Paramete_Set *FO_Kin, const double dy){
//		NS_Utils::DOUT dout("QDCS::update_xx");

		if(dy<=0.) return 0.;

		switch (thisres) {
			case RES_LL:
				QD_LL.update_zz(LL_Kin);
				break;
			case RES_LLO:
				QD_LLO.update_zz(LL_Kin);
				break;
			case RES_LO:
				QD_LO.update_xx(FO_Kin);
				break;
			case RES_MATCHED:
				QD_LL.update_zz(LL_Kin);
				/* no break */
			case RES_MATCHING_CORR:
				QD_LLO.update_zz(LL_Kin);
				QD_LO.update_xx(FO_Kin);
				break;
			case RES_RES_CORR:
				QD_LL.update_zz(LL_Kin);
				QD_LLO.update_zz(LL_Kin);
				break;
			default:
				break;
		}
		return _calc<thisres>(dy*A());
	}

	template< RES thisres>
	inline double update_y_xx(const Res_Kinematic_Parameters *LL_Kin, const FO_Kinematic_Paramete_Set *FO_Kin, const double dy){
//		NS_Utils::DOUT dout("QDCS::update_y_xx");

		if(dy<=0.) return 0.;

		set_y<thisres>(LL_Kin,FO_Kin);
		return update_xx<thisres>(LL_Kin,FO_Kin,dy);
	}


	template< RES thisres>
	inline double* update_xx_set_f(const Res_Kinematic_Parameters *LL_Kin, const FO_Kinematic_Paramete_Set *FO_Kin, const double dy, double f[]){
//		NS_Utils::DOUT dout("QDCS::update_xx");

		if(dy<=0.) return _unset_f<thisres>(f);

		switch (thisres) {
			case RES_LL:
				QD_LL.update_zz(LL_Kin);
				break;
			case RES_LLO:
				QD_LLO.update_zz(LL_Kin);
				break;
			case RES_LO:
				QD_LO.update_xx(FO_Kin);
				break;
			case RES_MATCHED:
				QD_LL.update_zz(LL_Kin);
				/* no break */
			case RES_MATCHING_CORR:
				QD_LLO.update_zz(LL_Kin);
				QD_LO.update_xx(FO_Kin);
				break;
			case RES_RES_CORR:
				QD_LL.update_zz(LL_Kin);
				QD_LLO.update_zz(LL_Kin);
				break;
			default:
				break;
		}

		*f	= _calc<thisres>(dy*A());	//result<thisres>


		switch (thisres) {
			case RES_MATCHED:
				if(without_combined_result)	--f;
				*(++f)	= results[RES_LL];	//LL
				*(++f)	= results[RES_LO];	//FO
				*(++f)	= results[RES_LLO];	//SQT
				*(++f)	= results[RES_MATCHING_CORR];	//MC
				*(++f)	= results[RES_RES_CORR];	//RC
				break;
			case RES_MATCHING_CORR:
				if(without_combined_result)	--f;
				*(++f)	= results[RES_LO];	//FO
				*(++f)	= results[RES_LLO];	//SQT
				break;
			case RES_RES_CORR:
				if(without_combined_result)	--f;
				*(++f)	= results[RES_LL];	//LL
				*(++f)	= results[RES_LLO];	//SQT
				break;
			default:
				break;
		}


		return f;
	}

	template< RES thisres>
	inline double* update_y_xx_set_f(const Res_Kinematic_Parameters *LL_Kin, const FO_Kinematic_Paramete_Set *FO_Kin, const double dy, double f[]){
//		NS_Utils::DOUT dout("QDCS::update_y_xx_set_f");

		if(dy<=0.) return _unset_f<thisres>(f);

		set_y<thisres>(LL_Kin,FO_Kin);
		return update_xx_set_f<thisres>(LL_Kin,FO_Kin,dy,f);
	}

	inline void print_pdf(const std::string s="")const{
		if(to_calc[RES_LL])		QD_LL.print_pdf(s+"LL");
		if(to_calc[RES_LLO])	QD_LLO.print_pdf(s+"LLO");
		if(to_calc[RES_LO])		QD_LO.print_pdf(s+"LO");
	}



	inline Quad_Differential_Crosssection(const bool without_combined_result=false)
	:_fc_op(NULL),_fc_pdfs(NULL),_fc_LL_Kin(NULL),_fc_FO_Kin(NULL)
	,results(N_RESULTS,0.)
	,to_calc(N_RESULTS,false)
//	,ms()
	,p_LL(),p_LLO(),p_LO()
	,order(N_RESULTS,0)
	,A()
	,H_RES(),H_SQT(),H_FO(),MS_F()
	,QD_LL(),QD_LLO(),QD_LO()
	,res_reduction()
	,not_exponent(false)
	,without_combined_result(without_combined_result){
//		NS_Utils::DOUT dout("Quad_Differential_Crosssection");
	}

	inline Quad_Differential_Crosssection(const SOPS_CS_Params& op,vector<NS_LHAFUNC::PDF>& pdfs, const Res_Kinematic_Parameters *LL_Kin,const FO_Kinematic_Paramete_Set *FO_Kin, const bool not_exponent, const bool without_combined_result)
	:_fc_op(NULL),_fc_pdfs(NULL),_fc_LL_Kin(NULL),_fc_FO_Kin(NULL)
	,results(N_RESULTS,0.)
	,to_calc(N_RESULTS,false)
//	,ms()
	,p_LL(),p_LLO(),p_LO()
	,order(N_RESULTS,0)
	,A()
	,H_RES(),H_SQT(),H_FO(),MS_F()
	,QD_LL(),QD_LLO(),QD_LO()
	,res_reduction()
	,not_exponent(not_exponent)
	,without_combined_result(without_combined_result){
//		NS_Utils::DOUT dout("Quad_Differential_Crosssection","op,pdfs,LL_Kin,FO_Kin");
		init(op,pdfs,LL_Kin,FO_Kin,not_exponent,without_combined_result);
//		print_pdf(dout.TAB());
	}

	inline Quad_Differential_Crosssection& operator=(const Quad_Differential_Crosssection& QDCS){
//		NS_Utils::DOUT dout("Quad_Differential_Crosssection::operator=","QDCS");

		if((QDCS._fc_op!=NULL)&&(QDCS._fc_pdfs!=NULL)){
			init(*(QDCS._fc_op),*(QDCS._fc_pdfs),QDCS._fc_LL_Kin,QDCS._fc_FO_Kin,QDCS.not_exponent,QDCS.without_combined_result);
		}else{
			_fc_op		= (NULL);
			_fc_pdfs	= (NULL);
			_fc_LL_Kin	= (NULL);
			_fc_FO_Kin	= (NULL);
		}

		results	= (QDCS.results);
//		to_calc	= (QDCS.to_calc);
//		mscheme	= (QDCS.mscheme);
//		p_LL	= (QDCS.p_LL);
//		p_LLO	= (QDCS.p_LLO);
//		p_LO	= (QDCS.p_LO);
//
//		order	= (QDCS.order);
//
//		A		= (QDCS.A);
//
//		H_LL	= (QDCS.H_LL);
//		H_LLO	= (QDCS.H_LLO);
//		H_LO	= (QDCS.H_LO);
//
//		QD_LL	= (QDCS.QD_LL);
//		QD_LLO	= (QDCS.QD_LLO);
//		QD_LO	= (QDCS.QD_LO);
//
////		print_pdf(dout.TAB());

		return *this;
	}

	inline Quad_Differential_Crosssection(const Quad_Differential_Crosssection& QDCS)
	:_fc_op(NULL),_fc_pdfs(NULL),_fc_LL_Kin(NULL),_fc_FO_Kin(NULL)
	,results(N_RESULTS,0.)
	,to_calc(N_RESULTS,false)
//	,ms()
	,p_LL(),p_LLO(),p_LO()
	,order(N_RESULTS,0)
	,A()
	,H_RES(),H_SQT(),H_FO(),MS_F()
	,QD_LL(),QD_LLO(),QD_LO()
	,res_reduction()
	,without_combined_result(false){
//	,results(QDCS.results),to_calc(QDCS.to_calc),mscheme(QDCS.mscheme)
//	,p_LL(QDCS.p_LL),p_LLO(QDCS.p_LLO),p_LO(QDCS.p_LO)
//	,order(QDCS.order)
//	,A(QDCS.A)
//	,H_LL(QDCS.H_LL),H_LLO(QDCS.H_LLO),H_LO(QDCS.H_LO)
//	,QD_LL(QDCS.QD_LL),QD_LLO(QDCS.QD_LLO),QD_LO(QDCS.QD_LO){
//		NS_Utils::DOUT dout("Quad_Differential_Crosssection","QDCS");
//		print_pdf(dout.TAB());
		operator=(QDCS);
	}

	inline ~Quad_Differential_Crosssection(){
//		NS_Utils::DOUT dout("~Quad_Differential_Crosssection");
	}

	inline double get_MSF()const{
		return MS_F.of_1();
	}

	inline double get_TF()const{
		return res_reduction;
	}

	inline void change_NPE(const double Lambda_NP, const NS_enums::NS_NP::CUTOFF cutoff){
		p_LL.change_NPE(Lambda_NP,cutoff);
		if(to_calc[RES_LL])	QD_LL.change_NPE();
	}

	inline void change_PDF(vector<PDF>& pdfs){

		_fc_pdfs	= &(pdfs);
		p_LL.change_PDF(pdfs);
		if(to_calc[RES_LL]){
			H_RES.change_AS(&p_LL);
			QD_LL.change_PDF(&p_LL);
		}
		if(to_calc[RES_LLO]){
			p_LLO.change_PDF(pdfs);
			H_SQT.change_AS(&p_LLO);
			QD_LLO.change_PDF(&p_LLO);
		}
		if(to_calc[RES_LO]){
			p_LO.change_PDF(pdfs);
			H_FO.change_AS(&p_LO);
			QD_LO.change_PDF(&p_LO);
		}
		MS_F.change_AS(&p_LL);

	}

};



template<  >
inline double Quad_Differential_Crosssection::_calc<RES_LL>(const double times){
	results[RES_LL]		= ( QD_LL.mul_this_by_Pol(H_RES,order[RES_LL])).of_1()*times;
	return results[RES_LL];
}

template<  >
inline double Quad_Differential_Crosssection::_calc<RES_LLO>(const double times){
	results[RES_LLO]	= (QD_LLO.mul_this_by_Pol(H_SQT,order[RES_LLO])).of_1()*times;
	return results[RES_LLO];
}

template<  >
inline double Quad_Differential_Crosssection::_calc<RES_LO>(const double times){
	results[RES_LO]	= (QD_LO.mul_this_by_Pol(H_FO,order[RES_LO])).of_1()*times;
	return results[RES_LO];
}

template<  >
inline double Quad_Differential_Crosssection::_calc<RES_RES_CORR>(const double times){
	results[RES_RES_CORR]	= res_reduction*(_calc<RES_LL>(times) - _calc<RES_LLO>(times));
	return results[RES_RES_CORR];
}

template<  >
inline double Quad_Differential_Crosssection::_calc<RES_MATCHING_CORR>(const double times){
	_calc<RES_LO>(times);
	_calc<RES_LLO>(times);
	results[RES_MATCHING_CORR]	 = ( QD_LO.mul_this_by_Pol(MS_F,order[RES_LO ])).of_1();
	results[RES_MATCHING_CORR]	-= (QD_LLO.mul_this_by_Pol(MS_F,order[RES_LLO])).of_1();
	results[RES_MATCHING_CORR]	*= times;
	return results[RES_MATCHING_CORR];
}

template<  >
inline double Quad_Differential_Crosssection::_calc<RES_MATCHED>(const double times){
	_calc<RES_LL>(times);
	_calc<RES_MATCHING_CORR>(times);

	results[RES_RES_CORR]		= res_reduction*(results[RES_LL] - results[RES_LLO]);

	results[RES_MATCHED]		= res_reduction*results[RES_LL];
	results[RES_MATCHED]		+= ( QD_LO.of_1() - res_reduction*QD_LLO.of_1() )*times;

	return results[RES_MATCHED];
}

}  // namespace NS_CROSSSECTION



}  // namespace CuTe

#endif
