#include "QDCS_Hardfunc.h"


namespace CuTe {
namespace NS_Hard {

typedef NS_Polynom::Pol< complex<double> > CPol_a;
typedef NS_Polynom::Pol< complex<double> > CPol_L;
typedef NS_Polynom::Pol< CPol_L > CPol_a_L;

using NS_Anomalous_Dimension::Anomalous_Dimension;
using NS_Anomalous_Dimension::AD_Beta;
using NS_Anomalous_Dimension::AD_GammaCusp;
using NS_Anomalous_Dimension::AD_gamma_q;
using NS_Anomalous_Dimension::AD_gamma_V;
using NS_Anomalous_Dimension::AD_gamma_g;
using NS_Anomalous_Dimension::AD_gamma_t;
using NS_Anomalous_Dimension::AD_gamma_S;

using namespace NS_enums::NS_RES;
using namespace NS_enums::NS_PROD;
using namespace NS_enums::NS_MS;

using NS_LHAFUNC::Parton;

using std::cerr;
using std::cout;
using std::endl;

class U_exp: public CPol_a{
protected:
	const A_strong* a_strong;
	constexpr static const double epsilon=1e-6;
	AD_Beta Beta;
	AD_GammaCusp Gamma;

public:
	Pol_a real;

	inline void change_AS(const A_strong* a_strong){
		this->a_strong	= a_strong;
	}

protected:

	inline virtual void v_AD_select(int n){
		cout << "*************************************************************************" << endl
			 << "**                      class U_exp::v_AD_select not defined -> exit(1)      **" << endl
			 << "*************************************************************************" << endl;
		exit(1);
	}

	inline void AD_select(int n){
		Beta.select(n);
		Gamma.select(n);
		v_AD_select(n);
	}


	inline CPol_a a_gamma(complex<double> a1, complex<double> a2, Anomalous_Dimension& gamma){
//		cout	<< "->a("<<a1<<","<<a2<<")" << endl;
		CPol_a erg(3,0.);
		erg[0]= log(a2/a1)*gamma(0);
		if(_order>0) erg[1]=(a2-a1)*(gamma(1)-gamma(0)*Beta(1,0));
		if(_order>1) erg[2]=0.5*(a2*a2-a1*a1)*(
									gamma(2)
									-gamma(0)*Beta(2,0)
									-Beta(1,0)*(gamma(1)-gamma(0)*Beta(1,0))
							);
		erg.order(_order);
		erg*=0.5*Beta(0,0);
//		cout 	<< "<-\t a = " << endl << "\t" << erg << endl;
		return erg;
	}

	inline CPol_a S(complex<double> a_1, complex<double> a_2){
		complex<double> r=a_2/a_1;
		complex<double> l=log(r);
//		cout<<"U_exp::S order="<<_order<<endl;
//		cout	<< "->S("<<a_1<<","<<a_2<<")" << endl
//				<< "\t r = " << r << "\t l = " << l << endl;
		CPol_a S(2,0.);
		S[0]=(r-1.0-r*l)/(r*a_1);
		S[0]+=(Gamma(1,0)-Beta(1,0))*(-r+l+1.)+0.5*Beta(1,0)*l*l;
		S[1]=(Beta(1,0)*Gamma(1,0)-Beta(2,0))*(1.-r+r*l)
			+(Beta(1,2)-Beta(2,0))*(1.-r)*l
			-0.5*(Beta(1,2)-Beta(2,0)-Beta(1,0)*Gamma(1,0)+Gamma(2,0))*(1.-r*2.+r*r);
		S[2]=(
				8.*Beta(1,3)
				-7.*Beta(1,0)*Beta(2,0)
				-Beta(3,0)
				-11.*Beta(1,2)*Gamma(1,0)
				+8.*Beta(2,0)*Gamma(1,0)
				+5.*Beta(1,0)*Gamma(2,0)
				-2.*Gamma(3,0)
			)/12.
			+0.5*(
				-Beta(1,3)
				+2.*Beta(1,0)*Beta(2,0)
				-Beta(3,0)
			)*l
			+(
				-Beta(1,3)
				+Beta(1,0)*Beta(2,0)
				+Beta(1,2)*Gamma(1,0)
				-Beta(2,0)*Gamma(1,0)
				+(
					0.25*(
						Beta(1,0)*Beta(2,0)
						-Beta(3,0)
						+Beta(1,2)*Gamma(1,0)
						-3.*Beta(1,0)*Gamma(2,0)
						+2.*Gamma(3,0)
					)
					+0.5*(
						Beta(1,3)
						-Beta(1,0)*Beta(2,0)
						-Beta(1,2)*Gamma(1,0)
						+Beta(1,0)*Gamma(2,0)
					)*l
					+(
						Beta(1,3)
						-2.*Beta(1,0)*Beta(2,0)
						+Beta(3,0)
						-Beta(1,2)*Gamma(1,0)
						+Beta(2,0)*Gamma(1,0)
						+Beta(1,0)*Gamma(2,0)
						-Gamma(3,0)
			)/3.*r	)*r	)*r;
		S.mul_this_by_scalar(0.25*Gamma(0)*Beta(0,2));
		S.order(_order);
		S.set_of(a_1);
//		cout 	<< "<-\t S = " << endl << "\t" << S << endl;
		return S;
	}

	inline virtual CPol_a v_u_exp(complex<double> a_h, complex<double> a_s, complex<double> L_h){
		cout << "*************************************************************************" << endl
			 << "**                      class U_exp::v_u_exp not defined -> exit(1)      **" << endl
			 << "*************************************************************************" << endl;
		exit(1);
	}

public:

	inline void update(complex<double> M, complex<double> a_h, int nf_h, complex<double> a_s, int nf_s, complex<double> L_h){
		mul_this_by_scalar(0.);

		for(Parton q(nf_h); q>Parton(nf_s); --q){

			AD_select(q.lha());
			double mu	= a_strong->quarkThreshold(q)+epsilon;
			double a	= a_strong->as(mu);

//			dout()	<<"mu="<<mu	<<"\t"<<"a="<<a	<<"\t"<<"n="<<q.flv()	<<endl;

			add_this_Pol(v_u_exp(a_h,a,L_h),_order);

			mu -= 2.*epsilon;

			a_h	= a_strong->as(mu);
			L_h	= 2.*log(M/mu);
		}
		AD_select(nf_s);
		add_this_Pol(v_u_exp(a_h,a_s,L_h),_order);
	}

	inline void update(complex<double> M, complex<double> mu_h, complex<double> mu_s){
		complex<double> L_h = 2.*log(M/mu_h);

		int nf_h = a_strong->nf(mu_h);
		complex<double> a_h = a_strong->as(mu_h);

		int nf_s = a_strong->nf( mu_s);
		complex<double> a_s = a_strong->as(mu_s);

//		dout()	<<"mu_h="<<mu_h	<<"\t"<<"a_h="<<a_h	<<"\t"<<"n_h="<<n_f_h	<<endl;
//		dout()	<<"mu_s="<<mu_s	<<"\t"<<"a_s="<<a_s	<<"\t"<<"n_s="<<n_f_s	<<endl;

		return update(M,a_h,nf_h,a_s,nf_s,L_h);

	}

	inline Pol_a& real_of(complex<double> M, complex<double> a_h, int nf_h, complex<double> a_s, int nf_s, complex<double> L_h){
//		DOUT dout("U_exp::real_of","M,mu_h,mu");
		update(M,a_h,nf_h,a_s,nf_s,L_h);
		real.order(_order);
		for(int i=0; i<=_order; ++i) real[i]= _P[i].real();
//		dout()	<<"M="<<M	<<"\t"<<"mu_h="<<mu_h	<<"\t"<<"mu="<<mu	<<endl;
//		dout()	<<"="<<real	<<endl;
		return real;
	}

	inline Pol_a& real_of(complex<double> M, complex<double> mu_h, complex<double> mu){
//		DOUT dout("U_exp::real_of","M,mu_h,mu");
		update(M,mu_h,mu);
		real.order(_order);
		for(int i=0; i<=_order; ++i) real[i]= _P[i].real();
//		dout()	<<"M="<<M	<<"\t"<<"mu_h="<<mu_h	<<"\t"<<"mu="<<mu	<<endl;
//		dout()	<<"="<<real	<<endl;
		return real;
	}

	inline U_exp(int order_,const A_strong* a_strong, const PROD prod)
	:CPol_a(order_,complex<double>(0.,0.))
	,a_strong(a_strong)
	,Beta()
	,Gamma(prod)
	,real(order_,0.){
	}


	virtual ~U_exp(){}
};

class U_exp_V:public U_exp{
protected:
	AD_gamma_q gamma_q;
	AD_gamma_V gamma_V;

	inline void v_AD_select(int n){
		gamma_q.select(n);
		gamma_V.select(n);
	}

	inline CPol_a v_u_exp(complex<double> a_h, complex<double> a_s, complex<double> L_h){
		return ( S(a_h,a_s)*2.*C_F
				 -a_gamma(a_h,a_s,gamma_V)
				 -a_gamma(a_h,a_s,Gamma)*L_h*C_F
				);
	}
public:
	inline U_exp_V(int order,const A_strong* a_strong)
	:U_exp(order,a_strong,PROD_V)
	,gamma_q(AD_gamma_q())
	,gamma_V(AD_gamma_V(&gamma_q)){

	}
};

class U_exp_S:public U_exp{
protected:
	AD_gamma_g gamma_g;
	AD_gamma_t gamma_t;
	AD_gamma_S gamma_S;

	inline void v_AD_select(int n){
		gamma_g.select(n);
		gamma_t.select(n);
		gamma_S.select(n);
	}

	inline CPol_a v_u_exp(complex<double> a_h, complex<double> a_s, complex<double> L_h){
//		cout << "U_exp_S::v_u_exp("<<a_h<<","<<a_s<<")" << endl;
//		cout << "\tS(a_h,a_s)*2.*C_A = " <<S(a_h,a_s)*2.*C_A <<endl;
//		cout << "\tagS(a_h,a_s) = " <<a_gamma(a_h,a_s,gamma_g)*2. <<endl;
//		cout << "\taG(a_h,a_s)*L_h*C_A = " <<a_gamma(a_h,a_s,Gamma)*L_h*C_A <<endl;
		return (	S(a_h,a_s)*2.*C_A
					-a_gamma(a_h,a_s,gamma_S)
//					-a_gamma(a_h,a_s,gamma_g)*2.//FIXME -a_gamma(a_h,a_s,gamma_S)
					-a_gamma(a_h,a_s,Gamma)*L_h*C_A
				);
	}
public:
	inline U_exp_S(int order,const A_strong* a_strong)
	:U_exp(order,a_strong,PROD_S)
	,gamma_g(AD_gamma_g())
	,gamma_t(AD_gamma_t(&Beta))
	,gamma_S(AD_gamma_S(&gamma_g, &gamma_t, &Beta)){

	}
};

class U_exp_t:public U_exp{
protected:
	AD_gamma_t gamma_t;

	inline void v_AD_select(int n){
		gamma_t.select(n);
	}

	inline CPol_a v_u_exp(complex<double> a_h, complex<double> a_s, complex<double> L_h){
//		DOUT dout("U_exp_t::v_u_exp","a_h,a_s,L_h");
//		dout()	<<"a_h="<<a_h	<<"\t"<<"a_s"<<a_s	<<"\t"<<"L_h="<<L_h	<<endl;
		return (-a_gamma(a_h,a_s,gamma_t));
	}
public:
	inline U_exp_t(int order,const A_strong* a_strong)
	:U_exp(order,a_strong,PROD_S)
	,gamma_t(AD_gamma_t(&Beta)){

	}
};


class C_exp:public CPol_a_L{
protected:
	int n_f_h;
	const A_strong* a_strong;
public:
	Pol_a real;

	inline void change_AS(const A_strong* a_strong){
		this->a_strong	= a_strong;
	}

	inline virtual void v_init(){
		cout << "*************************************************************************" << endl
			 << "**                      class C_exp::v_init() not defined -> exit(1)      **" << endl
			 << "*************************************************************************" << endl;
		exit(1);
	}

	inline C_exp(int order_,const A_strong* a_strong)
	:CPol_a_L(order_,CPol_a(0,complex<double>(0.,0.)))
	,n_f_h(0),a_strong(a_strong)
	,real(Pol_a(_order,0.)){
//		DOUT dout("C_exp","order_,a_strong");
//		dout()	<<".order = "<< order()	<<" = "<<_order	<<endl;
	}

public:

	inline Pol_a& real_of(complex<double> M, complex<double> mu_h){
//		DOUT dout("C_exp::real_of","M,mu_h");
		int n = a_strong->nf(mu_h);
		if(n_f_h!=n) {
			n_f_h=(n);
//			Anomalous_Dimension::select(n_f_h);
			v_init();
		}

		complex<double> a_h = a_strong->as(mu_h);
		complex<double> L_h = 2.*log(M/mu_h);

//		dout()	<<"M="<<M	<<"\t"<<"mu_h="<<mu_h		<<"\t"<<"a_h="<<a_h		<<"\t"<<"L_h="<<L_h	<<endl;


		real.order(_order);

//		using NS_Utils::operator <<;
//		dout()	<<"_P= "	<<_P	<<endl;

		real[0]	= _P[0].of(L_h).real();
		for(int i=1; i<=_order; ++i){
			real[i] = (_P[i].of(L_h)*a_h).real();
			a_h*=a_h;
		}
		//cout << "\ttmp("<< a <<"," <<L_h <<")\t=" << tmp << endl;
//		dout()	<<"="<<real	<<endl;
		return real;
	}

	inline const CPol_a_L operator() (complex<double> M, complex<double> mu_h){
		int n = a_strong->nf(mu_h);
		if(n_f_h!=n) {
			n_f_h=(n);
//			Anomalous_Dimension::select(n_f_h);
			v_init();
		}
		CPol_a_L tmp(*this);

		complex<double> a_h = a_strong->as(mu_h);
		complex<double> L_h = 2.*log(M/mu_h);
//		cout	<<"C_exp::operator("<<M<<","<<mu_h<<")" << endl
//				<<"\t"<<"a_h="<<a_h	<<"\t"<<"L_h="<<L_h	<<endl;

		tmp.set_of(a_h);
		for(int i=0; i<=_order; ++i) tmp[i].set_of(L_h);
		//cout << "\ttmp("<< a <<"," <<L_h <<")\t=" << tmp << endl;
		return tmp;
	}


	inline virtual ~C_exp(){}
};

class C_V_exp: public C_exp{
protected:

	inline CPol_L _HVF(){
		CPol_L _HVF(4,0.);
		_HVF[0]=0.125*255+0.5*7*PI2-83.0*PI4/360.0-30*zeta_3;
		_HVF[1]=-0.5*45-0.5*3*PI2+24*zeta_3;
		_HVF[2]=0.5*25-PI2/6.0;
		_HVF[3]=-3;
		_HVF[4]=0.5;
		return _HVF;
	}

	inline CPol_L _HVA(){
		CPol_L _HVA(4,0.);
		_HVA[0]=- 51157.0/648.0 - 337.0*PI2/108.0+11.0*PI4/45.0+313.0*zeta_3/9.0;
		_HVA[1]=2545.0/54.0+11.0*PI2/9.0-26*zeta_3;
		_HVA[2]=-233.0/18.0+PI2/3.0;
		_HVA[3]=11.0/9.0;
		_HVA[4]=0.0;
		return _HVA;
	}


	inline CPol_L _HVf(){
		CPol_L _HVf(4,0.);
		_HVf[0]=4085.0/162.0+23.0/27.0*PI2+4.0/9.0*zeta_3;
		_HVf[1]=-418.0/27.0-4.0*PI2/9.0;
		_HVf[2]=38.0/9.0;
		_HVf[3]=-4.0/9.0;
		_HVf[4]=0.0;
		return _HVf;
	}

	inline CPol_L _CV1(){
		CPol_L _CV1(2,0.);
		_CV1[0]=C_F*(-8+PI2/6.0);
		_CV1[1]=C_F*3;
		_CV1[2]=-C_F;
		return _CV1;
	}


	inline CPol_L _CV2(int n_f){
		return ((_HVF()*C_F + _HVA()*C_A + _HVf()*T_F*n_f)* C_F );
	}


	inline void v_init(){
//		DOUT dout("C_V_exp::v_init");
		int n=_order;
		order(2);
		_P[0]=CPol_L(0,0.);
		_P[1]=_CV1();
		_P[2]=_CV2(n_f_h)-_P[1]*_P[1]*0.5;
//		cout << "\tC_V_exp::update() -> " << *this << endl;
		order(n);
	}

public:
	inline C_V_exp(int order,const A_strong* a_strong):C_exp(order,a_strong){}
};

class C_S_exp: public C_exp{
protected:

	inline CPol_L _HSA(){
		CPol_L _HSA(4,0.);
		_HSA[0] = 5105./162. + 67.*PI2/36. + PI4/72. - 143.*zeta_3/9.;
		_HSA[1] = 80./27. - 11.*PI2/9. - 2.*zeta_3;
		_HSA[2] = -67./9. + PI2/6.;
		_HSA[3] = 11./9.;
		_HSA[4] = 0.5;
		return _HSA;
	}

	inline CPol_L _HSF(){
		CPol_L _HSF(1,0.);
		_HSF[0] = -67./3.+16.*zeta_3;
		_HSF[1] = 4.;
		return _HSF;
	}

	inline CPol_L _HSAF(){
		CPol_L _HSAF(3,0.);
		_HSAF[0] = (-1832./9.-5.*PI2-92*zeta_3)/9.;
		_HSAF[1] = (104./3.+4.*PI2)/9.;
		_HSAF[2] = 20./9.;
		_HSAF[3] = -4./9.;
		return _HSAF;
	}

	inline CPol_L _CS0(){
		CPol_L _CS0(0,1.);
		return _CS0;
	}

	inline CPol_L _CS1(){
		CPol_L _CS1(2,0.);
		_CS1[0]=C_A*PI2/6.0;
		_CS1[1]=0.0;
		_CS1[2]=-C_A;
		return _CS1;
	}

	inline CPol_L _CS2(int n_f){
		return (_HSA()*C_A2 + (_HSAF()*C_A + _HSF()*C_F)*T_F*n_f);
	}

	inline void v_init(){
		int n=_order;
		order(2);
		_P[0]=CPol_L(0,0.);
		_P[1]=_CS1();
		_P[2]=_CS2(n_f_h)-_P[1]*_P[1]*0.5;
		order(n);
	}

public:
	inline C_S_exp(int order, const A_strong* a_strong):C_exp(order,a_strong){}
};

class C_t_exp: public C_exp{
protected:

	inline CPol_L _HtA(int n_f){
		CPol_L _HtA(1,0.);
		_HtA[0] = 1063./36*C_A +(-15. - 94.*n_f)/18.*T_F;
		_HtA[1] = -7.*C_A;
		return _HtA;
	}

	inline CPol_L _HtF(int n_f){
		CPol_L _HtF(1,0.);
		_HtF[0] = 0.5*27.*C_F - 4./3*T_F - 5.*T_F*n_f;
		_HtF[1] =-8.*T_F*n_f;
		return _HtF;
	}

	inline CPol_L _HtAF(){
		CPol_L _HtAF(1,0.);
		_HtAF[0] = -100./3.;
		_HtAF[1] = 11.;
		return _HtAF;
	}

	inline CPol_L _Ct0(){
		CPol_L _Ct0(0,1.);
		return _Ct0;
	}

	inline CPol_L _Ct1(){
		CPol_L _Ct1(0,0.);
		_Ct1[0]=5.*C_A-3.*C_F;
		return _Ct1;
	}

	inline CPol_L _Ct2(int n_f){
		return (_HtA(n_f)*C_A + _HtAF()*C_A*C_F + _HtF(n_f)*C_F );
	}


	inline void v_init(){
		int n=_order;
		order(2);
		_P[0]=CPol_L(0,0.);
		_P[1]=_Ct1();
		_P[2]=_Ct2(n_f_h)-_P[1]*_P[1]*0.5;
		order(n);
	}

public:
	inline C_t_exp(int order, const A_strong* a_strong):C_exp(order,a_strong){}
};


void _Hardfunc_QCD5::_init(const CS_Parameter *p, const PROD prod, const bool resummation, const int order_a, const double M, const double M_t, const A_strong *AS){
	this->p		= p;
	this->M_t	= M_t;

	if(M_t>0.){
		this->M_t	= M_t;
	}else{
		this->M_t	= M;
	}

	delete C_t;		C_t	=NULL;
	delete U_t;		U_t	=NULL;

	this->order_a	= order_a;

	lnC.order(0);
	lnC[0] = 0.;
	lnU.order(0);
	lnU[0] = 0.;
	order(0);
	_P[0]=1.;

	switch (prod) {
	case PROD_Higgs:
		C_t		= new C_t_exp(order_a,AS);
		if(resummation){
			U_t		= new U_exp_t(order_a,AS);
		}
		break;
	case PROD_H_RS:
	case PROD_DY:
	case PROD_W:
	case PROD_Z:
		break;
	default:
		std::cerr	<<"_Hardfunc::init WRONG prod="<<prod<<" ->exit()"<<endl;
		exit(1);
		break;
	}
}

_Hardfunc_QCD5::~_Hardfunc_QCD5(){
	delete C_t;	// No = NULL
	delete U_t;	// No = NULL
}

void _Hardfunc_QCD5::change_AS(const CS_Parameter* p){
	_change_AS(p->a_strong);
}

void _Hardfunc_QCD5::_change_AS(const A_strong* AS){
	if(C_t!=NULL)	C_t->change_AS(AS);
	if(U_t!=NULL)	U_t->change_AS(AS);
}


void _Hardfunc_SCET::_init(const CS_Parameter *p, const PROD prod, const bool resummation, const int order_a, const double M, const double M_t, const A_strong *AS){
	_Hardfunc_QCD5::_init(p,prod,resummation,order_a,M,M_t,AS);

	this->M		= complex<double>(0.,M);

	delete C_SV;	C_SV=NULL;
	delete U_SV;	U_SV=NULL;

	switch (prod) {
	case PROD_Higgs:
	case PROD_H_RS:
		C_SV	= new C_S_exp(order_a,AS);
		if(resummation){
			U_SV	= new U_exp_S(order_a,AS);
		}
		break;
	case PROD_DY:
	case PROD_W:
	case PROD_Z:
		C_SV	= new C_V_exp(order_a,AS);
		if(resummation){
			U_SV	= new U_exp_V(order_a,AS);
		}
		break;
	default:
		std::cerr	<<"_Hardfunc::init WRONG prod="<<prod<<" ->exit()"<<endl;
		exit(1);
		break;
	}
}

Pol_a& _Hardfunc_SCET::_update_lnC(const complex<double> &M, const complex<double> &mu_h, const double M_t, const double mu_t){
	lnC.set_equalt_to(C_SV->real_of(M,mu_h),order_a);
	if(C_t!=NULL){
		lnC.add_this_Pol(C_t->real_of(M_t,mu_t),order_a);
	}
	return lnC;
}
Pol_a& _Hardfunc_SCET::_update_lnU(const complex<double> &M, const complex<double> &mu_h, const complex<double> &mu){
	return lnU.set_equalt_to(U_SV->real_of(M,mu_h,mu),order_a);
}
Pol_a& _Hardfunc_SCET::_update_lnU(const complex<double> &M, const complex<double> &mu_h, const double M_t, const double mu_t, const complex<double> &mu){
	_update_lnU(M,mu_h,mu);
	if(U_t!=NULL){
		lnU.add_this_Pol(U_t->real_of(M_t,mu_t,mu),order_a);
	}
	return lnU;
}

_Hardfunc_SCET::~_Hardfunc_SCET(){
	delete C_SV;	// No = NULL
	delete U_SV;	// No = NULL
}

void _Hardfunc_SCET::_change_AS(const A_strong* AS){
	_Hardfunc_QCD5::_change_AS(AS);
	if(C_SV!=NULL)	C_SV->change_AS(AS);
	if(U_SV!=NULL)	U_SV->change_AS(AS);
}


//Pol_a& Hardfunc::update_mu(const double mu, const double expfactor){
//	std::cerr	<<"virtual Hardfunc::update_mu ->exit()"<<endl;
//	exit(1);
//	return *this;
//}
//
//void Hardfunc::init(const CS_Parameter *p){
//	std::cerr	<<"virtual Hardfunc::init ->exit()"<<endl;
//	exit(1);
//}


void Hardfunc_FO::init(const CS_Parameter *p){
	const bool res(false);
	order_a	= p->order_as;
	if(order_a>0) --order_a;

	_init(p,p->prod,res,order_a,p->M,p->M_t,p->a_strong);
}

Pol_a& Hardfunc_FO::_update_mu(const double mu, const double expfactor){
	if(C_t!=NULL){

		lnC.set_equalt_to(C_t->real_of(M_t,mu),order_a);

		set_equalt_to(lnC,order_a);

		mul_this_by_scalar(+2.*expfactor);

		exponentiate();
	}

	return *this;
}

void Hardfunc_SQT::init(const CS_Parameter *p, const bool pi2){
	const bool res(pi2);
	order_a	= p->order_as;
	if(order_a>0) --order_a;

	_init(p,p->prod,res,order_a,p->M,p->M_t,p->a_strong);

	this->pi2	= pi2;
}

Pol_a& Hardfunc_SQT::update_mu(const double mu, const double expfactor){
//	std::stringstream ss;
//	ss	<<"mu="<<mu;
//	NS_Utils::DOUT	dout("Hardfunc_SQT::update_mu",ss.str());

	if(!pi2){
		_update_lnC(M,mu,M_t,mu);
		set_equalt_to(lnC,order_a);
	}else{
		complex<double> mu_h(0.,mu);

		_update_lnC(M,mu_h,M_t,mu);
		set_equalt_to(lnC,order_a);

		_update_lnU(M,mu_h,mu);
		add_this_Pol(lnU,order_a);

	}
//	dout()	<<"lnC(mu):"	 		<<lnC<<endl;
//	if(pi2)	dout()	<<"lnU(imu,mu):"<<lnU<<endl;
	mul_this_by_scalar(+2.*expfactor);

	return exponentiate();
}


void Hardfunc_LL::_init(const CS_Parameter *p, const int order_a,const bool exponent, const bool pi2){
	const bool res(true);

	_Hardfunc_SCET::_init(p,p->prod,res,order_a,p->M,p->M_t,p->a_strong);

	this->pi2		= pi2;
	this->exponent	= exponent;


	if(pi2){
		mu_h	= complex<double>(0.,p->mu_h);
	}else{
		mu_h	= complex<double>(p->mu_h,0.);
	}

	mu_t= p->mu_t;

}

void Hardfunc_LL::init(const CS_Parameter *p){
	_init(p,p->order_as,p->exponent,p->piquadrat);
}

Pol_a& Hardfunc_LL::update_mu(const double mu, const double expfactor){
//	std::stringstream ss;
//	ss	<<"mu="<<mu;
//	NS_Utils::DOUT	dout("Hardfunc_LL::update_mu",ss.str());

	_update_lnC(M,mu_h,M_t,mu_t);
	set_equalt_to(lnC,order_a);

	_update_lnU(M,mu_h,M_t,mu_t,mu);
	if(exponent){
		_P[0] += lnU.of_1(order_a);
	}else{
		add_this_Pol(lnU,order_a);
	}

	mul_this_by_scalar(+2.*expfactor);

//	dout()	<<"lnC(mu_h):"	 		<<lnC<<endl;
//	dout()	<<"lnU(mu_h,"<<mu<<"):"<<lnU<<endl;
	return exponentiate();
}


//void Hardfunc_MC::init(const CS_Parameter *p, const bool pi2){
//	order_a	= (p->order_as);
//	if(order_a>0) --order_a;
//
//	const bool exponent(false);
//
//	_init(p,order_a,exponent,pi2);
//}

const double MS_Factor::MU_LIMIT = 1.;

void MS_Factor::init(const CS_Parameter_LL *p, const NS_enums::NS_MS::MS ms, const bool as_number/*=true*/, const bool exponentiate_U/*=false*/, const bool order_mm/*=true*/){
	delete C_SV;	C_SV	= NULL;
	delete U_SV;	U_SV	= NULL;

	this->p		= p;
	this->ms	= ms;
	this->exponentiate_U	= exponentiate_U;
	this->as_number			= as_number;
	this->order_mm			= order_mm;
	this->M		= complex<double>(0.,p->M);
	this->mu_h	= p->mu_h;


	if(ms==MS_NAIVE){
		order_a = 0.;
	}{
		order_a	= p->order_as;
		if(order_a&&order_mm) --order_a;
		order(order_a);

		const A_strong* AS(p->a_strong);

		switch (p->prod) {
			case PROD_Higgs:
			case PROD_H_RS:
				C_SV	= new C_S_exp(order_a,AS);
				U_SV	= new U_exp_S(order_a,AS);
				break;
			case PROD_Z:
			case PROD_W:
			case PROD_DY:
				C_SV	= new C_V_exp(order_a,AS);
				U_SV	= new U_exp_V(order_a,AS);
				break;
			default:
				cerr	<<"MS_Factor::init wrong prod ="<<p->prod<<endl;
				exit(1);
				break;
		}

		switch (ms) {
			case MS_CONST_MU_PI2:
			case MS_CONST_QT_PI2:
			case MS_VAR_MU_PI2:
			case MS_VAR_QT_PI2:
				pi2	= true;
				mu_h	= complex<double>(0.,p->mu_h);
				break;
			default:
//			case MS_CONST_MU:
//			case MS_CONST_QT:
//			case MS_VAR_MU:
//			case MS_VAR_QT:
//			case MS_NAIVE:
//			case MS_WRONG:
				pi2 = false;
				break;
				break;
		}
	}
	mul_this_by_scalar(0.);
	_P[0] = 1.;

	lnC_RES = Pol_a(order_a,0.);
	lnC_SQT.set_equalt_to(lnC_RES);
	lnU_RES.set_equalt_to(lnC_RES);
//	lnU_SQT.set_equalt_to(lnC_RES);
//	if(!pi2) lnU_SQT.order(0);
}


void MS_Factor::_update_mu(const double mu, const double expfactor){
	lnC_RES.set_equalt_to(C_SV->real_of(M,mu_h),order_a);
	if(pi2){
		lnU_RES.set_equalt_to(U_SV->real_of(M,mu_h,complex<double>(0.,mu)),order_a);
		lnC_SQT.set_equalt_to(C_SV->real_of(M,complex<double>(0.,mu)),order_a);
	}else{
		lnU_RES.set_equalt_to(U_SV->real_of(M,mu_h,mu),order_a);
		lnC_SQT.set_equalt_to(C_SV->real_of(M,mu),order_a);
	}

	if(exponentiate_U){
		lnU_RES[0]	= lnU_RES.of_1(order_a);
		lnU_RES.order(0);
	}

	set_equalt_to(lnC_SQT,order_a);
//	if(pi2) add_this_Pol(lnU_SQT,order_a);
	mul_this_by_scalar(-1.);
	add_this_Pol(lnC_RES,order_a);
	add_this_Pol(lnU_RES,order_a);

	mul_this_by_scalar(2.*expfactor);

	exponentiate();

	if(as_number){
		_P[0]	= of_1(order_a);
		order(0);
	}
}


Pol_a& MS_Factor::update_mu(const double qT, double mu, const double mu_star, const double expfactor){
//	std::stringstream ss;
//	ss	<<"ms="<<ms<<"qT="<<qT<<",mu="<<mu<<",mu*="<<mu_star<<",t="<<expfactor;
//	NS_Utils::DOUT	dout("MS_Factor::update_mu",ss.str());

	switch (ms) {
		case MS_NAIVE:
			return *this;
			break;
		case MS_CONST_MU:
		case MS_CONST_MU_PI2:
			mu	= mu_star;
			break;
		case MS_CONST_QT:
		case MS_CONST_QT_PI2:
			mu	= qT;
			break;
		case MS_VAR_QT:
		case MS_VAR_QT_PI2:
			mu	*= qT/mu_star;	//= qT * mu/mu_star = qT*{ 1/2 , 1 , 2 }
			break;
		default:
//		case MS_VAR_MU_PI2:
//		case MS_VAR_MU:
//		case MS_WRONG:
//			mu	= mu;	//	= mu_star*{ 1/2 , 1 , 2 }
			break;
	}

	if(mu>=MU_LIMIT){
		_update_mu(mu,expfactor);
	}else{
		_update_mu(MU_LIMIT,expfactor);
		mul_this_by_scalar(NS_Utils::NS_pow::pow_3(mu/MU_LIMIT));
	}



//	to_stream(cout,dout.get_ntab());
	return *this;
}

std::ostream& MS_Factor::to_stream(std::ostream &os, const int ntabs){
	std::string tab("");
	for(int i=0; i<ntabs; ++i) tab += "\t";

	os	<<tab<<"MS_Factor:"<<endl;
	os	<<tab<<"\t"<<"o = "<<order_a<<"\t"<<"M = "<<M<<"\t"<<"mu_h = "<<mu_h<<"\t"<<endl;
	os	<<tab<<"\t"<<"ms = "<<ms	<<"\t["<<as_number<<exponentiate_U<<order_mm<<"]"	<<endl;
	os	<<tab<<"\t: = "<<*this<<endl;


	return os;
}

MS_Factor::~MS_Factor(){
	delete C_SV;
	delete U_SV;
}

void MS_Factor::change_AS(const CS_Parameter *p){
	if(ms!=MS_NAIVE){
		const A_strong *AS(p->a_strong);
		if(C_SV!=NULL)	C_SV->change_AS(AS);
		if(U_SV!=NULL)	U_SV->change_AS(AS);
	}
}


void MS_Factor_test::init(const CS_Parameter *p, const int order_as){
//	NS_Utils::DOUT dout("MS_Factor_test::init");
	delete C_SV;	C_SV=NULL;
	delete U_SV;	U_SV=NULL;

	this->p			= p;
	this->order_a	= order_as;
	this->M			= complex<double>(0.,p->M);
	this->mu_h		= p->mu_h;

//	dout()<<"as = "<<order_a<<endl;
//	dout()<<"M = "<<M<<"\t"<<"mu_h="<<mu_h<<endl;

	const A_strong* A(p->a_strong);

	if( (p->prod==PROD_Higgs)||(p->prod==PROD_H_RS) ){
//		dout()	<<PROD_H<<endl;
		C_SV	= new C_S_exp(order_a,A);
		U_SV	= new U_exp_S(order_a,A);
	}else{
//		dout()	<<"NOT "<<PROD_H<<endl;
		C_SV	= new C_V_exp(order_a,A);
		U_SV	= new U_exp_V(order_a,A);
	}


	results.resize((1+order_a)*8+2);
}

void MS_Factor_test::init(const CS_Parameter *p){
	init(p,p->order_as);
}

Pol_a& MS_Factor_test::ln_C(Pol_a & P, const complex<double> &mu_h)const{
	P.set_equalt_to(C_SV->real_of(M,mu_h),order_a);
	return P;
}

Pol_a& MS_Factor_test::ln_U(Pol_a & P, const complex<double> &mu_h, const complex<double> &mu)const{
	P.set_equalt_to(U_SV->real_of(M,mu_h,mu),order_a);
	return P;
}

const std::vector<double>& MS_Factor_test::update(const double mu){
	Pol_a lnC_SQT(order_a,0.);
	Pol_a lnC_RES(order_a,0.);
	Pol_a lnU_RES(order_a,0.);
	Pol_a lnC_SQT_PI2(order_a,0.);
	Pol_a lnU_SQT_PI2(order_a,0.);
	Pol_a lnC_RES_PI2(order_a,0.);
	Pol_a lnU_RES_PI2(order_a,0.);

	ln_C(lnC_SQT,mu);
	ln_C(lnC_RES,mu_h);
	ln_U(lnU_RES,mu_h,mu);

	ln_C(lnC_SQT_PI2,complex<double>(0.,mu));
	ln_U(lnU_SQT_PI2,complex<double>(0.,mu),mu);
	ln_C(lnC_RES_PI2,complex<double>(0.,mu_h));
	ln_U(lnU_RES_PI2,complex<double>(0.,mu_h),mu);

	lnC_SQT.mul_this_by_scalar(-2.);
	lnC_RES.mul_this_by_scalar(+2.);
	lnU_RES.mul_this_by_scalar(+2.);

	lnC_SQT_PI2.mul_this_by_scalar(-2.);
	lnU_SQT_PI2.mul_this_by_scalar(-2.);
	lnC_RES_PI2.mul_this_by_scalar(+2.);
	lnU_RES_PI2.mul_this_by_scalar(+2.);


	Pol_a lnCe_RES(lnC_RES);
	lnC_RES.add_this_Pol(lnU_RES,order_a);
	lnCe_RES[0] += lnU_RES.of_1(order_a);

	Pol_a lnCe_RES_PI2(lnC_RES_PI2);
	lnC_RES_PI2.add_this_Pol(lnU_RES_PI2,order_a);
	lnCe_RES_PI2[0] += lnU_RES_PI2.of_1(order_a);

	lnC_SQT_PI2.add_this_Pol(lnU_SQT_PI2,order_a);

	Pol_a lnC_SQT_mm(0,0.);
	if(order_a) lnC_SQT_mm.set_equalt_to(lnC_SQT,order_a-1);

	Pol_a lnC_SQT_PI2_mm(0,0.);
	if(order_a) lnC_SQT_PI2_mm.set_equalt_to(lnC_SQT_PI2,order_a-1);

	Pol_a fullE_mm(lnC_RES);
	fullE_mm.add_this_Pol(lnC_SQT_mm,order_a);

	Pol_a fullE_PI2_mm(lnC_RES_PI2);
	fullE_PI2_mm.add_this_Pol(lnC_SQT_PI2_mm,order_a);

	Pol_a fullE(lnC_RES);
	fullE.add_this_Pol(lnC_SQT,order_a);

	Pol_a fullE_PI2(lnC_RES_PI2);
	fullE_PI2.add_this_Pol(lnC_SQT_PI2,order_a);

	Pol_a fullEe_mm(lnCe_RES);
	fullEe_mm.add_this_Pol(lnC_SQT_mm,order_a);

	Pol_a fullEe_PI2_mm(lnCe_RES_PI2);
	fullEe_PI2_mm.add_this_Pol(lnC_SQT_PI2_mm,order_a);

	Pol_a fullEe(lnCe_RES);
	fullEe.add_this_Pol(lnC_SQT,order_a);

	Pol_a fullEe_PI2(lnCe_RES_PI2);
	fullEe_PI2.add_this_Pol(lnC_SQT_PI2,order_a);


	lnC_SQT.exponentiate();
	lnC_SQT_PI2.exponentiate();
	lnC_SQT_mm.exponentiate();
	lnC_SQT_PI2_mm.exponentiate();
	lnC_RES.exponentiate();
	lnC_RES_PI2.exponentiate();
	lnCe_RES.exponentiate();
	lnCe_RES_PI2.exponentiate();
	fullE.exponentiate();
	fullE_PI2.exponentiate();
	fullE_mm.exponentiate();
	fullE_PI2_mm.exponentiate();
	fullEe.exponentiate();
	fullEe_PI2.exponentiate();
	fullEe_mm.exponentiate();
	fullEe_PI2_mm.exponentiate();


	std::vector<double>::iterator it(results.begin());


//	"Hp/P"
	(*(it++))	=	fullE.of_1(order_a);
//	"HPe/P"
	(*(it++))	=	fullEe.of_1(order_a);
//	"HZ/Z"
	(*(it++))	=	(lnC_RES.of_1(order_a))*(lnC_SQT.of_1(order_a));
//	"HZe/Z"
	(*(it++))	=	(lnCe_RES.of_1(order_a))*(lnC_SQT.of_1(order_a));
//	"U"
	(*(it++))	=	exp(lnU_RES.of_1(order_a));

//	"Hp/P_PI2"
	(*(it++))	=	fullE_PI2.of_1(order_a);
//	"HPe/P_PI2"
	(*(it++))	=	fullEe_PI2.of_1(order_a);
//	"HZ/Z_PI2"
	(*(it++))	=	(lnC_RES_PI2.of_1(order_a))*(lnC_SQT_PI2.of_1(order_a));
//	"HZe/Z_PI2"
	(*(it++))	=	(lnCe_RES_PI2.of_1(order_a))*(lnC_SQT_PI2.of_1(order_a));
//	"U_PI2"
	(*(it++))	=	exp(lnU_RES_PI2.of_1(order_a));

	if(order_a){
		//	"Hp/P_mm"
			(*(it++))	=	fullE_mm.of_1(order_a);
		//	"HPe/P_mm"
			(*(it++))	=	fullEe_mm.of_1(order_a);
		//	"HZ/Z_mm"
			(*(it++))	=	(lnC_RES.of_1(order_a))*(lnC_SQT_mm.of_1(order_a));
		//	"HZe/Z_mm"
			(*(it++))	=	(lnCe_RES.of_1(order_a))*(lnC_SQT_mm.of_1(order_a));
//		//	"U_mm"
//			(*(it++))	=	exp(lnU_RES.of_1(order_a-1));

		//	"Hp/P_PI2_mm"
			(*(it++))	=	fullE_PI2_mm.of_1(order_a);
		//	"HPe/P_PI2_mm"
			(*(it++))	=	fullEe_PI2_mm.of_1(order_a);
		//	"HZ/Z_PI2_mm"
			(*(it++))	=	(lnC_RES_PI2.of_1(order_a))*(lnC_SQT_PI2_mm.of_1(order_a));
		//	"HZe/Z_PI2_mm"
			(*(it++))	=	(lnCe_RES_PI2.of_1(order_a))*(lnC_SQT_PI2_mm.of_1(order_a));
//		//	"U_PI2_mm"
//			(*(it++))	=	exp(lnU_RES_PI2.of_1(order_a-1));
	}


	return results;
}

std::ostream& MS_Factor_test::FL(std::ostream &os, const std::string &suffix)const{
	os<<"\t"<<"HP/P_"<<order_a<<order_a<<suffix;
	os<<"\t"<<"HPe/P_"<<order_a<<order_a<<suffix;
	os<<"\t"<<"HZ/Z_"<<order_a<<order_a<<suffix;
	os<<"\t"<<"HZe/Z_"<<order_a<<order_a<<suffix;
	os<<"\t"<<"U_"<<order_a<<suffix;
	os<<"\t"<<"HP/P_PI2_"<<order_a<<order_a<<suffix;
	os<<"\t"<<"HPe/P_PI2_"<<order_a<<order_a<<suffix;
	os<<"\t"<<"HZ/Z_PI2_"<<order_a<<order_a<<suffix;
	os<<"\t"<<"HZe/Z_PI2_"<<order_a<<order_a<<suffix;
	os<<"\t"<<"U_PI2_"<<order_a<<suffix;
	if(order_a){
		os<<"\t"<<"HP/P_"<<order_a<<order_a-1<<suffix;
		os<<"\t"<<"HPe/P_"<<order_a<<order_a-1<<suffix;
		os<<"\t"<<"HZ/Z_"<<order_a<<order_a-1<<suffix;
		os<<"\t"<<"HZe/Z_"<<order_a<<order_a-1<<suffix;
//		os<<"\t"<<"U_"<<order_a-1<<suffix;
		os<<"\t"<<"HP/P_PI2_"<<order_a<<order_a-1<<suffix;
		os<<"\t"<<"HPe/P_PI2_"<<order_a<<order_a-1<<suffix;
		os<<"\t"<<"HZ/Z_PI2_"<<order_a<<order_a-1<<suffix;
		os<<"\t"<<"HZe/Z_PI2_"<<order_a<<order_a-1<<suffix;
//		os<<"\t"<<"U_PI2_"<<order_a-1<<suffix;
	}
	return os;
}

MS_Factor_test::~MS_Factor_test(){
	delete C_SV;	C_SV=NULL;
	delete U_SV;	U_SV=NULL;
}


}  // namespace NS_Hard
}  // namespace CuTe


