#ifndef RES_QDK_H
#define RES_QDK_H

#include "polynom.h"
#include "Res_Parameter.h"

namespace CuTe {

namespace NS_CS_Params {
	class CS_Parameter;
	class CS_Parameter_LL;
	class CS_Parameter_LLO;
}  // namespace NS_CS_Params

namespace NS_Resumed {

//using NS_Polynom::Continious_Order_Reduction;

using namespace NS_enums::NS_PROD;
using namespace NS_enums::NS_RES;
using namespace NS_enums::NS_NP;

using NS_CS_Params::CS_Parameter;
using NS_CS_Params::CS_Parameter_LL;
using NS_CS_Params::CS_Parameter_LLO;

typedef NS_Polynom::Pol<double> Pol_ae;

class G_exp_Pol;
class IQI;
class _K_Integral_;
class K_LL;
class K_LLO;


class _Quad_Differential_Kernel:public Pol_ae {
protected:
	const RES res;

	const double *qT;

	G_exp_Pol		*gexp;
	IQI				*iqi;
	_K_Integral_	*K;

//	const Continious_Order_Reduction *reduction;

	void _init(const CS_Parameter *p, const Res_Kinematic_Parameters *RKP);

public:

	void set_qT_mu(const Res_Kinematic_Parameters *RKP);//, const Continious_Order_Reduction *reduction=NULL);

	void set_y(const Res_Kinematic_Parameters *RKP);

	Pol_ae& update_zz(const Res_Kinematic_Parameters *RKP);

private:
	_Quad_Differential_Kernel();
protected:

	_Quad_Differential_Kernel(const RES res);
	_Quad_Differential_Kernel(const _Quad_Differential_Kernel& rv);
	_Quad_Differential_Kernel& operator=(const _Quad_Differential_Kernel& rv);

	virtual ~_Quad_Differential_Kernel();

public:

	void print_pdf(const std::string s="")const;
	void change_PDF(const CS_Parameter *p);
};

class Quad_Differential_Kernel_LLO:public _Quad_Differential_Kernel {
protected:
	const double *Lmu;
	const double *etaB;

	void _del_K();
	void _init_K();

public:

	void init(const CS_Parameter_LLO *p, const Res_Kinematic_Parameters *RKP);

	Quad_Differential_Kernel_LLO()
	:_Quad_Differential_Kernel(RES_LLO)
	,Lmu(NULL),etaB(NULL){
//		NS_Utils::DOUT dout("Quad_Differential_Kernel");
	}

	Quad_Differential_Kernel_LLO(const Quad_Differential_Kernel_LLO& rv)
	:_Quad_Differential_Kernel(rv)
	,Lmu(rv.Lmu),etaB(rv.etaB){
//		NS_Utils::DOUT dout("Quad_Differential_Kernel","rv");
		if(rv.K!=NULL)	_init_K();
	}

	Quad_Differential_Kernel_LLO& operator=(const Quad_Differential_Kernel_LLO& rv){
//		NS_Utils::DOUT dout("Quad_Differential_Kernel::operator=","rv");
		_Quad_Differential_Kernel::operator=(rv);
		Lmu	= (rv.Lmu);
		etaB= (rv.etaB);

		if(rv.K!=NULL)	_init_K();
		else _del_K();

		return *this;
	}

	~Quad_Differential_Kernel_LLO(){
		_del_K();
//		NS_Utils::DOUT dout("~Quad_Differential_Kernel");
	}
};

class Quad_Differential_Kernel_LL:public _Quad_Differential_Kernel {
protected:
	const double *mu;
	const double *LM;
//	const double *q_star;
	const CUTOFF *cutoff;
	const double *Lambda;

	void _init_K();
	void _del_K();

public:

	void init(const CS_Parameter_LL *p, const Res_Kinematic_Parameters *RKP);

	Quad_Differential_Kernel_LL()
	:_Quad_Differential_Kernel(RES_LL)
	,mu(NULL),LM(NULL)//,q_star(NULL)
	,cutoff(NULL),Lambda(NULL){
//		NS_Utils::DOUT dout("Quad_Differential_Kernel");
	}

	Quad_Differential_Kernel_LL(const Quad_Differential_Kernel_LL& rv)
	:_Quad_Differential_Kernel(rv)
	,mu(rv.mu),LM(rv.LM)//,q_star(rv.q_star)
	,cutoff(rv.cutoff),Lambda(rv.Lambda){
//		NS_Utils::DOUT dout("Quad_Differential_Kernel","rv");
		if(rv.K!=NULL)	_init_K();
	}

	Quad_Differential_Kernel_LL& operator=(const Quad_Differential_Kernel_LL& rv){
//		NS_Utils::DOUT dout("Quad_Differential_Kernel::operator=","rv");
		_Quad_Differential_Kernel::operator=(rv);
		mu		= (rv.mu);
		LM		= (rv.LM);
//		q_star	= (rv.q_star);
		cutoff	= (rv.cutoff);
		Lambda	= (rv.Lambda);

		if(rv.K!=NULL)	_init_K();
		else _del_K();

		return *this;
	}

	~Quad_Differential_Kernel_LL(){
		_del_K();
//		NS_Utils::DOUT dout("~Quad_Differential_Kernel");
	}

	void change_NPE();
};



}  // namespace NS_RESUMED



}  // namespace CuTe


#endif
