#ifndef RES_EXPONENTS_H
#define RES_EXPONENTS_H

#include "polynom.h"
#include "Anomalous_Dim.h"


namespace CuTe {
namespace NS_Resumed {

using namespace NS_Utils::NS_pow;

using NS_Polynom::Pol;
//using NS_Polynom::Continious_Order_Reduction;

//using namespace NS_enums::NS_RES;
using namespace NS_enums::NS_PROD;
using namespace NS_enums::NS_PC;
using namespace NS_Utils::NS_min_max;

using NS_Anomalous_Dimension::AD_Beta;
using NS_Anomalous_Dimension::AD_Casimir_d;
using NS_Anomalous_Dimension::AD_GammaCusp;
using NS_Anomalous_Dimension::AD_gamma_C_over_C;

using std::cerr;
using std::complex;
using std::endl;
using std::string;
using std::stringstream;


class _exp_pol_:public Pol< Pol<double> >{
protected:
	const int _N_,_M_;

//	RES res;
	PROD prod;
	int ord_a_or_e;
//	bool improved;
	POWER_COUNTING pc;

	bool exponent;
	const double *as;

	const int *nf;	// real nf
	double C_B;	// C_A or C_F
	int n; 		//active nf

	inline void _init(int ord_a_or_e, POWER_COUNTING pc, bool exponent, const double *as, const int *nf, PROD prod){
//		NS_Utils::DOUT	dout("_exp_pol_::_init");
//		this->res			= res;
		this->prod			= prod;
		this->ord_a_or_e	= ord_a_or_e;
//		this->improved		= improved;
		this->pc			= pc;
		this->exponent		= exponent;
		this->as			= as;
		this->nf			= nf;


		switch (prod) {
			case PROD_Higgs:
			case PROD_H_RS:
				C_B	= C_A;
				break;
			case PROD_Z:
			case PROD_W:
			case PROD_DY:
				C_B	= C_F;
				break;
			default:
				cerr	<<"**********************************************"<<endl
						<<"**   virtual _exp_pol_::_init() Wrong prod  **"<<endl
						<<"**********************************************"<<endl;
				exit(1);
				break;
		}

//		dout()<<"prod="<<prod<<"\t"<<"C_B="<<C_B<<endl;
//		dout()<<"pc="<<pc<<"\t"<<"ord="<<ord_a_or_e<<endl;
//		dout()<<"exp="<<exponent<<endl;

		n	= -1;
	}

public:

	inline int get_order()const{ return ord_a_or_e;}
//	inline bool get_improved()const{ return improved;}
	inline const bool improved_pc()const{ return (pc==PC_IMP);}
	inline const POWER_COUNTING get_power_counting()const{ return pc;}
	inline bool get_exponent()const{ return exponent;}

	inline virtual void _set_mu(){//const Continious_Order_Reduction* reduction=NULL){
		cerr	<<"********************************************"<<endl
				<<"**   virtual _exp_pol_::_set_mu() called  **"<<endl
				<<"********************************************"<<endl;
		exit(1);
	}


	inline virtual void _set_mu(Pol< complex<double> > &Kexp, const complex<double> kexpfactor=1.){//,const Continious_Order_Reduction* reduction=NULL){
		cerr	<<"************************************************"<<endl
				<<"**   virtual _exp_pol_::_set_mu(Kexp) called  **"<<endl
				<<"************************************************"<<endl;
		exit(1);
	}


//	inline Pol< Pol<double> >& update_mu_a_l(const double as){
////		NS_Utils::DOUT dout("_exp_pol_::update_mu_a_l","as");
//		_set_mu();
////		dout()<<"*this:"<<endl<<*this<<endl;
//
//
//		set_of(as);
//		if( exponent ){
//
//			if(improved_pc()) improve_Pol_a_l(*this,ord_a_or_e);
//
//			for(int i=ord_a_or_e; i>0; --i) _P[0].add_this_Pol(_P[i]);
//			order(0);
//		}
//
//		return *this;
//	}

//	inline Pol< Pol<double> >& update_mu(){
////		NS_Utils::DOUT dout("_exp_pol_::update_mu");
//		update_mu_a_l(*as);
//
//		if( improved_pc() && (!exponent) ) improve_Pol_a_l(*this,ord_a_or_e);
//
//		return *this;
//	}
//
//	inline Pol< Pol<double> >& update_mu_exp(){
////		NS_Utils::DOUT dout("_exp_pol_::update_mu_exp");
////		dout()	<<"as="<<*as<<endl;
//
//		update_mu();
//
//		order(ord_a_or_e);	//order for exp()
//		exp();
//		reduce_order();
//
//		return *this;
//	}

	inline Pol< Pol<double> >& update_mu_a_l(const double as){//, const Continious_Order_Reduction* reduction=NULL){
//		NS_Utils::DOUT dout("_exp_pol_::update_mu_a_l","as,Kexp");
		_set_mu();//reduction);
//		dout()<<"*this:"<<endl;
//		dout()	<<*this<<endl;
		set_of(as);

		return *this;
	}

	inline Pol< Pol<double> >& update_mu(){//const Continious_Order_Reduction* reduction=NULL){
//		NS_Utils::DOUT dout("_exp_pol_::update_mu","Kexp");
		update_mu_a_l(*as);//,reduction);
//		dout()<<"*this:"<<endl;
//		dout()	<<*this<<endl;

		if(improved_pc())	improve_Pol_a_l(*this,ord_a_or_e);
		if(exponent){
			// still expand the highest order term, if its order in L_T is uneaven
			for(int i=_order; i>=0; --i){
//				dout()	<<"add(i)="<<i<<endl;
				_P[0].add_this_Pol(_P[i]);
				_P[i].order(0);
				_P[i][0]=0.;
			}
			reduce_order();
		}


//		dout()	<<"Kexp:"<<Kexp<<endl;
//		dout()	<<"this:"<<*this<<endl;

		return *this;
	}

	inline Pol< Pol<double> >& update_mu_exp(){//const Continious_Order_Reduction* reduction=NULL){
//		NS_Utils::DOUT dout("_exp_pol_::update_mu_exp","Kexp");
		update_mu();//reduction);
//		dout()	<<"*this:"<<endl;
//		dout()	<<*this<<endl;

		exponentiate();
		reduce_order();
//		dout()	<<"*this_exp:"<<endl;
//		dout()	<<*this<<endl;

		return *this;
	}

	inline Pol< Pol<double> >& update_mu_a_l(const double as,Pol< complex<double> > &Kexp){//, const Continious_Order_Reduction* reduction=NULL){
//		NS_Utils::DOUT dout("_exp_pol_::update_mu_a_l","as,Kexp");
		Kexp.order(2);
		Kexp.mul_this_by_scalar(0.);

		_set_mu(Kexp,1.);//,reduction);
//		dout()<<"*this:"<<endl;
//		dout()	<<*this<<endl;
		set_of(as);

		return *this;
	}

	inline Pol< Pol<double> >& update_mu(Pol< complex<double> > &Kexp){//,const Continious_Order_Reduction* reduction=NULL){
//		NS_Utils::DOUT dout("_exp_pol_::update_mu","Kexp");
		update_mu_a_l(*as,Kexp);//,reduction);
//		dout()	<<"Kexp:"<<Kexp<<endl;
//		dout()	<<"this:"<<*this<<endl;

		if(improved_pc()){
			improve_Pol_a_l(*this,ord_a_or_e);

			if(exponent){
				int i=ord_a_or_e;
				if((i>=1)&&(_P[i].order()%2)) --i;
				for(; i>=0; --i){
	//				dout()	<<"add(i)="<<i<<endl;
					Kexp.add_this_Pol(_P[i]);
					_P[i].order(0);
					_P[i][0]=0.;
				}
				reduce_order();
			}
		}else if(exponent){
//			dout()	<<"EXPONENT"<<endl;
//			dout()	<<"\t"<<"ord_a_or_e="<<ord_a_or_e<<"\t"<<"_order="<<_order<<endl;
			// still expand the highest order term, if its order in L_T is uneaven
			int i=ord_a_or_e;

//			// If the last term of P[ord] is uneaven in L_T add only all remaining terms to Kexp
//			int l=_P[i].order();
//			if( (l%2)&&(l>Kexp.order())){
//				Kexp.add_this_Pol(_P[i],--l);
//				for(;l>=0;--l) _P[i][l] = 0.;
//				--i;
//			}

			// If the last term of P[ord] is uneaven do not add it to Kexp
			if((i>=2)&&(_P[i].order()%2)) --i;

			for(; i>=0; --i){
//				dout()	<<"add(i)="<<i<<endl;
				Kexp.add_this_Pol(_P[i]);
				_P[i].order(0);
				_P[i][0]=0.;
			}
			reduce_order();
		}else{
			order(ord_a_or_e);	//order for exp()
		}

		Kexp.reduce_order();


//		dout()	<<"Kexp:"<<Kexp<<endl;
//		dout()	<<"this:"<<*this<<endl;

		return *this;
	}

	inline Pol< Pol<double> >& update_mu_exp(Pol< complex<double> > &Kexp){//, const Continious_Order_Reduction* reduction=NULL){
//		NS_Utils::DOUT dout("_exp_pol_::update_mu_exp","Kexp");
		update_mu(Kexp);//,reduction);
//		dout()	<<"*this:"<<endl;
//		dout()	<<*this<<endl;

		exponentiate();
		reduce_order();
//		dout()	<<"E:"<<*this<<endl;
//		dout()	<<"k:"<<Kexp<<endl;

		return *this;
	}

private:
	_exp_pol_()
	:Pol< Pol<double> >()
	,_N_(5),_M_(5)
	,prod(),ord_a_or_e(),pc(),exponent()
	,as(NULL),nf(NULL),C_B()
	,n(-1){

	}

public:
	_exp_pol_(int _N_, int _M_)
	:Pol< Pol<double> >(_N_,Pol<double>(_M_,0.))
	,_N_(_N_),_M_(_M_)
	,prod(),ord_a_or_e(),pc(),exponent()
	,as(NULL),nf(NULL),C_B()
	,n(-1){

	}

	_exp_pol_(const _exp_pol_ &epol)
	:Pol< Pol<double> >(epol._N_,Pol<double>(epol._M_,0.))
	,_N_(epol._N_),_M_(epol._M_)
	,prod(epol.prod),ord_a_or_e(epol.ord_a_or_e),pc(epol.pc),exponent(epol.exponent)
	,as(epol.as),nf(epol.nf),C_B(epol.C_B)
	,n(-1){
//		NS_Utils::DOUT dout("_exp_pol_","epol");
		set_equalt_to(epol);
	}

	inline _exp_pol_& operator=(const _exp_pol_ &epol){
//		NS_Utils::DOUT dout("_exp_pol_::operator=","epol");
		if(_N_!=epol._N_){
			std::cerr	<<"_exp_pol_::operator=(epol) _N_ = "<<_N_<<" != epol._N_ = "<<epol._N_<<endl;
			exit(1);
		}
		if(_M_!=epol._M_){
			std::cerr	<<"_exp_pol_::operator=(epol) _M_ = "<<_M_<<" != epol._M_ = "<<epol._M_<<endl;
			exit(1);
		}

		prod		= (epol.prod);
		ord_a_or_e	= (epol.ord_a_or_e);
		pc			= (epol.pc);
		exponent	= (epol.exponent);
		as			= (epol.as);
		nf			= (epol.nf);
		C_B			= (epol.C_B);

		set_equalt_to(epol);
		n			= (-1);

		return *this;
	}

	virtual ~_exp_pol_(){

	}

	inline virtual const string properties(){
		stringstream ss;
		ss	<<"_exp_pol_("<<_N_<<","<<_M_<<"):"	<<endl
			<<"\t"<<"prod="<<prod	<<"\t"<<"ord="<<ord_a_or_e	<<"\t"<<"PC="<<pc	<<"\t"<<"exp="<<exponent	<<endl
			<<"\t"<<"as="<<*as	<<"\t"<<"nf="<<*nf	<<"\t"<<"n="<<n	<<"\t"<<"C_B="<<C_B	<<endl;

		return ss.str();
	}

};


/*! \class H_exp_Pol
    \brief Calculate the exponent h: B = exp(h)*Bbar

    Calculate the exponent h which is taken out of the TDPDFs B to generate the RGE-invariant versions Bbar
    B = exp(h)*Bbar

*/
class H_exp_Pol:public _exp_pol_{
protected:
	Pol< Pol<double> > _h_a_l;

	AD_Beta Beta;
	AD_GammaCusp Gamma;
	AD_gamma_C_over_C gamma_C;

	inline void _AD_select(int n){
//		NS_Utils::DOUT dout("H_exp_Pol::_AD_select","n");

		Beta.select(n);
		Gamma.select(n);
		gamma_C.select(n);

//		dout()	<<"C_B="<<C_B<<"\t"<<"g_C(0)="<<gamma_C(0)<<"\t"<<"g_C(1)="<<gamma_C(1)<<"\t"<<"g_C(2)="<<gamma_C(2)<<endl;
	}

public:

	inline void _set_mu(){//const Continious_Order_Reduction* reduction=NULL){
//		NS_Utils::DOUT dout("hpol::_set_mu");
//		dout()	<<"as="<<*as	<<"\t"<<"nf="<<*nf<<endl;
		if(n!=(*nf)){
//			dout()	<<"n="<<n	<<"\t"<<"n_f="<<*nf	<<"\t"<<"&n_f="<<nf	<<endl;
			n	= (*nf);
			_AD_select(n);

			// set h[a][l]/(C_B)

			switch (_h_a_l.order()) {
				case 6:
					_h_a_l[6].order(7);		//a_s^6
					_h_a_l[6][0]	=	0;	//epsilon^12
					_h_a_l[6][1]	=	0;	//epsilon^11
					_h_a_l[6][2]	=	0;	//epsilon^10
					_h_a_l[6][3]	=	0;	//epsilon^9
					_h_a_l[6][4]	=	0;	//epsilon^8
					_h_a_l[6][5]	=	0;	//epsilon^7
					_h_a_l[6][6]	=	0;	//epsilon^6
					_h_a_l[6][7]	=	(pow_5(Beta(0))*Gamma(0))/84.;//epsilon^5
					/* no break */
				case 5:
					_h_a_l[5].order(6);		//a_s^5
					_h_a_l[5][0]	=	0;	//epsilon^10
					_h_a_l[5][1]	=	0;	//epsilon^9
					_h_a_l[5][2]	=	0;	//epsilon^8
					_h_a_l[5][3]	=	0;	//epsilon^7
					_h_a_l[5][4]	=	0;	//epsilon^6
					_h_a_l[5][5]	=	-(pow_3(Beta(0))*(-13*Beta(1,0)*Gamma(0) - 12*Gamma(1) + 24*Beta(0)*gamma_C(0)))/120.;	//epsilon^5
					_h_a_l[5][6]	=	(pow_4(Beta(0))*Gamma(0))/60.;	//epsilon^4
					/* no break */
				case 4:
					_h_a_l[4].order(5);		//a_s^4
					_h_a_l[4][0]	=	0.;	//epsilon^8
					_h_a_l[4][1]	=	0.;	//epsilon^7
					_h_a_l[4][2]	=	0.;	//epsilon^6
					_h_a_l[4][3]	=	-(Beta(0)*(-(Beta(2,0)*Gamma(0)) - 2*Beta(1,0)*Gamma(1) - 3*Gamma(2) + 10*Beta(1)*gamma_C(0) + 12*Beta(0)*gamma_C(1)))/12.;	//epsilon^5
					_h_a_l[4][4]	=	+pow_2(Beta(0))*(5.*Beta(1,0)*Gamma(0) + 6.*Gamma(1) - 12.*Beta(0)*(gamma_C(0)))/48.;	//epsilon^4
					_h_a_l[4][5]	=	+pow_3(Beta(0))*Gamma(0)/40.;	//epsilon^3
					/* no break */
				case 3:
					_h_a_l[3].order(4);		//a_s^3
					_h_a_l[3][0]	=	0.;	//epsilon^6
					_h_a_l[3][1]	=	-gamma_C(2);	//epsilon^5
					_h_a_l[3][2]	=	+0.25*( Gamma(2) - 2.*Beta(1)*gamma_C(0) - 4.*Beta(0)*gamma_C(1));	//epsilon^4
					_h_a_l[3][3]	=	+Beta(0)*(Beta(1,0)*Gamma(0) + 2.*Gamma(1) - 4.*Beta(0)*gamma_C(0))/12.;	//epsilon^3
					_h_a_l[3][4]	=	+pow_2(Beta(0))*Gamma(0)/24.;	//epsilon^2
					/* no break */
				case 2:
					_h_a_l[2].order(3);		//a_s^2
					_h_a_l[2][0]	=	0.;	//epsilon^4
					_h_a_l[2][1]	=	-(gamma_C(1));	//epsilon^3
					_h_a_l[2][2]	=	+0.25*( Gamma(1) - 2.*Beta(0)*(gamma_C(0)));	//epsilon^2
					_h_a_l[2][3]	=	+Beta(0)*Gamma(0)/12.;	//epsilon^1
					/* no break */
				case 1:
					_h_a_l[1].order(2);		//a_s^1
					_h_a_l[1][0]	=	0.;	//epsilon^2
					_h_a_l[1][1]	=	-gamma_C(0);	//epsilon^1
					_h_a_l[1][2]	=	+0.25*Gamma(0);		//epsilon^0	_h_a_l[1][2]->k_exp[2]
					/* no break */
				case 0:
					break;
				default:
					std::cerr	<<" WRONG order("<<_order<<") in hpol::_set_nf"<< endl;
					break;
			}

			_h_a_l.mul_this_by_scalar(C_B);
//			dout()	<<"h:"<<_h_a_l<<endl;
		}

		switch (pc) {
		case PC_NO_RES:
			set_equalt_to(_h_a_l,ord_a_or_e);
			break;
		case PC_NAIVE:
			set_equalt_to(_h_a_l,ord_a_or_e);
			break;
		case PC_STD:
			set_equalt_to(_h_a_l,max_of(1,ord_a_or_e));
			break;
		case PC_IMP:
			set_equalt_to(_h_a_l,ord_a_or_e+1);
			break;
		default:
			break;
		}
//		if(reduction!=NULL) reduction->reduce_this_Pol_a_l(*this);
//		dout()	<<"this:"<<*this<<endl;
	}

	inline void _set_mu(Pol< complex<double> > &Kexp, const complex<double> kexpfactor=1.){//, const Continious_Order_Reduction* reduction=NULL){
//		NS_Utils::DOUT dout("hpol::_set_mu","Kexp");
		_set_mu();//reduction);

//		dout()	<<"this:"<<endl;
//		dout()	<<*this<<endl;

		if( (pc==PC_STD) || (pc==PC_IMP) ){
			// always add Leading epsilon terms
			Kexp[2]	+= kexpfactor*(*as)*_P[1][2];

			// remove them from h
			_P[1][2] = 0.;
			_P[1].reduce_order();
			if(_order==1) reduce_order();
		}

//		dout()	<<"Kexp("<<kexpfactor<<"):"<<Kexp<<endl;
//		dout()	<<"this_after_Kexp:"<<endl;
//		dout()	<<*this<<endl;
	}

	inline void init(
			int order, POWER_COUNTING pc, bool exponent
			, const int *nf, const double *as
			, PROD prod
			){
//		NS_Utils::DOUT dout("H_exp_Pol::init");
		_init(order,pc,exponent,as,nf,prod);	//sets n=-1

		switch (pc) {
			case PC_NO_RES:
				_h_a_l.order(order);
				break;
			case PC_NAIVE:
				_h_a_l.order(order);
				break;
			case PC_STD:
				_h_a_l.order(max_of(order,1));
				break;
			case PC_IMP:
				_h_a_l.order(order+1);	//max(e) = a+1
				break;
			default:
				break;
		}

		_h_a_l[0].order(0);		//a_s^0
		_h_a_l[0][0]	=	0.;	//epsilon^0

		Gamma.init(prod);
		gamma_C.init(prod);
	}

	H_exp_Pol()
	:_exp_pol_(6,7)
	,_h_a_l(Pol< Pol<double> >(6,Pol<double>(7,0.)))
	,Beta(),Gamma(PROD_WRONG_STATEMENT)
	,gamma_C(PROD_WRONG_STATEMENT){

	}

	H_exp_Pol(
			int order, POWER_COUNTING pc, bool exponent
			, const int *nf, const double *as
			, PROD prod
			)
	:_exp_pol_(6,7)
	,_h_a_l(Pol< Pol<double> >(6,Pol<double>(7,0.)))
	,Beta(),Gamma(prod)
	,gamma_C(prod){
		init(order,pc,exponent,nf,as,prod);
	}

	H_exp_Pol(const H_exp_Pol &h)
	:_exp_pol_(6,7)
	,_h_a_l(Pol< Pol<double> >(6,Pol<double>(7,0.)))
	,Beta(),Gamma(h.prod)
	,gamma_C(h.gamma_C){
//		NS_Utils::DOUT dout("H_exp_Pol","h");

		_exp_pol_::operator=(h);	//sets n=-1
		_h_a_l		= (h._h_a_l);
	}

	inline H_exp_Pol& operator=(const H_exp_Pol &h){
		_exp_pol_::operator=(h);	//sets n=-1

		_h_a_l		= (h._h_a_l);
		Gamma		= (h.Gamma);
		gamma_C		= (h.gamma_C);

		return *this;
	}

	~H_exp_Pol(){
//		delete gamma_C;	// No = NULL
	}

	inline virtual const string properties(){
		stringstream ss;
		ss	<<"H_exp_Pol:"<<_exp_pol_::properties();
//		ss	<<"\t_h_a_l = "<<_h_a_l<<endl;

		return ss.str();
	}

};

/*! \class Anomaly_exp_Pol
    \brief Calculate the Anomaly Exponent = -(LT+LM)*F_{i,j}

    Calculate the exponent which is generated by the collinear anomaly
    given by: -(LT+LM)*F_{i,j}

*/
class Anomaly_exp_Pol:public _exp_pol_{
protected:
	Pol< Pol<double> > _F_a_l;
	Pol< Pol<double> > _LM_F_a_l;
	Pol< Pol<double> > _LT_F_a_l;


	AD_Beta Beta;
	AD_GammaCusp Gamma;
	AD_Casimir_d d;

	const double* LM;//= log(M^2/mu^2)

	inline void _AD_select(int n){
//		NS_Utils::DOUT dout("Anomaly_exp_Pol::_AD_select","n");
//		dout()	<<"C_B="<<C_B<<endl;
		Beta.select(n);
		Gamma.select(n);
		d.select(n);
	}

public:

	inline void _set_mu(){//const Continious_Order_Reduction* reduction=NULL){
//		NS_Utils::DOUT dout("Anomaly_exp_Pol::_set_mu");
//		dout()	<<"as="<<*as	<<"\t"<<"nf="<<*nf<<	"\t"<<"LM="<<*LM	<<endl;
		if(n!=(*nf)){
//	//		NS_Utils::DOUT dout("n != nf");
//			dout()	<<"n="<<n	<<"\t"<<"nf="<<*nf	<<endl;
			n	= (*nf);
			_AD_select(n);

			// set F /( C_B * Gamma_0 )

			//	Order ( a , b , c )	-> b =  Order(a)*L_T	c = Order(a)* eta/a_s

			switch (_F_a_l.order()) {
				case 7:
					_F_a_l[7].order(7);		//a_s^(7,7,6)
					_F_a_l[7][0]	=	0;	//epsilon^(14,13,12) > 5
					_F_a_l[7][1]	=	0;	//epsilon^(13,12,11) > 5
					_F_a_l[7][2]	=	0;	//epsilon^(12,11,10) > 5
					_F_a_l[7][3]	=	0;	//epsilon^(11,10,09) > 5
					_F_a_l[7][4]	=	0;	//epsilon^(10,09,08) > 5
					_F_a_l[7][5]	=	0;	//epsilon^(09,08,07) > 5
					_F_a_l[7][6]	=	0;	//epsilon^(08,07,06) > 5
					_F_a_l[7][7]	=	pow_6(Beta(0))/7.;//epsilon^(07,06,05)
					/* no break */
				case 6:
					_F_a_l[6].order(6);		//a_s^(6,6,5)
					_F_a_l[6][0]	=	0;	//epsilon^(12,11,10) > 5
					_F_a_l[6][1]	=	0;	//epsilon^(11,10,09) > 5
					_F_a_l[6][2]	=	0;	//epsilon^(10,09,08) > 5
					_F_a_l[6][3]	=	0;	//epsilon^(09,08,07) > 5
					_F_a_l[6][4]	=	0;	//epsilon^(08,07,06) > 5
					_F_a_l[6][5]	=	pow_4(Beta(0))*(77*Beta(1,0) + 60*Gamma(1,0))/60.;//epsilon^(07,06,05)
					_F_a_l[6][6]	=	pow_5(Beta(0))/6.;//epsilon^(06,05,04)
					/* no break */
				case 5:
					_F_a_l[5].order(5);		//a_s^(5,5,4)
					_F_a_l[5][0]	=	0;	//epsilon^(10,09,08) > 5
					_F_a_l[5][1]	=	0;	//epsilon^(09,08,07) > 5
					_F_a_l[5][2]	=	0;	//epsilon^(08,07,06) > 5
					_F_a_l[5][3]	=	pow_2(Beta(0))*(3*pow_2(Beta(1,0)) + 24*d(2)*Beta(0) + 6*Beta(2,0) + 14*Beta(1,0)*Gamma(1,0) + 12*Gamma(2,0))/6.;//epsilon^(07,06,05)
					_F_a_l[5][4]	=	pow_3(Beta(0))*(13*Beta(1,0) + 12*Gamma(1,0))/12.;//epsilon^(06,05,04)
					_F_a_l[5][5]	=	pow_4(Beta(0))/5.;//epsilon^(05,04,03)
					/* no break */
				case 4:
					_F_a_l[4].order(4);		//a_s^(4,4,3)
					_F_a_l[4][0]	=	0.;	//epsilon^(08,07,06) > 5
					_F_a_l[4][1]	=	3*d(3)*Beta(0) + 2*d(2)*Beta(1) + Gamma(3,0);//epsilon^(07,06,05)
					_F_a_l[4][2]	=	Beta(0)*(6*d(2)*Beta(0) + Beta(2,0) + 2*Beta(1,0)*Gamma(1,0) + 3*Gamma(2,0))/2.;//epsilon^(06,05,04)
					_F_a_l[4][3]	=	pow_2(Beta(0))*(5*Beta(1,0) + 6*Gamma(1,0))/6.;//epsilon^(05,04,03)
					_F_a_l[4][4]	=	pow_3(Beta(0))/4.;//epsilon^(04,03,02)
					/* no break */
				case 3:
					_F_a_l[3].order(3);		//a_s^(3,3,2)
					_F_a_l[3][0]	=	d(3);//epsilon^(06,05,04)
					_F_a_l[3][1]	=	2*d(2)*Beta(0) + Gamma(2,0);//epsilon^(05,04,03)
					_F_a_l[3][2]	=	Beta(0)*(Beta(1,0) + 2*Gamma(1,0))/2.;//epsilon^(04,03,02)
					_F_a_l[3][3]	=	pow_2(Beta(0))/3.;//epsilon^(03,02,01)
					/* no break */
				case 2:
					_F_a_l[2].order(2);				//a_s^(2,2,1)
					_F_a_l[2][0]	=	d(2);		//epsilon^(04,03,02)
					_F_a_l[2][1]	=	Gamma(1,0);	//epsilon^(03,02,01)
					_F_a_l[2][2]	=	Beta(0)/2.;	//epsilon^(02,01,00)	//->Kexp
					/* no break */
				case 1:
					_F_a_l[1].order(1);		//a_s^(1,1,0)
					_F_a_l[1][0]	=	0.;	//epsilon^(02,01,00)
					_F_a_l[1][1]	=	1.;	//epsilon^(01,00,-1)	//->Kexp
					/* no break */
					break;
				default:
					std::cerr	<<" WRONG order("<<_order<<") in Anomaly_exp_Pol::_set_nf"<< endl;
					break;
			}
			_F_a_l.mul_this_by_scalar(-C_B*Gamma(0));

//			dout()	<<"_F_a_l:"<<_F_a_l<<endl;
//
//			dout()	<<"ord_a_or_e="<<ord_a_or_e
//					<<"\t"<<"Ord(F)="<<_F_a_l.order()
//					<<"\t"<<"Ord(LM_F)="<<_LM_F_a_l.order()
//					<<"\t"<<"Ord(LT_F)="<<_LT_F_a_l.order()<<endl;

			switch (pc) {
			case PC_NO_RES:{
				const int o(max_of(1,ord_a_or_e));

				_LM_F_a_l.set_equalt_to(_F_a_l,o);

				_LT_F_a_l.set_equalt_to(_F_a_l,ord_a_or_e);
				for(int i=1; i<=ord_a_or_e; ++i) _LT_F_a_l[i].timesX(); //additional power of LT
			}
			break;
			case PC_NAIVE:{

				_LM_F_a_l.order(ord_a_or_e);
				for(int i=0; i<=ord_a_or_e; ++i){
					_LM_F_a_l[i].set_equalt_to(_F_a_l[i+1]);	//multiplie later with additional as
				}

//				dout()	<<"_LM_F_a_l:";
//				dout()	<<_LM_F_a_l<<endl;

				_LT_F_a_l.set_equalt_to(_F_a_l,ord_a_or_e);
				for(int i=1; i<=ord_a_or_e; ++i) _LT_F_a_l[i].timesX(); //additional power of LT
			}
			break;
			case PC_STD:{
				const int o(max_of(1,ord_a_or_e));
//		//		NS_Utils::DOUT dout("improved==false");

				_LM_F_a_l.order(o);
				for(int i=0; i<=o; ++i){
					_LM_F_a_l[i].set_equalt_to(_F_a_l[i+1]);	//multiplie later with additional as
				}

//				dout()	<<"_LM_F_a_l:";
//				dout()	<<_LM_F_a_l<<endl;

				_LT_F_a_l.set_equalt_to(_F_a_l,o);
				for(int i=1; i<=o; ++i) _LT_F_a_l[i].timesX(); //additional power of LT
			}
			break;
			case PC_IMP:{
				const int o(ord_a_or_e+1);

				_LM_F_a_l.order(o);
				for(int i=0; i<=o; ++i){
					_LM_F_a_l[i].set_equalt_to(_F_a_l[i+1]);	//multiplie later with additional as
				}

				_LT_F_a_l.set_equalt_to(_F_a_l,o);
				for(int i=1; i<=o; ++i) _LT_F_a_l[i].timesX(); //additional power of LT
			}
			break;
			default:
			break;
			}

//			dout()	<<"LM*F:"<<_LM_F_a_l<<endl;
//			dout()	<<"LT*F:"<<_LT_F_a_l<<endl;
		}

//		dout()	<<"as="<<*as
//				<<"\t"<<"LM="<<*LM
//				<<"\t"<<"Ord(this)="<<order()
//				<<"\t"<<"Ord(LM_F)="<<_LM_F_a_l.order()
//				<<"\t"<<"Ord(LT_F)="<<_LT_F_a_l.order()<<endl;

		double mul;
		int o1;
		int o2;

		switch (pc) {
		case PC_NO_RES:
			mul	= (*LM);
			o1	= max_of(1,ord_a_or_e);
			o2	= ord_a_or_e;
			break;
		case PC_NAIVE:
			mul	= (*as)*(*LM);
			o1	= ord_a_or_e;
			o2	= ord_a_or_e;
			break;
		case PC_STD:
			mul	= (*as)*(*LM);
			o1	= max_of(1,ord_a_or_e);
			o2	= max_of(1,ord_a_or_e);
			break;
		case PC_IMP:
			mul	= (*as)*(*LM);
			o1	= ord_a_or_e+1;
			o2	= ord_a_or_e+1;
			break;
		default:
			break;
		}
		set_equalt_to(_LM_F_a_l,o1);
		mul_this_by_scalar(mul);
		add_this_Pol(_LT_F_a_l,o2);
//		if( (reduction!=NULL)&&(pc!=PC_NO_RES) ){
//			reduction->reduce_this_Pol_aEta_l(*this);
//			add_this_Pol(reduction->reduce_Pol_a_l(_LT_F_a_l),o2);
//		}else{
//			add_this_Pol(_LT_F_a_l,o2);
//		}

//		dout()	<<"this:"<<*this<<endl;
	}

	inline void _set_mu(Pol< complex<double> > &Kexp, const complex<double> kexpfactor=1){//,const Continious_Order_Reduction* reduction=NULL){
//		NS_Utils::DOUT dout("Anomaly_exp_Pol::_set_mu","Kexp");

		_set_mu();//reduction);
//		dout()	<<"this:"<<endl;
//		dout()	<<*this<<endl;


		switch (pc) {
		case PC_NO_RES:
			Kexp[1]	+= kexpfactor*(*as)*_P[1][1];	//<- as * LM * _F_a_l[1][1] = eta*L_T e^-1
			_P[1][1]	= 0.;
			_P[1].reduce_order();
			if(_order<=1) reduce_order();
			break;
		case PC_NAIVE:
			Kexp[1]	+= kexpfactor*_P[0][1];	//<- as * LM * _F_a_l[1][1] = eta*L_T e^-1
			_P[0][1]	= 0.;
			_P[0].reduce_order();
			if(_order==0) reduce_order();
			break;
		case PC_STD:
		case PC_IMP:
			Kexp[1]	+= kexpfactor*_P[0][1];	//<- as * LM * _F_a_l[1][1] = eta*L_T e^-1
			_P[0][1]	= 0.;
			_P[0].reduce_order();

			Kexp[2]	+= kexpfactor*(*as)*_P[1][2];	//<- as^2 * LM * _F_a_l[2][2] + as * LT _F_a_l[1][1] = ( eta*Beta(0)/2. + C_B*G_0 )as^1 LT^2 e^0
			_P[1][2]	= 0.;
			_P[1].reduce_order();
			if(_order<=1) reduce_order();
			break;
		default:
			break;
		}

//		dout()	<<"this_after_Kexp:"<<endl;
//		dout()	<<*this<<endl;
	}

	inline void init(
			int order, POWER_COUNTING pc, bool exponent
			, const int *nf, const double *as, const double *LM
			, PROD prod
			){

		_init(order,pc,exponent,as,nf,prod);	//sets n=-1

//		if(res==RES_LLO){
//			_F_a_l.order(max_of(order,1));
//		}else{//RES_LL
		switch (pc) {
			case PC_NO_RES:
				_F_a_l.order(max_of(order,1));
				break;
			case PC_NAIVE:
				_F_a_l.order(order+1);// eta=as*LM sim ord(1)
				break;
			case PC_STD:
				_F_a_l.order(max_of(order+1,2));// --"-- and include eps^0
				break;
			case PC_IMP:
				_F_a_l.order(order+2);// as~eps^{2}
				break;
			default:
				break;
		}
//		}
		_F_a_l[0].order(0);
		_F_a_l[0][0]	=	0.;


		this->LM= LM;
		Gamma.init(prod);
	}

	Anomaly_exp_Pol()
	:_exp_pol_(7,8)
	,_F_a_l(Pol< Pol<double> >(7,Pol<double>(7,0.)))
	,_LM_F_a_l(Pol< Pol<double> >(7,Pol<double>(7,0.)))
	,_LT_F_a_l(Pol< Pol<double> >(7,Pol<double>(8,0.)))
	,Beta(),Gamma(PROD_WRONG_STATEMENT),d()
	,LM(NULL){

	}

	Anomaly_exp_Pol(const Anomaly_exp_Pol &F)
	:_exp_pol_(7,8)
	,_F_a_l(Pol< Pol<double> >(7,Pol<double>(7,0.)))
	,_LM_F_a_l(Pol< Pol<double> >(7,Pol<double>(7,0.)))
	,_LT_F_a_l(Pol< Pol<double> >(7,Pol<double>(8,0.)))
	,Beta(),Gamma(F.prod),d()
	,LM(F.LM){
//		NS_Utils::DOUT dout("Anomaly_exp_Pol","F");

		_exp_pol_::operator=(F);	//sets n=-1

		_F_a_l		= (F._F_a_l);
		_LM_F_a_l	= (F._LM_F_a_l);
		_LT_F_a_l	= (F._LT_F_a_l);

	}

	inline Anomaly_exp_Pol& operator=(const Anomaly_exp_Pol &F){
		_exp_pol_::operator=(F);	//sets n=-1

		_F_a_l		= (F._F_a_l);
		_LM_F_a_l	= (F._LM_F_a_l);
		_LT_F_a_l	= (F._LT_F_a_l);
		LM			= (F.LM);
		Gamma       = (F.Gamma);

		return *this;
	}

	~Anomaly_exp_Pol(){
	}

	inline virtual const string properties(){
		stringstream ss;
		ss	<<"F_exp_Pol:"<<_exp_pol_::properties();
//		ss	<<"\t"<<"LM="<<*LM<<endl;
//		ss	<<"\t"<<"   F = "<<_F_a_l<<endl;
//		ss	<<"\t"<<"LM*F = "<<_LM_F_a_l<<endl;
//		ss	<<"\t"<<"LT*F = "<<_LT_F_a_l<<endl;

		return ss.str();
	}

};


/*! \class G_exp_Pol
    \brief The exponent g = -(LT+LM)*F_{i,j} + h_{i} + h_{j}

*/
class G_exp_Pol:public _exp_pol_{
public:
	H_exp_Pol hpol;
	Anomaly_exp_Pol fpol;


	inline void _set_mu(){//const Continious_Order_Reduction* reduction=NULL){
//		NS_Utils::DOUT dout("G_exp_Pol::_set_mu");
//		dout()	<<"as="<<*as	<<"\t"<<"nf="<<*nf	<<endl;
		hpol._set_mu();//reduction);
		fpol._set_mu();//reduction);

//		dout()	<<"as="<<*as<<endl;

		set_equalt_to(hpol);
		mul_this_by_scalar(2.);
		add_this_Pol(fpol);
	}

	inline void _set_mu(Pol< complex<double> > &Kexp,const complex<double> kexpfactor=1.){//, const Continious_Order_Reduction* reduction=NULL){
//		NS_Utils::DOUT dout("G_exp_Pol::_set_mu","Kexp");
//		dout()	<<"as="<<*as	<<"\t"<<"nf="<<*nf	<<endl;

		hpol._set_mu(Kexp,kexpfactor*2.);//,reduction);
		fpol._set_mu(Kexp,kexpfactor);//,reduction);

		set_equalt_to(hpol);
		mul_this_by_scalar(2.);
		add_this_Pol(fpol);;
//		dout()	<<"gexp:"<<*this<<endl;
//		dout()	<<"kexp:"<<Kexp<<endl;
	}

	inline void init(
			int order, POWER_COUNTING pc, bool exponent
			,const int *nf, const double *as, const double *LM
			,PROD prod
			){
//		NS_Utils::DOUT dout("G_exp_Pol::init");
//		dout()	<<"res="<<res	<<"\t"<<"ord="<<order	<<"\timp="<<improved	<<endl;
//		NS_Utils::DOUT dout("G_exp_Pol::init");
//		dout()<<"prod="<<prod<<endl;

		hpol.init(order,pc,exponent,nf,as   ,prod);
		fpol.init(order,pc,exponent,nf,as,LM,prod);
		_init(order,pc,exponent,as,nf,prod);
	}


//	template<RES res>
//	inline void init(const CS_params<res>* p){
//		init(res,p->order,p->improved,p->exponent,&(p->nf),&(p->a),&(p->L_M),p->prod,p->Gamma_Cusp_3);
//	}

	G_exp_Pol()
	:_exp_pol_(7,8)
	,hpol(),fpol(){

	}

	G_exp_Pol(const G_exp_Pol& g)
	:_exp_pol_(7,8)
	,hpol(g.hpol),fpol(g.fpol){
//		NS_Utils::DOUT dout("G_exp_Pol","g");
		_exp_pol_::operator=(g);
	}

	inline virtual const string properties(){
		stringstream ss;
		ss	<<"G_exp_Pol:"<<_exp_pol_::properties();
		ss	<<hpol.properties()<<fpol.properties();

		return ss.str();
	}

};

}  // namespace NS_RESUMED



}  // namespace CuTe


#endif

