#ifndef TDPDF_H
#define TDPDF_H

#include "TDPDF_integrand.h"
#include "TDPDF_options.h"
//#include "cuba_int.h"

namespace CuTe {

namespace NS_TDPDF {

using NS_TDPDF_Integrand::E_INT;
using namespace NS_enums::NS_RES;
using namespace NS_enums::NS_BFUNCS;
using namespace NS_enums::NS_NP;

using NS_Utils::NS_pow::pow_2;
using NS_Utils::NS_min_max::min_of;

class TD_PDF{
public:
	static const int NB= NS_TDPDF_Integrand::_TDD_PDF_Integrand::NB;
	vector<double> results;
	vector<double> integration_errors;
	vector<double> integration_probabilities;

	const bool improved_pc()const{ return (pc==PC_IMP); }
protected:
	NS_TDPDF_Integrand::_TMD_PDF_Integrand _td_integrand;
	NS_Ibar::Ibar_G_Parameter IG;
	NS_Ibar::Ibar_Q_Parameter IQ;
	NS_LHAFUNC::VectorPDF pdf;
	NS_CUBA_integration::Cuba_Params cuba;

	double accuracy;

	RES res;
	double M;

	double xi;
	double qT;
	double xT;
	double mu;

	int		nf;	//nf(mu)
	double	as;	//alpha_s(mu)/(4*pi)
	double	LT;	//ln( xT^2 mu^2 b_0^(-2) )
	double	Lmu;//ln(mu^2/qT^2)
	double	q_star;

	ENUM_B_FUNCS B;
	int		order;
//	bool	improved;
	POWER_COUNTING		pc;
	bool	exponent;

	CUTOFF cutoff;
	double Lambda;
	double LM;		//ln(M^2/mu^2)

	inline void _init_cuba(){
		cuba.set(1,NB);
	}

	inline void _init_integrand(){
		_td_integrand.init(B,pdf,&qT,&xT,&mu,&nf,&as,&LT,&Lmu,&q_star,res,order,pc,exponent,cutoff,&accuracy,&Lambda,&LM,1.,0.);
	}

	inline void _init_parameter(){
		IG.init(order,improved_pc());
		IQ.init(order,improved_pc());
	}

	inline void _init_pdf(const string &pdfname, int member){
		pdf.create(pdfname,member);
	}

	inline void _init_pdf(const string &pdfname,int a, int b, int member){
		pdf.create(pdfname,member);
		pdf.set_to_dymmy(a,b);
	}

	inline void _set_xi_xT_mu(double xi, double xT, double mu){
//		NS_Utils::DOUT	dout("TD_PDF::_set_xi_xT_mu","xi,xT,mu");
		this->xi	= xi;
		this->xT	= xT;
		this->mu	= mu;

		nf	= pdf.nf(mu);
		as	= pdf.as(mu);
		LT	= log(pow_2(xT*mu/b_0));
//		Lmu	= log(gsl_pow_2(mu/qT));

		if(M>0.){
			LM	= log(pow_2(M/mu));
		}else{
			LM	= 0.;
		}

		IG.set_x(xi);
		IQ.set_x(xi);

		_td_integrand.set_xi_xT_mu(IG,IQ);
	}

	inline void _set_xi_qT_mu(double xi, double qT, double mu){
//		NS_Utils::DOUT	dout("TD_PDF::_set_xi_qT_mu","xi,qT,mu");
		this->xi	= xi;
		this->qT	= qT;
		this->mu	= mu;

		nf	= pdf.nf(mu);
		as	= pdf.as(mu);
//		LT	= log(gsl_pow_2(xT*mu/b_0));
		Lmu	= log(pow_2(mu/qT));

		if(M>0.){
			LM	= log(pow_2(M/mu));
		}else{
			LM	= 0.;
		}

		IG.set_x(xi);
		IQ.set_x(xi);

		_td_integrand.set_xi_qT_mu(IG,IQ);
	}

	template< E_INT e_int >
	inline double* _integrand(double ff[], double z, double Jac){
//		NS_Utils::DOUT dout("TD_PDF::_integrand<e_int>","ff,z,Jac");
		IG.update(z);
		IQ.update(z);

		return _td_integrand.integrand<e_int>(IG,IQ,ff,Jac);
	}

	template< E_INT e_int >
	inline static int _cuba_int(const int *ndim, const double xx[],const int *ncomp, double ff[], void *userdata){
//		NS_Utils::DOUT dout("TD_PDF::_cuba_int<e_int>","...");
		TD_PDF* td_pdf = static_cast<TD_PDF*>(userdata);
		double xi(td_pdf->xi);
		double Jac(1.-xi);
		double z(xi + Jac*xx[0] );

		td_pdf->_integrand<e_int>(ff,z,Jac);

		return 0;
	}

	template< E_INT e_int >
	inline void _integrate(){
//		NS_Utils::DOUT dout("TD_PDF::_integrate<e_int>");
		cuba.cube(_cuba_int<e_int>,results,integration_errors,integration_probabilities,this);
//		cuba.test(1,_cuba_int<e_int>,results,integration_errors,integration_probabilities,this);
	}

//	inline double* _integrand_B(double ff[], double z, double Jac){
////		NS_Utils::DOUT	dout("TD_PDF::_cuba_int()","ff,z,Jac");
//
//		IG.update(z);
//		IQ.update(z);
//
//		return _td_integrand.integrand_B_times_xi(IG,IQ,ff,Jac);
//	}
//
//	inline static int _cuba_int_B(const int *ndim, const double xx[],const int *ncomp, double ff[], void *userdata){
////		NS_Utils::DOUT	dout("TD_PDF::_cuba_int_()","*ndim...");
//		TD_PDF* tdd_pdf = static_cast<TD_PDF*>(userdata);
//		double xi(tdd_pdf->xi);
//		double Jac(1.-xi);
//		double z(xi + Jac*xx[0] );
//
//		tdd_pdf->_integrand_B(ff,z,Jac);
//
//		return 0;
//	}
//
//	inline void _integrate_B(){
////		NS_Utils::DOUT	dout("TD_PDF::_integrate()");
//		cuba.cube(_cuba_int_B,results,integration_errors,integration_probabilities,this);
////		cuba.test(1,_cuba_int_B,results,integration_errors,integration_probabilities,this);
//	}
//
//	inline double* _integrand_BTilde(double ff[], double z, double Jac){
////		NS_Utils::DOUT	dout("TD_PDF::_cuba_int()","ff,z,Jac");
//
//		IG.update(z);
//		IQ.update(z);
//
//		return _td_integrand.integrand_BTilde_times_xi(IG,IQ,ff,Jac);
//	}
//
//	inline static int _cuba_int_BTilde(const int *ndim, const double xx[],const int *ncomp, double ff[], void *userdata){
////		NS_Utils::DOUT	dout("TD_PDF::_cuba_int_()","*ndim...");
//		TD_PDF* tdd_pdf = static_cast<TD_PDF*>(userdata);
//		double xi(tdd_pdf->xi);
//		double Jac(1.-xi);
//		double z(xi + Jac*xx[0] );
//
//		tdd_pdf->_integrand_BTilde(ff,z,Jac);
//
//		return 0;
//	}
//
//	inline void _integrate_BTilde(){
////		NS_Utils::DOUT	dout("TD_PDF::_integrate()");
//		cuba.cube(_cuba_int_BTilde,results,integration_errors,integration_probabilities,this);
////		cuba.test(1,_cuba_int_BTilde,results,integration_errors,integration_probabilities,this);
//	}

public:
	TD_PDF()
	:results(NB,0.)
	,integration_errors(NB,0.)
	,integration_probabilities(NB,0.)
	,_td_integrand()
	,IG(),IQ()
	,pdf()
	,cuba()
	,accuracy(1e-3)
	,res(RES_LL),M(-1)
	,xi(),qT(),xT(),mu()
	,nf(),as(),LT(),Lmu(),q_star(5.)
	,B(),order(),pc(),exponent()
	,cutoff(),Lambda(),LM(){
	}

	TD_PDF(const string &lhapdf_name, int member=0)
	:results(NB,0.)
	,integration_errors(NB,0.)
	,integration_probabilities(NB,0.)
	,_td_integrand()
	,IG(),IQ()
	,pdf()
	,cuba()
	,accuracy(1e-3)
	,res(RES_LL),M(-1)
	,xi(),qT(),xT(),mu()
	,nf(),as(),LT(),Lmu(),q_star(5.)
	,B(E_BFUNC_B),order(2),pc(PC_STD),exponent(0)
	,cutoff(CUTOFF_NO),Lambda(0.),LM(){
		_init_cuba();
		_init_pdf(lhapdf_name,member);
		_init_parameter();
		_init_integrand();
	}

	TD_PDF(const TD_PDF &T)
	:results(T.results)
	,integration_errors(T.integration_errors)
	,integration_probabilities(T.integration_probabilities)
	,_td_integrand()
	,IG(T.IG),IQ(T.IQ)
	,pdf(T.pdf)
	,cuba(T.cuba)
	,accuracy(T.accuracy)
	,res(T.res),M(T.M)
	,xi(T.xi),qT(T.qT),xT(T.xT),mu(T.mu)
	,nf(T.nf),as(T.as),LT(T.LT),Lmu(T.Lmu),q_star(T.q_star)
	,B(T.B),order(T.order),pc(T.pc),exponent(T.exponent)
	,cutoff(T.cutoff),Lambda(T.Lambda),LM(T.LM){
		_init_integrand();
	}

	TD_PDF& operator=(const TD_PDF &T){
		integration_errors			= (T.integration_errors);
		integration_probabilities	= (T.integration_probabilities);
		IG		= (T.IG);
		IQ		= (T.IQ);
		pdf		= (T.pdf);
		cuba	= (T.cuba);
		accuracy= (T.accuracy);
		res	= (T.res);
		M	= (T.M);
		xi	= (T.xi);
		qT	= (T.qT);
		xT	= (T.xT);
		mu	= (T.mu);
		nf	= (T.nf);
		as	= (T.as);
		LT	= (T.LT);
		Lmu	= (T.Lmu);
		q_star		= (T.q_star);
		B			= (T.B);
		order		= (T.order);
		pc			= (T.pc);
		exponent	= (T.exponent);
		cutoff		= (T.cutoff);
		Lambda		= (T.Lambda);
		LM			= (T.LM);
		_init_integrand();
		return *this;
	}

	//	Eliminates all dependences on the hard scale M, by setting them to 0
	inline void reset_Hard_Scale(){
		M	= -1.;
	}

	/*	Specifies the Hard Scale M
	 * 	-> Important for
	 * 		- the Anomaly Factor in E_BFUNC_FB
	 * 		- Enhanced non-perturbative effects
	 */
	inline void set_Hard_Scale(double M){
		if(M>0.)	this->M	= M;
	}

	/* Define the function to Calculate
	 * E_BFUNC_B:
	 * 	- Transverse-Distance-Dependent-PDF = B
	 * 	- Transverse-Momentum-Dependent-PDF = tilde(B)
	 * E_BFUNC_Bbar
	 * 	- the RGE-independent part of B		= bar(B)
	 * 	- analogously						= tilde(bar(B))
	 * E_BFUNC_FB
	 * 	- B times the square root of the anomaly-factor	!!(M dependent)!!
	 * 	- analogously !!(M dependent)!!
	 */
	inline void set_TD_function(ENUM_B_FUNCS B, double M=-1.){
		this->B	= B;
		set_Hard_Scale(M);
		_init_integrand();
	}

	/*
	 * specify the PDF- and alpha_s-function to use, by LHAPDF-name and -member
	 */
	inline void set_lhapdf(const string &lhapdf_name, int member=0){
		_init_pdf(lhapdf_name,member);
		_init_integrand();
	}


	/*
	 * specify only the alpha_s-function to use, by LHAPDF-name and -member
	 * and use dummy pdfs
	 * phi_i(x) = x^a * (1-x)^b	 for all i<nf(mu)
	 */
	inline void set_lhapdf_dummy(const string &lhapdf_name, int a, int b, int member=0){
		_init_pdf(lhapdf_name,member,a,b);
		_init_integrand();
	}

	inline void set_relative_integration_accuracy(double acc=1e-3){
		this->accuracy	= acc;
		_init_cuba();
	}


	/* Sets the perturbative order
	 *
	 * order -> see Improved
	 *
	 * Improved specifies the power-counting
	 * 	- False	-> count in alpha_s	-> order in (0,1,2)
	 * 	- True	-> count in epsilon (alpha_s ~ epsilon^2, LT ~ 1/epsilon )	-> order in (0,1,2,3,4,5)
	 *
	 * exponent specifies the expansion scheme of terms in the exponent
	 * 	- False	-> they will be expanded
	 * 	- True	-> they stay in the exponent
	 *
	 * resummation specifies how the Fourier-Integral has to be evaluated
	 * 	- RES_LL	-> gives the full resummed result
	 * 	- RES_LLO	-> gives the fixed-order small_qT expansion
	 * 		->	improved = false
	 * 		->	order in (0,1,2)
	 *
	 */
	inline void set_perturbative_order(int order, POWER_COUNTING pc=PC_STD, bool exponent = false, RES resummation=RES_LL){
		this->order		= order;
		this->pc		= pc;
		this->exponent	= exponent;
		this->res		= resummation;

		if(res==RES_LLO){
			pc		 = PC_STD;
			order	 = min_of(order,2);
		}

		_init_parameter();
		_init_integrand();
	}


	/* Initializes a non-perturbative cutoff proportional to Lambda
	 * CUTOFF_NO
	 * CUTOFF_ENHANCED	<- needs the specification of the hard scale M
	 * CUTOFF_GAUSS
	 * CUTOFF_DIPOL
	 */
	inline void set_non_perturbative_effects(CUTOFF cutoff, double Lambda, double M=-1.){
		this->cutoff	= cutoff;
		this->Lambda	= Lambda;
		set_Hard_Scale(M);
		if(Lambda==0.) this->cutoff = CUTOFF_NO;
		_init_integrand();
	}

	inline vector<double>& xi_times_B_of_(double xi, double xT, double mu){
		_set_xi_xT_mu(xi,xT,mu);
//		_integrate_B();
		_integrate<NS_TDPDF_Integrand::INT_XI_B>();
		return results;
	}

	inline vector<double>& xi_times_BTilde_of_(double xi, double qT, double mu){
		_set_xi_qT_mu(xi,qT,mu);
		_integrate<NS_TDPDF_Integrand::INT_XI_Btilde>();
		return results;
	}

	inline int get_nf(double mu)const{ return pdf.nf(mu);}
	inline double get_as(double mu)const{ return pdf.as(mu);}
	inline double get_alpha_s(double mu)const{ return 4.*M_PI*pdf.as(mu);}
	inline complex<double> get_as(complex<double> mu)const{ return pdf.as(mu);}
	inline complex<double> get_alpha_s(complex<double> mu)const{ return 4.*M_PI*pdf.as(mu);}

};



}

}  // namespace CuTe
#endif
