/*
 * cuba_int.h
 *
 *  Created on: May 1, 2014
 *      Author: daniel
 */

#ifndef CUBA_INT_H_
#define CUBA_INT_H_

#include "cuba_options.h"
#include "utils.h"
#include "gsl_rand.h"



namespace CuTe {
namespace NS_CUBA_integration {


typedef int (*cuba_integrand_t)(
								 const int *ndim
								,const double x[]
								,const int *ncomp
								,double f[]
								,void *userdata);

typedef void(*CUBA_IN_EX_FUNC)();
typedef void(*CUBA_TRUE_IN_EX_FUNC)(void* params, int& core);
typedef int (*cuba_integrand_t_vegas)(	 const int* ndim
								,const double x[]
								,const int* ncomp
								,double f[]
								,void* userdata
								,const int* nvec
								,const double* weight
								,const int* iter
								);

//using namespace NS_enums;

using std::vector;
using std::cout;
using std::endl;
//using Shared_Memory_Control::SM_Serial_ID;


/*namespace CUBACORE_TEST_FUNCS {

	static const int MASTER = 32768;

	static int test_integrand(const int *ndim, const double xx[],const int *ncomp, double ff[], void *userdata);

	static void test_init(void* params, int &core);


}  // namespace CUBACORE_TEST F

class CUBACORE_GETTER{
	static const int MASTER = 32768;
protected:

//	static int test_integrand(const int *ndim, const double xx[],const int *ncomp, double ff[], void *userdata);
//
//	static void test_init(void* params, int &core);

	int _test_CUBACORES(int points_per_core);
public:

	inline CUBACORE_GETTER(){
	}

	inline int get_ncores(int points_per_core=10000){
		return _test_CUBACORES(points_per_core);
	}

	void set_ncores(int ncores,int points_per_core=10000);

};*/


class Cuba_Params{
protected:
	NS_Utils::Timer timer;
	static const int cuba_uninitialized;
	//CUBACORE_GETTER CCG;
public:
	int nregions;
	int neval;
	int fail;
	vector<double> integral;
	vector<double> error;
	vector<double> prob;

public:
//	enum e_routine {e_vegas, e_suave, e_divonne, e_cuhre};
	int ncores; // # of CUBACORES
	int ncoresmax;	// # Points Per Core
	CUBA_R routine;

	int ndim;	//	Dimension of Volume
	int ncomp;	//	Dimension of integrand
	int nvec;	//	The maximum number of points to be given to the integrand
				//	routine in each invocation. Usually this is 1 but if the integrand can profit from e.g.
				//	SIMD vectorization, a larger value can be chosen.
	double epsrel;	// Relative error
	double epsabs; // Asolute error
	int flags; // flags governing the integration
	int seed;	// Seed for random number generator
	int mineval; // Minimum Number of integrand evaluations
	int maxeval; // Maximum Number of integrand evaluations
	char* statefile; // a filename for storing the internal state

//	VEGAS specific
	int nstart; // the number of integrand evaluations per iteration to start with
	int nincrease; // the increase in the number of integrand evaluations per iteration
	int nbatch; // the batch size for sampling
	int gridno; // the slot in the internal grid table


//	SUAVE specific
	int nnew;	// the number of new integrand evaluations in each subdivision
	int nmin;	// minimum number of samples a former pass must contribute to a
				// subregion to be considered in that region’s compound integral value.
	double flatness; // the parameter p

//	DIVONNE specific
	int key1; // determines sampling in the partitioning phase
	int key2; // ;determines sampling in the final integration phase
	int key3; // sets the strategy for the refinement phase
	int maxpass; // controls the thoroughness of the partitioning phase
	double border; // the width of the border of the integration region
	double maxchisq; // the maximum χ2 value a single subregion is allowed to have in the final integration phase
	double mindeviation; // a bound, given as the fraction of the requested error of the entire integral


	int NGIVEN; // the number of points in the xgiven array
	int LDXGIVEN; // the leading dimension of xgiven, i.e. the offset between one point and the next in memory
	int NEXTRA;

//	CUHRE specific
	int key; // chooses the basic integration rule

protected:

	cuba_integrand_t integrand;
	void* userdata;
	CUBA_TRUE_IN_EX_FUNC initfunc,exitfunc;

	int _cube(vector<double> &out_results,vector<double> &out_errors,vector<double> &out_probs);
	int _test(int n,vector<double> &out_results,vector<double> &out_errors,vector<double> &out_probs);
	int _test_vegas(int n,vector<double> &out_results,vector<double> &out_errors,vector<double> &out_probs);
	void _initCubaCores();
	void _set_ncores(const int ncores,const int points_per_core)const;




public:

	int cube(cuba_integrand_t integrand, vector<double> &out_results,vector<double> &out_errors,vector<double> &out_probs, void* userdata=NULL){
		this->integrand	= integrand;
		this->userdata	= userdata;
		_set_ncores(ncores,ncoresmax);
		return _cube(out_results,out_errors,out_probs);
	}

	int cube(cuba_integrand_t integrand, void* userdata=NULL){
		this->integrand	= integrand;
		this->userdata	= userdata;
		_set_ncores(ncores,ncoresmax);
		return _cube(integral,error,prob);
	}

	int cube(cuba_integrand_t_vegas integrand, void* userdata=NULL){
		this->integrand	= (cuba_integrand_t)integrand;
		this->userdata	= userdata;
		_set_ncores(ncores,ncoresmax);
		return _cube(integral,error,prob);
	}
	int test(int n, cuba_integrand_t integrand, void* userdata=NULL){
		this->integrand	= integrand;
		this->userdata	= userdata;
		_set_ncores(0,ncoresmax);
		return _test(n,integral,error,prob);
	}
	int test(int n, cuba_integrand_t integrand, vector<double> &out_results,vector<double> &out_errors,vector<double> &out_probs, void* userdata=NULL){
		this->integrand	= integrand;
		this->userdata	= userdata;
		_set_ncores(0,ncoresmax);
		return _test(n,out_results,out_errors,out_probs);
	}

	int test(int n, cuba_integrand_t_vegas integrand, void* userdata=NULL){
		this->integrand	= (cuba_integrand_t)integrand;
		this->userdata	= userdata;
		_set_ncores(0,ncoresmax);
		return 	_test_vegas(n,integral,error,prob);
	}

	inline void set(int dim_, int comp_){
		ndim	= (dim_);
		ncomp	= (comp_);
		if(ndim==1){
			if((routine==CUBA_R_CUHRE)||(routine==CUBA_R_DIVONNE)) ++ndim;
		}
	}

	inline void set_in_ex( CUBA_TRUE_IN_EX_FUNC initfunc=NULL,CUBA_TRUE_IN_EX_FUNC exitfunc=NULL){
		this->initfunc	= initfunc;
		this->exitfunc	= exitfunc;
	}

	inline void set(CUBA_R c_routine, int dim_, int comp_){
		routine	= c_routine;
		set(dim_,comp_);
	}

//	inline void init(const Options& op,int dim_=1, int comp_=1){
	inline void init(const OPS_CUBA& op,int dim_=1, int comp_=1){
//		cout	<< "-> Cuba_Params::init(config,fail,"<<dim_<<","<<comp_<<")" << endl;

//		delete ccs;	// No = NULL
//		ccs = new CUBACORE_SETTER(op);
//		ncores = atoi(getenv("CUBACORES"));
		set(dim_,comp_);

		ncoresmax	= 10000;

		if(op.CUBA_cores.is_set){
			set_ncores(op.CUBA_cores());
		}else{
			set_ncores(-1);
		}

		routine 	= op.CUBA_routine();
		nvec 		= 1;
		epsrel 		= op.accuracy();
		epsabs 		= 1e-12;
		flags 		= op.CUBA_flags();
		seed 		= op.CUBA_seed();
		mineval 	= op.CUBA_mineval();
		maxeval 	= op.CUBA_maxeval();
		statefile 	= NULL;
		//VEGAS
		nstart 		= op.CUBA_VEGAS_nstart();
		nincrease 	= op.CUBA_VEGAS_nincrease();
		nbatch 		= op.CUBA_VEGAS_nbatch();
		gridno 		= 0;
		//SUAVE
		nnew 		= op.CUBA_SUAVE_nnew();
		nmin 		= op.CUBA_SUAVE_nmin();
		flatness 	= op.CUBA_SUAVE_flatness();
		//DIVONNE
		key1		= op.CUBA_DIVONNE_key1();
		key2		= op.CUBA_DIVONNE_key2();
		key3		= op.CUBA_DIVONNE_key3();
		maxpass		= op.CUBA_DIVONNE_maxpass();
		border		= op.CUBA_DIVONNE_border();
		maxchisq	= op.CUBA_DIVONNE_maxchisq();
		mindeviation= op.CUBA_DIVONNE_mindeviation();
		NGIVEN 		= 0;
		LDXGIVEN 	= ndim;
		NEXTRA 		= 0;
		//CUHRE
		key			= op.CUBA_CUHRE_key();
//		cout	<< "<- Cuba_Params::init(config,fail)" << endl;
	}

	void set_ncores(const int ncores){
		if(ncores<0){
			this->ncores	= -1;
		}else{
			this->ncores	= ncores;
		}
		return	_set_ncores(ncores,ncoresmax);
	}

	inline Cuba_Params()
	:nregions(0),neval(0),fail(0)
	,integral(),error(),prob()
	,ncores(-1),ncoresmax(10000)
	,routine(CUBA_R_CUHRE)
	,ndim(0),ncomp(0),nvec(1)
	,epsrel(0),epsabs(0.)
	,flags(0),seed(0)
	,mineval(2000),maxeval(1000000)
	,statefile(NULL)
	//VEGAS
	,nstart(1000),nincrease(1000),nbatch(2000),gridno(0)
	//SUAVE
	,nnew(1000),nmin(2),flatness(25.)
	//DIVONNE
	,key1(47),key2(1),key3(1),maxpass(5),border(0.001),maxchisq(10)
	,mindeviation(0.25),NGIVEN(0),LDXGIVEN(0),NEXTRA(0)
	//CUHRE
	,key(9)
	,integrand(NULL),userdata(NULL),initfunc(NULL),exitfunc(NULL){
	}

	inline Cuba_Params(const OPS_CUBA& op,int dim_=1, int comp_=1)
	:nregions(0),neval(0),fail(0)
	,integral(),error(),prob()
	,ncores(-1),ncoresmax(10000)
	,routine(CUBA_R_WRONG)
	,ndim(0),ncomp(0),nvec(1)
	,epsrel(0),epsabs(0.)
	,flags(0),seed(0)
	,mineval(1000),maxeval(100000000)
	,statefile(NULL)
	//VEGAS
	,nstart(1000),nincrease(1000),nbatch(2000),gridno(0)
	//SUAVE
	,nnew(1000),nmin(2),flatness(25.)
	//DIVONNE
	,key1(47),key2(1),key3(1),maxpass(5),border(0.01),maxchisq(10)
	,mindeviation(0.25),NGIVEN(0),LDXGIVEN(0),NEXTRA(0)
	//CUHRE
	,key(9)
	,integrand(NULL),userdata(NULL),initfunc(NULL),exitfunc(NULL){
		init(op,dim_,comp_);
	}

	inline Cuba_Params& operator=(const Cuba_Params& CP){
		nregions	= (CP.nregions);
		neval		= (CP.neval);
		fail		= (CP.fail);
		integral	= (CP.integral);
		error		= (CP.error);
		prob		= (CP.error);
		ncores		= (CP.ncores);
		ncoresmax	= (CP.ncoresmax);
		routine		= (CP.routine);
		ndim		= (CP.ndim);
		ncomp		= (CP.ncomp);
		nvec		= (CP.nvec);
		epsrel		= (CP.epsrel);
		epsabs		= (CP.epsabs);
		flags		= (CP.flags);
		seed		= (CP.seed);
		mineval		= (CP.mineval);
		maxeval		= (CP.maxeval);
		statefile	= (CP.statefile);
		//VEGAS;
		nstart		= (CP.nstart);
		nincrease	= (CP.nincrease);
		nbatch		= (CP.nbatch);
		gridno		= (CP.gridno);
		//SUAVE;
		nnew		= (CP.nnew);
		nmin		= (CP.nmin);
		flatness	= (CP.flatness);
		//DIVONNE;
		key1		= (CP.key1);
		key2		= (CP.key2);
		key3		= (CP.key3);
		maxpass		= (CP.maxpass);
		border		= (CP.border);
		maxchisq	= (CP.maxchisq);
		mindeviation= (CP.mindeviation);
		NGIVEN		= (CP.NGIVEN);
		LDXGIVEN	= (CP.LDXGIVEN);
		NEXTRA		= (CP.NEXTRA);
		//CUHRE;
		key			= (CP.key);

		return *this;
	}

	inline Cuba_Params(const Cuba_Params& CP)
	:nregions(CP.nregions),neval(CP.neval),fail(CP.fail)
	,integral(CP.integral),error(CP.error),prob(CP.error)
	,ncores(CP.ncores),ncoresmax(CP.ncoresmax)
	,routine(CP.routine)
	,ndim(CP.ndim),ncomp(CP.ncomp),nvec(CP.nvec)
	,epsrel(CP.epsrel),epsabs(CP.epsabs)
	,flags(CP.flags),seed(CP.seed)
	,mineval(CP.mineval),maxeval(CP.maxeval),statefile(CP.statefile)
	//VEGAS
	,nstart(CP.nstart),nincrease(CP.nincrease),nbatch(CP.nbatch),gridno(CP.gridno)
	//SUAVE
	,nnew(CP.nnew),nmin(CP.nmin),flatness(CP.flatness)
	//DIVONNE
	,key1(CP.key1),key2(CP.key2),key3(CP.key3)
	,maxpass(CP.maxpass),border(CP.border)
	,maxchisq(CP.maxchisq),mindeviation(CP.mindeviation)
	,NGIVEN(CP.NGIVEN),LDXGIVEN(CP.LDXGIVEN),NEXTRA(CP.NEXTRA)
	//CUHRE
	,key(CP.key)
	,integrand(NULL),userdata(NULL),initfunc(NULL),exitfunc(NULL){
	}

	inline ~Cuba_Params(){
	}
};


}  // namespace CUBA_ing
}  // namespace CuTe




#endif /* CUBA_INT_H_ */
