#ifndef GSL_INT_H_
#define GSL_INT_H_
#include <string>

namespace CuTe {
namespace NS_GSL_INT {


typedef double (*GSL_Integrand)(double, void*);

class GSL_Integrator {
public:
	GSL_Integrand function;
	void* params;
	double result,error;
	int gsl_status;
protected:

	size_t memory;
	void* workspace;

	GSL_Integrator()
	:function(NULL),params(NULL)
	,result(0.),error(0.),gsl_status(0)
	,memory(0),workspace(NULL){

	}

	void _alloc_workspace();
	void _free_workspace();
	bool _status(const std::string &name)const;

public:

	GSL_Integrator(const size_t memory)
	:function(NULL),params(NULL)
	,result(0.),error(0.),gsl_status(0)
	,memory(memory),workspace(NULL){
		_alloc_workspace();
	}

	GSL_Integrator(const GSL_Integrator &other)
	:function(other.function),params(other.params)
	,result(other.result),error(other.error),gsl_status(other.gsl_status)
	,memory(other.memory),workspace(NULL){
		_alloc_workspace();
	}

	GSL_Integrator& operator=(const GSL_Integrator &other){

		function	= (other.function);
		params		= (other.params);
		result		= (other.result);
		error		= (other.error);
		gsl_status	= (other.gsl_status);
		if(memory!=other.memory){
			_free_workspace();
			memory		= (other.memory);
			_alloc_workspace();
		}

		return *this;
	}

	~GSL_Integrator(){
		_free_workspace();
	}

	double integrate_a_to_b(const double a, const double b, const double acc_rel, const double acc_abs=0., const int key=3);
	double integrate_a_to_inf(const double a, const double acc_rel, const double acc_abs=0.);
	double integrate_minf_to_inf(const double acc_rel, const double acc_abs=0.);
	double integrate_minf_to_b(const double b, const double acc_rel, const double acc_abs=0.);

	double test_a_to_b(const double a, const double b, const double acc_rel, const int N);

	inline bool has_failed()const{ return (gsl_status!=0);}

	inline bool status(const std::string &name=std::string("f"))const{
		if(has_failed()) return _status(name);
		return false;
	}


};

}  // namespace NS_GSL_INT
}  // namespace CuTe

#endif //GSL_INT_H_
