#ifndef OP_EXCEPTIONS_H
#define OP_EXCEPTIONS_H

#include <exception>
#include <string>
#include <sstream>

namespace NS_OPTIONS {

using std::exception;
using std::string;
using std::stringstream;

class file_not_found_exception: public exception{
	string filename;
public:
	virtual const char* what() const throw(){
		return filename.c_str();
	}
	file_not_found_exception(const string& filename)
	:filename(filename){
	}

	virtual ~file_not_found_exception() throw(){}
};

class key_not_found_exception: public exception{
	string key;
public:
	virtual const char* what() const throw(){
		return key.c_str();
	}
	key_not_found_exception(const string& key)
	:key(key){
	}

	virtual ~key_not_found_exception() throw(){}
};

class bad_in_exception: public exception{
	string s;
public:
	virtual const char* what() const throw(){
		return s.c_str();
	}
	bad_in_exception(const string& value)
	:s(value){
	}

	bad_in_exception(const string& value, const string& name, const string& expl){
		stringstream ss;
		ss << "Wrong \"" << name << "\" Parameter: " << value << " != " << expl;
		s=ss.str();
	}

	bad_in_exception(const string& value, const string& name, const string& expl, int N, int n){
		stringstream ss;
		ss << "Wrong \"" << name << "\" Number of Parameters: "<<n<<" != "<<N<< "\t:" << value << " != " << expl;
		s=ss.str();
	}

	bad_in_exception(const bad_in_exception& e, const string& name, const string& expl){
		stringstream ss;
		ss << "Wrong \"" << name << "\" Parameter: " << e.what() << " != " << expl;
		s=ss.str();
	}

	virtual ~bad_in_exception() throw(){}
};

}  // namespace NS_OPTIONS
#endif
