#ifndef OPTION_H
#define OPTION_H

//#include "op_exceptions.h"
#include "ConfigFile.h"
#include "File_Path.h"
#include <cstdlib>

namespace NS_OPTIONS{
using std::cout;
using std::cerr;
using std::endl;
using std::ostream;
using std::string;
using std::stringstream;
using std::vector;


static const int name_width=20;
static const int val_width=5;




enum e_read {e_to_val,e_to_def};


class _Option_Interface {
public:
	string name;
	string explanation;

	bool is_set;
	bool has_defv;

	bool has_defe;
	string def_explanation;

	bool verbose_defv;
	bool verbose_defe;

protected:

	inline _Option_Interface()
	:name(),explanation()
	,is_set(false)
	,has_defv(false)
	,has_defe(false),def_explanation()
	,verbose_defv(true),verbose_defe(true){
	}

	inline _Option_Interface(const string& name, const string& explanation)
	:name(name),explanation(explanation)
	,is_set(false)
	,has_defv(false)
	,has_defe(false),def_explanation()
	,verbose_defv(true),verbose_defe(true){
	}

	inline _Option_Interface(const _Option_Interface& other)
	:name(other.name),explanation(other.explanation)
	,is_set(other.is_set)
	,has_defv(other.has_defv)
	,has_defe(other.has_defe),def_explanation(other.def_explanation)
	,verbose_defv(other.verbose_defv),verbose_defe(other.verbose_defe){

	}

public:

	virtual bool read(ConfigFile& config,e_read to_which){
		cerr	<<"*********************************************"	<<endl
				<<"**    virtual _Option_Interface::read()    **"	<<endl
				<<"*********************************************"	<<endl
				;
		return false;
	}


	virtual ~_Option_Interface(){

	}

	virtual stringstream& properties(stringstream &ss)const{
		ss		<<"*********************************************"	<<endl
				<<"** virtual _Option_Interface::properties() **"	<<endl
				<<"*********************************************"	<<endl
				;
		return ss;
	}

	virtual const string properties()const{
		stringstream ss;
		string s = properties(ss).str();
		cerr	<< s;
		return s;
	}

	virtual ostream& to_stream(ostream& os)const{
		cerr	<<"*********************************************"	<<endl
				<<"** virtual _Option_Interface::to_stream()  **"	<<endl
				<<"*********************************************"	<<endl
				;
		return os;
	}
};

template< typename T >
class Option: public _Option_Interface {
public:
	T value;
	T def_val;

	inline Option()
	:_Option_Interface()
	,value(),def_val(){
	}

	inline Option(const string& name, const string& explanation)
	:_Option_Interface(name,explanation)
	,value(),def_val(){
	}

	inline Option(const string& name, const string& explanation, const T& initialization_value)
	:_Option_Interface(name, explanation)
	,value(initialization_value),def_val(){
		is_set = true;
	}

	inline Option(const Option& other)
	:_Option_Interface(other)
	,value(other.value),def_val(other.def_val){

	}

	inline void set_def_val(const T& value){
		has_defv		= true;
		def_val			= value;
		has_defe		= false;
		def_explanation = string();
	}

	void set_def_exp(const string& def_explanation){
		has_defv		= false;
		def_val			= T();
		has_defe		= true;
		this->def_explanation = def_explanation;
	}

	virtual bool read(ConfigFile& config,e_read to_which){
		if(to_which == e_to_val){
			is_set = true;

			try {
				value = config.read<T>(name);
			}catch(const key_not_found_exception& e){
				if(has_defv){
					int tmp_w(cout.width());
					if(verbose_defv){
						cout << "Parameter \"";
						cout.width(name_width);
						cout << name;
						cout.width(tmp_w);
						cout << "\" not found: -> default = " << def_val << endl;
					}
					value = def_val;
				}else if(has_defe){
						int tmp_w(cout.width());
						if(verbose_defe){
							cout << "Parameter \"";
							cout.width(name_width);
							cout << name;
							cout.width(tmp_w);
							cout << "\" not found: -> " << def_explanation << endl;
						}
						is_set = false;
				}else{
					is_set = false;
					throw e;
				}
			}catch (const bad_in_exception& e) {
				is_set = false;
				throw bad_in_exception(e,name,explanation);
			}
			return is_set;
		}else if(to_which==e_to_def){
			try {
				if(has_defv){
					bool tmp	= config.readInto(def_val,name,def_val);
					if(tmp) verbose_defv = false;
				}else{
					has_defv	= config.readInto(def_val,name);
					if(has_defv) verbose_defv = false;
				}
			}catch(const key_not_found_exception& e){
				//nothing
			}catch (const bad_in_exception& e) {
				has_defv = false;
				throw bad_in_exception(e,name,explanation);
			}

			return has_defv;
		}
		return false;
	}


//	inline T& val(){
//		if(!is_set){
//			cerr	<<"*** !!NOT CONST!! OPTION "<<name<<" NOT SET ***"<<endl;
//		}
//		return value;
//	}
	inline const T& val()const{
		if(!is_set){
			cerr	<<"*** OPTION "<<name<<" NOT SET ***"<<endl;
		}
		return value;
	}

//	inline T& operator()(){
//		return val();
//	}
	inline const T& operator()()const{
		return val();
	}

	virtual stringstream& properties(stringstream &ss)const{
		ss	<<"\t\t"<<name<<" "<<explanation<<":"	<<endl
			<<"\t\t"<<"is_set="<<is_set;
		if(is_set) ss	<<"\t"<<"to: "<<value;
		ss	<<endl
			<<"\t\t"<<"has_defv="<<has_defv;
		if(has_defv) ss	<<"\t"<<"to: "<<def_val;
		ss	<<endl
			<<"\t\t"<<"has_defe="<<has_defe;
		if(has_defe) ss	<<"\t"<<"to: "<<def_explanation;
		return ss;
	}

	virtual const string properties()const{
		stringstream ss;
		return properties(ss).str();
	}

	virtual ostream& to_stream(ostream& os)const{
		int w = os.width();
		if(is_set){
			os.width(name_width);
			os <<name;
			os.width(w);
			os	<<" = ";
			os.width(val_width);
			os	<<std::left;
			os	<<value;
		}else{
			os	<<"#";
			os.width(name_width-1);
			os <<name;
			os.width(w);
			os	<<" = ";
			if(has_defv){
				os.width(val_width);
				os	<<std::left<<def_val;
			}else{
				if(has_defe){
					os.width(val_width);
					os	<<std::left<<def_explanation;
				}else{
					os.width(val_width);
					os	<<"";
				}
			}
		}
		os.width(w);
		os	<<std::right;
		os	<<"\t#"<<explanation;
		return os;
	}

	virtual ~Option(){

	}

	inline void unset(){
		is_set		= false;
	}

	inline void set(const T& value){
		is_set		= true;
		this->value	= value;
	}
};

template< typename T >
ostream& operator<< (ostream& os, const Option<T>& op){
	return op.to_stream(os);
}

template< typename T, int N >
class VOption: public Option< vector<T> > {
public:
	using Option< vector<T> >::_Option_Interface::name;
	using Option< vector<T> >::_Option_Interface::explanation;
	using Option< vector<T> >::_Option_Interface::is_set;
	using Option< vector<T> >::_Option_Interface::has_defv;
	using Option< vector<T> >::_Option_Interface::has_defe;
	using Option< vector<T> >::_Option_Interface::def_explanation;
	using Option< vector<T> >::_Option_Interface::verbose_defv;
	using Option< vector<T> >::_Option_Interface::verbose_defe;

	using Option< vector<T> >::value;
	using Option< vector<T> >::def_val;
	using Option< vector<T> >::val;

	inline VOption()
	:Option< vector<T> >(){
	}

	inline VOption(const string& name, const string& explanation)
	:Option< vector<T> >(name, explanation){
	}

	inline VOption(const string& name, const string& explanation, const vector<T>& initialization_value)
	:Option< vector<T> >(name, explanation,initialization_value){
		int n	= initialization_value.size();
		if(n!=N){
			cerr	<<"VOption("<<name<<","<<explanation<<","<<initialization_value<<")"	<<endl;
			cerr	<<"\tWRONG # = "<<n <<" != "<<N	<<endl;
			std::exit(1);
		}
	}

	inline VOption(const VOption& other)
	:Option< vector<T> >(other){

	}


//	inline T& val(int i){
//		if(i>=N){
//			cerr	<<"*** OPTION "<<name<<" value["<<i<<" >= "<<N<<"] called" <<endl;
//		}
//		return val()[i];
//	}
//	inline const T& val(int i)const{
//		if(i>=N){
//			cerr	<<"*** OPTION "<<name<<" value["<<i<<" >= "<<N<<"] called" <<endl;
//		}
//		return val()[i];
//	}
//	inline T& operator[](int i){ return val()[i];}
//	inline const T& operator[](int i)const{return val()[i];}
	inline const T operator[](int i)const{
		if(i>=N){
			cerr	<<"*** OPTION "<<name<<" value["<<i<<" >= "<<N<<"] called" <<endl;
		}
		return val()[i];
	}

	virtual bool read(ConfigFile& config,e_read to_which){
		bool OK = Option< vector<T> >::read(config,to_which);
		if(OK){
			int n;
			if(to_which == e_to_val){
				n = value.size();
			}else if(to_which == e_to_def){
				n = def_val.size();
			}
			if(n!=N){
				stringstream svalue;
				svalue	<<value;
				OK	= false;
				throw bad_in_exception(svalue.str(),name,explanation,N,n);
			}
		}
		return OK;
	}


	virtual stringstream& properties(stringstream &ss)const{
		Option< vector<T> >::properties(ss);
		ss	<<"\t\t"	<<"# = "<<N	<<endl;
		return ss;
	}

	virtual const string properties()const{
		stringstream ss;

		return properties(ss).str();
	}

	virtual ~VOption(){

	}

};

template< typename T, int N >
ostream& operator<< (ostream& os, const VOption<T,N>& op){
	return op.to_stream(os);
}

using NS_File_Path::File_Path;

class _Options_Set{
protected:
	string OP_Set_Name;
	vector<_Option_Interface*> options;

	File_Path infilename;
	Option<File_Path> Basefile;

	/*
	 * define all options
	 */
	virtual void v_init_ops(){
		cerr	<<"*******************************************"	<<endl
				<<"*****  virtual _Options_Set::v_init() *****"	<<endl
				<<"*****  OP_Set_Name: "<<OP_Set_Name<<" *****"	<<endl
				<<"*******************************************"	<<endl
				;
		exit(1);
	}

	/*
	 * push ops into options* vector
	 */
	virtual void v_init_vec(){
		cerr	<<"*******************************************"	<<endl
				<<"*****  virtual _Options_Set::v_init() *****"	<<endl
				<<"*****  OP_Set_Name: "<<OP_Set_Name<<" *****"	<<endl
				<<"*******************************************"	<<endl
				;
		exit(1);
	}

	inline void init_vec(){
		options.clear();
		options.push_back(&Basefile);
		v_init_vec();
	}

	inline void init(){
//		cout	<<"->_Options_Set::init()"<<endl;
		Basefile	= Option<File_Path>("Basefile","[string]");
		Basefile.set_def_exp("No Basefile");
		Basefile.verbose_defe = false;
		v_init_ops();
		init_vec();
//		cout	<<"<-_Options_Set::init()"<<endl;
	}

	inline void init(const string& infilename){
//		cout	<<"->_Options_Set::init(infilename)"<<endl;
		init();
		if(!open(infilename)){
			cerr	<<"_Options_Set::init("<<infilename<<")"	<<endl;
		}
//		cout	<<"<-_Options_Set::init(infilename)"<<endl;
	}

	_Options_Set()
	:OP_Set_Name("New _Options_Set")
	,options()
	,infilename()
	,Basefile(){
//		init();
	}

	_Options_Set(const string& set_Name)
	:OP_Set_Name(set_Name)
	,options()
	,infilename()
	,Basefile(){
//		init();
	}


	bool _set_defaults_from_Basefile(File_Path& Basefile,const File_Path& above,int depths);
	bool _read(ConfigFile& config,const File_Path& fp, e_read to_which, int depths=0);

	_Options_Set(const _Options_Set& OS)
	:OP_Set_Name(OS.OP_Set_Name)
	,options()
	,infilename(OS.infilename)
	,Basefile(OS.Basefile){
//		std::cout<<"_Options_Set("<<OP_Set_Name<<")::(OS)"<<std::endl;
	}

public:

	_Options_Set& operator=(const _Options_Set& OS){
//		std::cout<<"_Options_Set("<<OS.OP_Set_Name<<")::operator=(OS)"<<std::endl;

		OP_Set_Name	= (OS.OP_Set_Name);
		infilename	= (OS.infilename);
		Basefile	= (OS.Basefile);

		return *this;
	}

	stringstream& to_ss(stringstream& ss, const string pre=string(""))const;
	ostream& into_stream(ostream& os, const string pre=string(""))const;

	/*
	 * Open filename
	 * Recursive open Basefile
	 * Save to options
	 */
	bool open(const string& filename);

	/*
	 * Print properties of all single Options
	 */
	const string properties()const;


	const File_Path& get_infilename()const{
		return infilename;
	}

	inline virtual ~_Options_Set(){

	}
};

ostream& operator<<(ostream& os, const _Options_Set& op);

class _Set_of_Options_Sets{
protected:
	vector<_Options_Set*> _all;
	File_Path infilename;
	int N;

	inline int _init(){
		N	= _all.size();
		return N;
	}

//	virtual void v_init_all(){
//		cerr	<<"********************************************\n"
//				<<"**** _Set_Options_Sets::v_init() called ****\n"
//				<<"********************************************\n";
//		exit(1);
//	}
//
//	inline void init_all(){
////		_all.clear();
//		v_init_all();
//		std::cout<<"_Set_of_Options_Sets::init_all() "<<N;
//		N	= _all.size();
//		std::cout<<" -> "<<N<<endl;
//	}
//
//	inline void init_all(const string& infilename){
//		std::cout<<"_Set_of_Options_Sets::init_all("<<infilename<<")"<<std::endl;
//		init_all();
//		if(!open(infilename)){
//			cerr	<<"_Set_of_Options_Sets::_Set_of_Options_Sets("<<infilename<<")"	<<endl;
//		}
//	}

	_Set_of_Options_Sets()
	:_all(),N(_init()){
//		std::cout<<"_Set_of_Options_Sets()"<<std::endl;
	}

	_Set_of_Options_Sets(const _Set_of_Options_Sets& OP)
	:_all(),N(_init()),infilename(OP.infilename){
//		std::cout<<"_Set_of_Options_Sets(const _Set_of_Options_Sets& OP)"<<std::endl;
	}

public:

	_Set_of_Options_Sets& operator=(const _Set_of_Options_Sets& OP){
		infilename = OP.infilename;
		return *this;
	}

	bool open(const string& filename){
//		std::cout<<"_Set_of_Options_Sets::open("<<filename<<")"<<std::endl;
		infilename	= filename;
		bool OK(true);


		for(int i=0; i<N; ++i){
			OK = OK && _all[i]->open(filename);
		}


		return OK;
	}


	stringstream& to_ss(stringstream& ss, const string pre=string(""))const{

		for(int i=0; i<N; ++i){
			_all[i]->to_ss(ss,pre);
		}

		return ss;
	}
	ostream& into_stream(ostream& os, const string pre=string(""))const{
		stringstream ss;
		to_ss(ss,pre);
		os << ss.str();

		return os;

	}

	const string properties()const{
			stringstream ss;

			for(int i=0; i<N; ++i){
				ss	<< _all[i]->properties();
			}

			return ss.str();
		}

	virtual ~_Set_of_Options_Sets(){

	}

	const File_Path& get_infilename()const{
		return infilename;
	}

	void write_to_file(const string& ofname)const{
		std::ofstream of(ofname.c_str());

		of	<<"#\tInfile: "<<ofname<<endl
			<<"#\t<- comments"	<<endl
			<<"#\tAll mass afflicted parameters in GeV"	<<endl
			<<endl;

		into_stream(of);


		of.close();
	}


	void write_example()const{
		write_to_file("example.in");
	}

};

ostream& operator<<(ostream& os, const _Set_of_Options_Sets& op);

}  // namespace NS_OPTIONS
#endif
