#include <Kinematics_Two_P_PS.h>
#include <fstream>

using CuTe::NS_Kinematics::Two_P_PS;
using CuTe::NS_Kinematics::Two_P_PS_Interpolated;
using std::ofstream;
using std::cout;
using std::endl;

int main(int argc, char **argv) {
	const double sS(13000);
	const double M(90);
	const double mp_max((sS*sS+M*M)/(2.*sS));
	const double qT(10);
	const double mp(sqrt(M*M+qT*qT));
	const double y_max(acosh(mp_max/mp));
	const double y(0.6);
	const double a(20.);
	const double b(2.4);
	const double b0(1.2);
	const double b1(1.52);

	Two_P_PS ps;

	ps.init_and_return_min_mp(M);
	ps.update_qT_y(qT,y);
	cout<<ps<<endl;

	ps.init_and_return_min_mp(M,a,-1.);
	ps.update_qT_y(qT,y);
	cout<<ps<<endl;

	ps.init_and_return_min_mp(M,-1,b);
	ps.update_qT_y(qT,y);
	cout<<ps<<endl;

	ps.init_and_return_min_mp(M,a,b);
	ps.update_qT_y(qT,y);
	cout<<ps<<endl;

	ps.init_and_return_min_mp(M,-1.,-1.,b0,b1);
	ps.update_qT_y(qT,y);
	cout<<ps<<endl;

	ps.init_and_return_min_mp(M,-1.,b,b0,b1);
	ps.update_qT_y(qT,y);
	cout<<ps<<endl;

	Two_P_PS_Interpolated psI;
	ofstream of;
	char name[64];

	sprintf(name,"PS_full.txt");
	of.open(name);
	psI.init_and_return_min_mp(M);
	psI.update_qT_get_ipol_range(qT,y_max);
	psI.into_stream(of);
	of.close();

	sprintf(name,"PS_a.txt");
	of.open(name);
	psI.init_and_return_min_mp(M,a,-1.);
	psI.update_qT_get_ipol_range(qT,y_max);
	psI.into_stream(of);
	of.close();

	sprintf(name,"PS_b.txt");
	of.open(name);
	psI.init_and_return_min_mp(M,-1,b);
	psI.update_qT_get_ipol_range(qT,y_max);
	psI.into_stream(of);
	of.close();

	sprintf(name,"PS_a_b.txt");
	of.open(name);
	psI.init_and_return_min_mp(M,a,b);
	psI.update_qT_get_ipol_range(qT,y_max);
	psI.into_stream(of);
	of.close();

	sprintf(name,"PS_b0b1.txt");
	of.open(name);
	psI.init_and_return_min_mp(M,-1.,-1.,b0,b1,2000);
	psI.update_qT_get_ipol_range(qT,y_max);
	psI.into_stream(of);
	of.close();

	sprintf(name,"PS_a_b_b0b1.txt");
	of.open(name);
	psI.init_and_return_min_mp(M,-1.,b,b0,b1,2000);
	psI.update_qT_get_ipol_range(qT,y_max);
	psI.into_stream(of);
	of.close();



}
