/**
 * @file 	utils.cpp
 * @date	27.05.2014
 * @author	daniel
 * @brief	Datei_Beschreibung_Kurz
 * 
 * Datei_Beschreibung_Lang
 */

/** Funktion_Beschreibung_Kurz.
 * @param	Paramter_Name Parameter_Beschreibung
 * @return Return_Beschreibung
 * 
 * Funktion_Beschreibung_Lang
 */

#include <ostream>

#include "utils.h"
#ifdef __cplusplus
extern "C" {
#endif
	#include "gsl/gsl_errno.h"
#ifdef __cplusplus
}
#endif

using std::cerr;

void CuTe::NS_Utils::GSL_ERROR_CODE(const string& expl, const int status){
	switch(status)
	{
	case GSL_SUCCESS:	break;
	case GSL_CONTINUE: 	break;
	case GSL_EBADFUNC:	cerr << expl << "\t" <<  "GSL_EBADFUNC" << endl;	break;
	case GSL_EBADLEN: 	cerr << expl << "\t" <<  "GSL_EBADLEN" << endl;	break;
	case GSL_EBADTOL: 	cerr << expl << "\t" <<  "GSL_EBADTOL" << endl;	break;
	case GSL_ECACHE: 	cerr << expl << "\t" <<  "GSL_ECACHE" << endl;	break;
	case GSL_EDOM: 		cerr << expl << "\t" <<  "GSL_EDOM" << endl;	break;
	case GSL_EFACTOR: 	cerr << expl << "\t" <<  "GSL_EFACTOR" << endl;	break;
	case GSL_EFAILED: 	cerr << expl << "\t" <<  "GSL_EFAILED" << endl;	break;
	case GSL_EFAULT: 	cerr << expl << "\t" <<  "GSL_EFAULT" << endl;	break;
	case GSL_EINVAL: 	cerr << expl << "\t" <<  "GSL_EINVAL" << endl;	break;
	case GSL_ELOSS: 	cerr << expl << "\t" <<  "GSL_ELOSS" << endl;	break;
	case GSL_ENOMEM: 	cerr << expl << "\t" <<  "GSL_ENOMEM" << endl;	break;
	case GSL_ENOPROG: 	cerr << expl << "\t" <<  "GSL_ENOPROG" << endl;	break;
	case GSL_ENOPROGJ: 	cerr << expl << "\t" <<  "GSL_ENOPROGJ" << endl;	break;
	case GSL_ENOTSQR:	cerr << expl << "\t" <<  "GSL_ENOTSQR" << endl;	break;
	case GSL_EOF: 		cerr << expl << "\t" <<  "GSL_EOF" << endl;	break;
	case GSL_EOVRFLW:	cerr << expl << "\t" <<  "GSL_EOVRFLW" << endl;	break;
	case GSL_ERANGE:	cerr << expl << "\t" <<  "GSL_ERANGE" << endl;	break;
	case GSL_ERUNAWAY:	cerr << expl << "\t" <<  "GSL_ERUNAWAY" << endl;	break;
	case GSL_ESANITY:	cerr << expl << "\t" <<  "GSL_ESANITY" << endl;	break;
	case GSL_ETABLE:	cerr << expl << "\t" <<  "GSL_ETABLE" << endl;	break;
	case GSL_ETOL:		cerr << expl << "\t" <<  "GSL_ETOL" << endl;	break;
	case GSL_ETOLF:		cerr << expl << "\t" <<  "GSL_ETOLF" << endl;	break;
	case GSL_ETOLG:		cerr << expl << "\t" <<  "GSL_ETOLG" << endl;	break;
	case GSL_ETOLX:		cerr << expl << "\t" <<  "GSL_ETOLX" << endl;	break;
	case GSL_EUNDRFLW:	cerr << expl << "\t" <<  "GSL_EUNDRFLW" << endl;	break;
	case GSL_EUNIMPL:	cerr << expl << "\t" <<  "GSL_EUNIMPL" << endl;	break;
	case GSL_EUNSUP:	cerr << expl << "\t" <<  "GSL_EUNSUP" << endl;	break;
	case GSL_EZERODIV:	cerr << expl << "\t" <<  "GSL_EZERODIV" << endl;	break;
	case GSL_EMAXITER:	cerr << expl << "\t" <<  "the maximum number of subdivisions was exceeded." << endl;	break;
	case GSL_EROUND:	cerr << expl << "\t" <<  "cannot reach tolerance because of roundoff error, or roundoff error was detected in the extrapolation table." << endl;	break;
	case GSL_ESING:		cerr << expl << "\t" <<  "a non-integrable singularity or other bad integrand behavior was found in the integration interval." << endl;	break;
	case GSL_EDIVERGE:	cerr << expl << "\t" <<  "the integral is divergent, or too slowly convergent to be integrated numerically." << endl;	break;
	default:			cerr << expl << "\t" <<  "ERROR_CODE: " << status << endl;	break;
	}
}

int CuTe::NS_Utils::DOUT::tab = 0;

void CuTe::NS_Utils::DOUT::_init(){
	os	= &(cout);
//	os	= &(NULLos);
}


std::ostream& CuTe::NS_Utils::operator<< ( std::ostream& os,const CuTe::NS_Utils::Timer& t){
	os	<< t.t<<" sec.";
	return os;
}


bool CuTe::NS_Utils::interpolate_parabola(const double x0, const double y0, const double x1, const double y1, const double x2, const double y2, double &a, double &b, double &c){
//	DOUT dout("interpolate_parabola");

	const double den( (x0-x1)*(x0-x2)*(x1-x2) );
	if(den==0.) return false;
	const double iden(1./den);

	a	= -(    x0*(y1-y2) +    x1*(y2-y0) +    x2*(y0-y1) )*iden;
	b	= +( x0*x0*(y1-y2) + x1*x1*(y2-y0) + x2*x2*(y0-y1) )*iden;
	c	= y0 - x0*(b + a*x0);


//	dout()	<<"{{"<<x0<<","<<y0<<"},{"<<x1<<","<<y1<<"},{"<<x2<<","<<y2<<"}}"	<<endl;
//	dout()	<<"a = "<<a	<<"\t"<<"b = "<<b	<<"\t"<<"c = "<<c	<<endl;

	return true;
}


bool CuTe::NS_Utils::parabola_min_max(const double x0, const double y0, const double x1, const double y1, const double x2, const double y2, double &min, double &max){
//	DOUT dout("parabola_min_max");
	double a,b,c;
	if(!interpolate_parabola(x0,y0,x1,y1,x2,y2,a,b,c)) return false;

	double xmin,xmax;

	if( x0 < x1 ){ // x0 < x1
		if( x1 < x2 ){	// x0 < x1 < x2
			xmin	= x0;
			min		= y0;
			xmax	= x2;
			max		= y2;
//			centre	= y1;
		}else{	// (x0 < x1)&&(x2 < x1)
			if( x0 < x2 ){ // x0 < x2 < x1
				xmin	= x0;
				min		= y0;
//				centre	= y2;
			}else{ // x2 < x0 < x1
				xmin	= x2;
				min		= y2;
//				centre	= y0;
			}
			xmax	= x1;
			max		= y1;
		}
	}else{// x1 < x0
		if( x0 < x2 ){	// x1 < x0 < x2
			xmin	= x1;
			min		= y1;
			xmax	= x2;
			max		= y2;
		}else{	// (x1 < x0)&&(x2 < x0)
			if( x1 < x2 ){ // x1 < x2 < x0
				xmin	= x1;
				min		= y1;
			}else{ // x2 < x1 < x0
				xmin	= x2;
				min		= y2;
			}
			xmax	= x0;
			max		= y0;
		}
	}

	if(max<min) std::swap(min,max);

//	dout()	<<"x in ["<<xmin<<","<<xmax<<"]"	<<"\t"<<"y in ["<<min<<","<<max<<"]"	<<endl;

	if(a!=0.){	// exists extremum

		const double x_extremum = -0.5*b/a;

//		dout()	<<"extremum at "<<x_extremum	<<endl;

		if((xmin<x_extremum)&&(x_extremum<xmax)){	//x_extremum in [xmin,xmax]

			double y_extremum = 0.5*b*x_extremum+c;
			if(a>0.){	// minimum
				min = y_extremum;
//				dout()	<<"min "<<min	<<endl;
			}else{		// maximum
				max = y_extremum;
//				dout()	<<"max "<<max	<<endl;
			}
		}
	}

	return true;
}





