/**
 * @file 	utils.h
 * @date	27.05.2014
 * @author	daniel
 * @brief	Datei_Beschreibung_Kurz
 * 
 * Datei_Beschreibung_Lang
 */

/** Funktion_Beschreibung_Kurz.
 * @param	Paramter_Name Parameter_Beschreibung
 * @return Return_Beschreibung
 * 
 * Funktion_Beschreibung_Lang
 */
#ifndef UTILS_H_
#define UTILS_H_

#include <cstdlib>
#include <sys/time.h>
#include <iostream>
#include <sstream>
#include <vector>
#include <algorithm>
#include "utils_simple_math.h"

namespace CuTe {
namespace NS_Utils {

using std::cout;
using std::endl;
using std::vector;
using std::string;


void GSL_ERROR_CODE(const string& expl, const int status);

template<class T>
std::ostream& operator<<( std::ostream& os, const vector< vector<T> >& V ){
	const int ni(V.size());
	for(int i=0; i<ni; ++i){
		const int nj(V[i].size());
		for(int j=0; j<nj; ++j) os << V[i][j] << " ";
		os << endl;
	}
	return os;
}

template<typename T>
std::ostream& operator<<( std::ostream& os, const vector<T>& V ){
	const int ni(V.size());
	for(int i=0; i<ni; ++i) os << V[i] << " ";
	return os;
}

//template<typename T>
//const T max_of_vec(const vector<T>& V){
//	typename vector<T>::const_iterator it(V.begin());
//	int n(V.size());
//	T max(*(it));
//	for(int i=1; i<n; ++i){
//		if(max<(*(++it))) max = *it;
//	}
//	return max;
//}
//
//template<typename T>
//const T min_of_vec(const vector<T>& V){
//	typename vector<T>::const_iterator it(V.begin());
//	int n(V.size());
//	T min(*(it));
////	cout	<<"min="<<min	<<"\tV["<<0<<"] = "<<*it	<<endl;
//	for(int i=1; i<n; ++i){
//		if(min>(*(++it))) min = *it;
////		cout	<<"min="<<min	<<"\tV["<<i<<"] = "<<*it	<<endl;
//	}
////	cout	<<"min_of_vec::return("<<min<<")"	<<endl;
//	return min;
//}

template< typename T >
typename vector<T>::iterator replace(typename vector<T>::iterator position, typename vector<T>::const_iterator begin,const typename vector<T>::const_iterator& end){
	while(begin!=end){
		*(position++)	= *(begin++);
	}
	return position;
}

class Timer {
protected:
    timeval _time;
public:

	double t;

	inline Timer()
	:t(0.){
		start();
	}

	inline double time(){
	    if (gettimeofday(&_time,NULL)){
	        std::cerr	<< "**** Timer Error ****"<< std::endl;
	        return 0;
	    }
	    return (double)_time.tv_sec + (double)_time.tv_usec * .000001;
	}

	inline Timer& start(){
	    t	= time();
	    return *this;
	}

	inline Timer& stop(){
		t	= time()	- t;
	    return *this;
	}
};

std::ostream& operator<< ( std::ostream& os,const Timer& t);


class NulStreambuf : public std::streambuf
{
    char                dummyBuffer[ 64 ];
protected:
    virtual int         overflow( int c )
    {
        setp( dummyBuffer, dummyBuffer + sizeof( dummyBuffer ) );
        return (c == traits_type::eof()) ? '\0' : c;
    }
};

//class NulOStream : private NulStreambuf, public std::ostream
//{
//public:
//    NulOStream() : std::ostream( this ) {}
//    NulStreambuf* rdbuf() const { return this; }
//};

class DOUT {
public:
	enum IO {in,out};
protected:

	string name;
	string params;
	static int tab;
	NulStreambuf NULLbuff;
	std::ostream NULLos;
	std::ostream *os;

	void _init();

	inline DOUT()
	:NULLbuff(),NULLos(&NULLbuff),os(NULL){
		_init();
	}

	inline void print(IO io){
		cout<<TAB();
		switch (io) {
			case in:
				cout	<<"-> ";
				break;
			case out:
				cout	<<"<- ";
				break;
			default:
				std::cerr	<<"DOUT print("<<io<<")"	<<endl;
				break;
		}
		(*os)	<<name<<"("<<params<<")"<<	endl;
	}
public:

	inline std::ostream& operator() ()const{
		return (*os) << TAB();
	}

	inline std::ostream& operator() (const uint ntab)const{
		if(ntab==0) return (*os);
		else return (*this)(ntab-1)<<"\t";
	}

	inline static const string TAB(){
		std::stringstream ss;
		for(int i=0; i<tab; ++i) ss<<"\t";
		return ss.str();
	}

	inline int get_ntab()const{return tab;}

	inline DOUT(const string& name)
	:name(name),params(),NULLbuff(),NULLos(&NULLbuff),os(NULL){
		_init();
		print(in);
		++tab;
	}

	inline DOUT(const string& name,const string& params)
	:name(name),params(params),NULLbuff(),NULLos(&NULLbuff),os(NULL){
		_init();
		print(in);
		++tab;
	}

	inline ~DOUT(){
		--tab;
		print(out);
	}

	inline DOUT& operator<< (const string& s){
		params += s;
		return *this;
	}

};

/*
 * y(x)	= a x^2 + b x + c
 */
bool interpolate_parabola(const double x0, const double y0, const double x1, const double y1, const double x2, const double y2, double &a, double &b, double &c);

/*
 * get min & max
 * of the parabola defined by y_i = a x_i^2 + b x_i + c
 * inside the interval [min{x_i},max{x_i}]
 */
bool parabola_min_max(const double x0, const double y0, const double x1, const double y1, const double x2, const double y2, double &min, double &max);

template< typename T >
bool insert_N_values(vector<T> in_vals, const int N, vector<T> &out_vals){
	bool OK(true);
	if(in_vals.size()<1){
		std::cerr	<<"No in values"	<<endl;
		return false;
	}
	if(N<1){
		std::cerr	<<"N = "<<N<<" < 1"	<<endl;
		return false;
	}

	typename vector<T>::iterator it(in_vals.begin());
	typename vector<T>::iterator end(in_vals.end());
	std::sort(it,end);

	out_vals.clear();

	T val(*(it++));
	T Delta;
	T r(N);
	r	= 1./r;

	while(it!=end){
		Delta	= (*(it++)-val)*r;

		for(int j=0; j<N; ++j){
			out_vals.push_back(val);
			val += Delta;
		}
	}
	out_vals.push_back(val);

	return OK;
	}

}  // namespace NS_Utils



}  // namespace CuTe



#endif /* UTILS_H_ */
