#ifndef UTILS_SIMPLE_MATH_H_
#define UTILS_SIMPLE_MATH_H_
#include <cmath>

namespace CuTe {
namespace NS_Utils {

namespace NS_min_max {

	template< typename T >
	inline const T& min_of(const T &a, const T &b){ return ( ( (a)<(b) ) ? (a) : (b) );}

	template< typename T >
	inline const T& max_of(const T &a, const T &b){ return ( ( (a)>(b) ) ? (a) : (b) );}

	template <typename T>
	inline const T min_of_Vec(const std::vector<T>& Vec){
		typename std::vector<T>::const_iterator it(Vec.begin());
		const typename std::vector<T>::const_iterator end(Vec.end());
		T tmp = *it;
		while((++it)!=end){
			tmp	= min_of<T>(tmp,*it);
		}
		return tmp;
	}

	template <typename T>
	inline const T max_of_Vec(const std::vector<T>& Vec){
		typename std::vector<T>::const_iterator it(Vec.begin());
		const typename std::vector<T>::const_iterator end(Vec.end());
		T tmp = *it;
		while((++it)!=end){
			tmp	= max_of<T>(tmp,*it);
		}
		return tmp;
	}

	template< typename T >
	inline const int sign_of(const T &x){ return ( (x >= T(0)) ? +1 : -1 );}

}  // namespace NS_min_max

namespace NS_pow {

	template< typename T >
	inline const T pow_2(const T &x) { return x*x;   }
	template< typename T >
	inline const T pow_3(const T &x) { return x*x*x; }
	template< typename T >
	inline const T pow_4(const T &x) { const T x2(x*x);   return x2*x2;    }
	template< typename T >
	inline const T pow_5(const T &x) { const T x2(x*x);   return x2*x2*x;  }
	template< typename T >
	inline const T pow_6(const T &x) { const T x2(x*x);   return x2*x2*x2; }
	template< typename T >
	inline const T pow_7(const T &x) { const T x3(x*x*x); return x3*x3*x;  }
	template< typename T >
	inline const T pow_8(const T &x) { const T x2(x*x);   T x4 = x2*x2; return x4*x4; }
	template< typename T >
	inline const T pow_9(const T &x) { const T x3(x*x*x); return x3*x3*x3; }

	template< typename T>
	inline const T pow_n(const T &x, const uint n) {
		if(n==0) return 1;
		if(n==1) return x;
		const T xN2(pow_n(x, n/2));
		if (n % 2) {
			return xN2 * xN2 * x;
		}
		return xN2 * xN2;
	}

}  // namespace NS_pow




}  // namespace NS_Utils
}  // namespace CuTe

#endif
