#ifndef VECMATH_H
#define VECMATH_H

/*#ifndef int
#define int unsigned int
#endif*/

#include <vector>
#include <ostream>
//#include <array>
//#include <cmath>
//#include <complex>
//using std::complex;

//typedef unsigned int  int;
//
template <int N, typename T>
class Matrix;

template <int N, typename T>
class Vector{
public:
typedef T*        	iterator;
typedef const T*	const_iterator;
protected:
	T* _V;
	T* _end;
public:
//---------------Konstruktoren
	inline Vector()
		:_V(new T[N])
		,_end(_V+N){
//		for(int i=N-1; i>=0; --i) _V[i]=T();
	}

	explicit inline Vector(T const &t)
		:_V(new T[N])
		,_end(_V+N){
		for(iterator it(begin()); it!=_end; ++it)	(*it)=t;
//		for(int i=0; i<N; ++i) _V[i]=t;
	}

	inline Vector(Vector const& V)
		:_V(new T[N])
		,_end(_V+N){
		const_iterator it2(V.begin());
		for(iterator it(_V); it!=_end; ++it){
			(*it)=(*it2);
			++it2;
		}
//		for(int i=0; i<N; ++i)_V[i]=V[i];
	}

	inline Vector(T (&array)[N])
		:_V(new T[N])
		,_end(_V+N){
		const_iterator it2(array);
		for(iterator it(_V); it!=_end; ++it){
			(*it)=(*it2);
			++it2;
		}
//		for(int i=0; i<N; ++i)_V[i]=array[i];
	}

	inline Vector(const std::vector<T>& v)
		:_V(new T[N])
		,_end(_V+N){
		const_iterator it2(v.begin());
		for(iterator it(_V); it!=_end; ++it){
			(*it)=(*it2);
			++it2;
		}
//		for(int i=0; i<N; ++i)_V[i]=v[i];
	}

	~Vector(){
//		std::cout	<<"~V<"<<N<<"> ";
		delete[] _V;	// No = NULL
	}
//----------------	get
	inline int n()const{return N;}
	inline T& operator[](int i){ return _V[i];}
	inline const T& operator[](int i)const{ return _V[i];	}
//----------------	iterator
	inline iterator begin(){ return _V;}
	inline iterator end(){ return _end;}
	inline const_iterator begin()const{ return _V;}
	inline const_iterator end()const{ return _end;}
	inline iterator rbegin(){ return _end-1;}
	inline iterator rend(){ return _V-1;}
	inline const_iterator rbegin()const{ return _end-1;}
	inline const_iterator rend()const{ return _V-1;}
//----------------	set
	inline Vector& operator=(const Vector& V){
		const_iterator it2(V.begin());
		for(iterator it(_V); it!=_end; ++it){
			(*it)=(*it2);
			++it2;
		}
//		for(int i=0; i<N; ++i)_V[i]=V[i];
		return *this;
	}
//----------------	funcs
	inline const T norm2()const{ return scalar_product(*this);}

	template< int p >
	inline const T p_norm()const{
		const_iterator it(begin());
		T norm(pow(*(it++),p));
		while(it!=_end){
			norm += pow(*(it++),p);
		}
		const double ip(1./static_cast<double>(p));
		return pow(norm,ip);
	}
//-----------------  standard calc

	inline const Vector operator-()const{
		Vector tmp(*this);
		for(iterator it(tmp.begin()); it!=tmp.end(); ++it){
			(*it)=-(*it);
		}
//		for(int i=0; i<N; ++i) tmp[i]=-tmp[i];
		return tmp;
	}

	template< typename T2 >
	inline Vector& add_Vec_to_this(const Vector< N,T2 >& V2){
		typename Vector<N,T2>::const_iterator it2(V2.begin());
		for(iterator it(_V); it!=_end; ++it){
			(*it)+=(*it2);
			++it2;
		}
//		for(int i=0; i<N; ++i) _V[i]+=V2[i];
		return *this;
	}

	template< typename T2 >
	inline const Vector add_Vec(const Vector< N,T2 >& V2)const{
		Vector tmp(*this);
		tmp.add_Vec_to_this(V2);
		return tmp;
	}

	template< typename T2 >
	inline Vector& sub_Vec_to_this(const Vector< N,T2 >& V2){
		typename Vector<N,T2>::const_iterator it2(V2.begin());
		for(iterator it(_V); it!=_end; ++it){
			(*it)-=(*it2);
			++it2;
		}
//		for(int i=0; i<N; ++i) _V[i]-=V2[i];
		return *this;
	}

	template< typename T2 >
	inline const Vector sub_Vec(const Vector< N,T2 >& V2)const{
		Vector tmp(*this);
		tmp.sub_Vec_to_this(V2);
		return tmp;
	}

	inline Vector& operator+=(const Vector& V){
		return add_Vec_to_this(V);
	}

	inline Vector& operator-=(const Vector& V){
		return sub_Vec_to_this(V);
	}

	inline const Vector operator+(const Vector& V)const{
		return add_Vec(V);
	}

	inline const Vector operator-(const Vector& V)const{
		return sub_Vec(V);
	}

//-----------------  scalar multiplikation
	template< typename T2 >
	inline Vector& mul_this_by_scalar(const T2& t2){
		for(iterator it(_V); it!=_end; ++it){
			(*it)*=t2;
		}
//		for(int i=0; i<N; ++i) _V[i]*=t2;
		return *this;
	}
	template< typename T2 >
	inline const Vector mul_by_scalar(const T2& t2)const{
		Vector tmp(*this);
		tmp.mul_this_by_scalar(t2);
		return tmp;
	}

	inline Vector& operator*=(const T& t){
		return mul_this_by_scalar(t);
	}

	inline const Vector operator*(const T& t)const{
		return mul_by_scalar(t);
	}

//------------------------	skalarproduct

	template < typename T2 >
	inline const T scalar_product(const Vector< N,T2 >& V2)const{
		iterator it(_V);
		typename Vector<N,T2>::const_iterator it2(V2.begin());
		T tmp((*it)*(*it2));
		++it; ++it2;
		for(; it!=_end; ++it){
			tmp += (*it)*(*it2);
			++it2;
		}
//		T tmp(_V[0]*V2[0]);
//		for(int i=1; i<N; ++i) tmp+=_V[i]*V2[i];
		return tmp;
	}

	inline const T operator*(const Vector& V)const{
			return scalar_product(V);
		}


	//----------------- Vector Matrix multiplikation
		template< typename T2 >
		inline Vector& mul_this_by_Matrix_from_right(Matrix<N,T2> M2);

		template< typename T2 >
		inline const Vector mul_by_Matrix_from_right(const Matrix<N,T2>& M2)const{
			Vector tmp(*this);
			return tmp.mul_this_by_Matrix_from_right(M2);
		}

		template< typename T2 >
		inline Vector<N,T>& mul_this_by_Matrix_from_left(const Matrix<N,T2>& M2);
		template< typename T2 >
		inline const Vector<N,T> mul_by_Matrix_from_left(const Matrix<N,T2>& M2)const{
			Vector tmp(*this);
			return tmp.mul_this_by_Matrix_from_left(M2);
		}

		inline Vector<N,T>& operator*=(const Matrix<N,T>& M){
			return mul_this_by_Matrix_from_right(M);
		}

		inline const Vector<N,T> operator*(const Matrix<N,T>& M)const{
			return mul_by_Matrix_from_right(M);
		}

};

template < int N, typename T >
inline std::ostream& operator<<(std::ostream& os,  Vector< N,T > const & V)
{
	os << "( " << V[0] << ",";
    for(int i=1; i<(N-1); ++i) os << "\t" << V[i] << ",";
    os << "\t" << V[N-1] << " )";
    return os;
}


//////////////////////////////////////////////////////////
/////////////		Matrix	/////////////
////////////////////////////////////////////////////////////

template < int N, typename T >
class Matrix{
protected:
	Vector< N,T > *_M;

public:
//---------------Konstruktoren
	inline Matrix()
		:_M(new Vector< N,T>[N]){
	}

	explicit inline Matrix(const T& t)
		:_M(new Vector< N,T>[N]){
		for(int i=0; i<N; ++i)for(int j=0; j<N; ++j) _M[i][j]=t;
	}

	inline Matrix(const Matrix& M)
		:_M(new Vector< N,T>[N]){
		for(int i=0; i<N; ++i) _M[i]=M[i];
	}


	~Matrix< N,T >(){
//		std::cout	<<"~M<"<<N<<">\t";
		delete[] _M;	// No = NULL
//		std::cout	<<std::endl;
	}

//----------------	iterator
	inline typename Vector<N,T>::iterator begin(int n){ return _M[n].begin();}
	inline typename Vector<N,T>::iterator end(int n){ return _M[n].end();}
	inline typename Vector<N,T>::const_iterator begin(int n)const{ return _M[n].begin();}
	inline typename Vector<N,T>::const_iterator end(int n)const{ return _M[n].end();}

//---------------	get
	inline int n()const{return N;}
	inline Vector<N,T>& operator[](int i){ return _M[i];}
	inline const Vector<N,T>& operator[](int i)const{return _M[i];	}


//---------------	set
	inline Matrix& operator=(const Matrix& M){
		for(int i=0; i<N; ++i) _M[i]=M[i];
		return *this;
	}

	inline Matrix& set_this_to_identity(const T& one = T(1), const T& zero = T()){
		for(int i=0; i<N;++i){
			for(int j=0; j<i;++j) _M[i][j]=zero;
			_M[i][i]=one;
			for(int j=i+1; j<N;++j) _M[i][j]=zero;
		}
		return *this;
	}

	inline const Matrix get_identity(const T& one = 1., const T& zero = T())const{
		Matrix tmp;
		return tmp.set_this_to_identity(one,zero);
	}
//---------------	funcs

	inline Matrix& transpose_this(){
		for(int i=0; i<N; ++i) for(int j=i+1; j<N; j++){
			std::swap(_M[i][j],_M[j][i]);
		}
		return *this;
	}

	inline const Matrix get_transpose()const{
		Matrix tmp(*this);
		return tmp.transpose_this();
	}

//-----------------  standard calc

	inline const Matrix operator-()const{
		Matrix tmp(*this);
		for(int i=0; i<N; ++i) tmp[i]=-tmp[i];
		return tmp;
	}

	template< typename T2 >
	inline Matrix& add_Matrix_to_this(const Matrix< N,T2 >& M2){
		for(int i=0; i<N; ++i) _M[i].add_Vec_to_this(M2[i]);
		return *this;
	}

	template< typename T2 >
	inline const Matrix add_Matrix(const Matrix< N,T2 >& M2)const{
		Matrix tmp(*this);
		return tmp.add_Matrix_to_this(M2);
	}

	template< typename T2 >
	inline Matrix& sub_Matrix_to_this(const Matrix< N,T2 >& M2){
		for(int i=0; i<N; ++i) _M[i].sub_Vec_to_this(M2[i]);
		return *this;
	}

	template< typename T2 >
	inline const Matrix sub_Matrix(const Matrix< N,T2 >& M2)const{
		Matrix tmp(*this);
		return tmp.sub_Matrix_to_this(M2);
	}

	inline Matrix& operator+=(const Matrix& M){
		return add_Matrix_to_this(M);
	}

	inline Matrix& operator-=(const Matrix& M){
		return sub_Matrix_to_this(M);
	}

	inline const Matrix operator+(const Matrix& M)const{
		return add_Matrix(M);
	}

	inline const Matrix operator-(const Matrix& M)const{
		return sub_Matrix(M);
	}

//-----------------  scalar multiplikation
	template< typename T2 >
	inline Matrix& mul_this_by_scalar(const T2& t2){
		for(int i=0; i<N; ++i) _M[i].mul_this_by_scalar(t2);
		return *this;
	}
	template< typename T2 >
	inline const Matrix mul_by_scalar(const T2& t2)const{
		Matrix tmp(*this);
		tmp.mul_this_by_scalar(t2);
		return tmp;
	}

	inline Matrix& operator*=(const T& t){
		return mul_this_by_scalar(t);
	}

	inline const Matrix operator*(const T& t)const{
		return mul_by_scalar(t);
	}
//-----------------  Matrix multiplikation
	template< typename T2 >
	inline Matrix& mul_this_by_Matrix_from_right(Matrix<N,T2> M2){
		Matrix tmp(*this);
		M2.transpose_this();
		for(int i=0; i<N; ++i)for(int j=0; j<N; ++j) _M[i][j]=tmp[i].scalar_product(M2[j]);
		return *this;
	}
	template< typename T2 >
	inline const Matrix mul_by_Matrix_from_right(const Matrix<N,T2>& M2)const{
		Matrix tmp(*this);
		return tmp.mul_this_by_Matrix_from_right(M2);
	}

	template< typename T2 >
	inline Matrix& mul_this_by_Matrix_from_left(Matrix<N,T2> M2){
		transpose_this();
		Matrix tmp(*this);
		for(int i=0; i<N; ++i)for(int j=0; j<N; ++j) _M[i][j]=tmp[i].scalar_product(M2[j]);
		return *this;
	}
	template< typename T2 >
	inline const Matrix mul_by_Matrix_from_left(const Matrix<N,T2>& M2)const{
		Matrix tmp(*this);
		return tmp.mul_this_by_Matrix_from_left(M2);
	}

	inline Matrix& operator*=(const Matrix& M){
		return mul_this_by_Matrix_from_right(M);
	}

	inline const Matrix operator*(const Matrix& M)const{
		return mul_by_Matrix_from_right(M);
	}

//----------------- Matrix Vector multiplikation
	template< typename T2 >
	inline const Vector<N,T> mul_by_Vector_from_right(const Vector<N,T2>& V2)const{
		Vector<N,T> tmp;
		for(int i=0; i<N; ++i) tmp[i]=_M[i].scalar_product(V2);
		return tmp;
	}
	template< typename T2 >
	inline const Vector<N,T> mul_by_Vector_from_left(const Vector<N,T2>& V2)const{
		Matrix tmp(*this);
		tmp.transpose_this();
		return tmp.mul_by_Vector_from_right(V2);
	}

	inline const Vector<N,T> operator*(const Vector<N,T>& V)const{
		return mul_by_Vector_from_right(V);
	}
};


template< int N,typename T >
template< typename T2 >
inline Vector<N,T>& Vector<N,T>::mul_this_by_Matrix_from_right(Matrix<N,T2> M2){
	Vector  tmp(*this);
	M2.transpose_this();
	for(int i=0; i<N; ++i) _V[i]=tmp.scalar_product(M2[i]);
	return *this;
}

template< int N,typename T >
template< typename T2 >
inline Vector<N,T>& Vector<N,T>::mul_this_by_Matrix_from_left(const Matrix<N,T2>& M2){
	Vector  tmp(*this);
	for(int i=0; i<N; ++i) _V[i]=tmp.scalar_product(M2[i]);
	return *this;
}

template < int N, typename T >
inline std::ostream& operator<<(std::ostream& os, const Matrix< N,T >& M)
{
	for(int i=0; i<(N-1); ++i){
		os << M[i] << "\n";
	}
	os << M[N-1];

    return os;
}

#endif
