#ifndef PLOT_H
#define PLOT_H

#include "CuTe.h"
#include <iomanip>

using namespace CuTe;
using NS_CS_Params::MU_star;
using NS_File_Path::File_Path;
using namespace NS_enums;
using std::cerr;
using std::ofstream;
using std::ostream;
using std::stringstream;
using std::vector;

using namespace CuTe::NS_Utils::NS_min_max;
using NS_Utils::Timer;


class Plot{
protected:

	string progname;

	typedef vector<double>::const_iterator vdcit;

	enum e_var {e_qT=0, e_mu=1, e_y=2 };
	enum E_Plot {E_Plot_qT, E_Plot_qT_y, E_Grid_Plot_qT, E_Grid_Plot_qT_y } e_plot;
	int N_min_grid;

	SOPS_CuTe OP;

	File_Path infilename;
	File_Path outfilename;

	vector<double> qT_values;
	vector<double> y_values;

	vector<double> qT_cumulant_values;

	int ofprec;
	int ofwidth;

	vector<bool> toplot;
	Crosssection CS;
	RES res;


	inline const string _FirstLine(){
//		NS_Utils::DOUT dout("Plot::_FirstLine");
		stringstream FL;

		time_t now = time(0);
		char* dt = ctime(&now);

		FL << "########################################################\n";
		FL << "##  ";											FL << "\n";
		FL << "##  ";					FL << PACKAGE_STRING;	FL << "\n";
		FL << "##  ";					FL << dt;
		FL << "##  ";											FL << "\n";
		FL << "########################################################\n";
		FL << "##  ";											FL << "\n";
		if(progname!=string()){
			FL << "##  Command: ";			FL << progname;		FL << "\n";
		}
		FL << "##  Infile: ";			FL << infilename;		FL << "\n";
		FL << "##  ";											FL << "\n";
		FL << "########################################################\n";


		OP.into_stream(FL,"##\t");

		FL << "########################################################\n";
		FL << "##  Outfile: ";									FL << "\n";
		FL << "########################################################\n";
		FL << "########################################################\n";

		FL << "##\t"<<ER_qT<<"\t: "<<"Transverse Momentum of the  Boson"<<endl;
		FL << "##\t"<<ER_mu<<"\t: "<<"Factorization Scale"<<endl;
		if(toplot[ER_y]){
			FL << "##\t"<<ER_y<<"\t: "<<"Pseudo-Rapidity of the Boson"<<endl;
			FL << "##\t"<<ER_erg_result<<"\t: "<<"(d^2sigma)/(dqT*dy)	("<<long_str(res)<<")"<<endl;
		}else{
			FL << "##\t"<<ER_erg_result<<"\t: "<<"dsigma/dqT	("<<long_str(res)<<")"<<endl;
		}
		for(int i=static_cast<int>(ER_erg_RES); i<=static_cast<int>(ER_erg_RC); ++i){
			if(toplot[i])	FL << "##\t\t"<<static_cast<E_RESULTS>(i)<<"\t: "<<long_str(static_cast<E_RESULTS>(i))<<endl;

		}
		FL << "##\t"<<ER_erg_Delta<<"\t: "<<"Absolute, symmetric Integration Error"<<endl;
		if(toplot[ER_erg_mu_P]){
//			FL << "##\tmu+\t: "<<"Approximated maximum in [mu/mu_factor,mu*mu_factor]"<<endl;
//			FL << "##\tmu-\t: "<<"Approximated minimum in [mu/mu_factor,mu*mu_factor]"<<endl;
			FL << "##\tScale-errors from varying mu in [mu/mu_factor,mu*mu_factor]. Result + D_mu+-"<<endl;
			FL << "##\t\t"<<ER_erg_mu_P<<"\t: "<<"Positive scale error"<<endl;
			FL << "##\t\t"<<ER_erg_mu_M<<"\t: "<<"Negative scale error"<<endl;
		}
		if(toplot[ER_erg_PDF_Delta]){
			FL << "##\tLHAPDF errors corresponding to one sigma:"<<endl;
			FL << "##\t\t"<<ER_erg_PDF_Delta<<"\t: "	<<"Absolute, Symmetric"<<endl;
			FL << "##\t\t"<<ER_erg_PDF_DeltaP<<"\t: "	<<"Positive"<<endl;
			FL << "##\t\t"<<ER_erg_PDF_DeltaM<<"\t: "	<<"Negative"<<endl;
//			FL << "##\t\tD_PDF_mu\t: "<<"Scale"<<endl;
		}
		if(toplot[ER_erg_exp]){
			FL << "##\t"<<ER_erg_exp<<"\t: "<<"Signed exponent error"<<endl;
		}

		FL << "##\tD_C\t: "<<"Signed quadratically combined errors (=+-sqrt(e1+-^2 + ... + en+-^2))"<<endl;
		FL << "##\t\t"<<ER_erg_DCP<<"\t: "<<"Positive"<<endl;
		FL << "##\t\t"<<ER_erg_DCM<<"\t: "<<"Negative"<<endl;

		FL	<< "##\talpha_s("<<DEFAULT_MZ<<") = "<< CS.get_alpha_s_MZ()<<endl;
		FL	<< "##\tq_star(M) = "<< CS.get_q_star()<<endl;
		FL << "########################################################\n";


		CS.FL_into_stream(FL,toplot,ofwidth);


		return FL.str();
	}

	inline const string plotpoint(Timer &time){
		stringstream ss;

		ss.precision(ofprec);

		CS.into_stream(ss,toplot,ofwidth,ofprec);

		cout	<<std::setw(ofwidth+1)<<time	<<" "<<ss.str()<<endl;

		return ss.str();

	}


	inline void _plot_qT_qTmu(ofstream &of, Timer &timer, const double qT, const double qTmu){
		timer.start();
		CS.update_qT(qT,qTmu);
		timer.stop();
		of << plotpoint(timer) << endl;
	}

	inline void _plot_qT_qTmu_y(ofstream &of, Timer &timer, const double qT, const double qTmu, const double y){
		timer.start();
		CS.update_qT_y(qT,qTmu,y);
		timer.stop();
		of << plotpoint(timer) << endl;
	}

	template< E_Plot thisplot >
	inline void make_cumulant_values(const int i_qT, const double qT){
		qT_cumulant_values.assign(qT_values.begin(),qT_values.begin()+i_qT+1);
		if( (0<i_qT)&&((i_qT+1)<N_min_grid) ){

			int N((N_min_grid-1)/i_qT);
			if(N*i_qT+1<N_min_grid) ++N;

			NS_Utils::insert_N_values(qT_cumulant_values,N,qT_cumulant_values);
		}
	}

	template< E_Plot thisplot >
	inline void tplot_qT_y(ofstream &of, Timer &timer, const double qT, const double y){
		_plot_qT_qTmu_y(of,timer,qT,qT,y);
	}

	template< E_Plot thisplot >
	inline void tplot_qT(ofstream &of, Timer &timer, const double qT){
		_plot_qT_qTmu(of,timer,qT,qT);
	}

	template< E_Plot thisplot >
	inline void tplot(ofstream &of){
		Timer timer;
		vdcit itr_qT(qT_values.begin());
		const vdcit end_qT(qT_values.end());

		int i_qT(0);

		while(itr_qT!=end_qT){
			const double qT(*(itr_qT++));
			if(qT>0.){
				make_cumulant_values<thisplot>(i_qT,qT);
				tplot_qT<thisplot>(of,timer,qT);
			}
			++i_qT;
		}
	}




public:

	Plot(const string& progname="")
	:progname(progname)
	,e_plot(E_Plot_qT),N_min_grid(0)
	,OP()
	,infilename(),outfilename()
//	,qTmin(0.),qTmax(0.),qTN(0.),qTDelta(0.)
	,qT_values(),y_values()
	,ofprec(0),ofwidth(0)
	,toplot(vector<bool>(ER_Last,false))//,tocalc(vector<bool>(tp_Last,false))
	,CS()
	,res(RES_WRONG_STATEMENT){

	}
	~Plot(){
//		delete PP;
	}



	bool init(const SOPS_CuTe& OP_in){
		bool OK(true);
		this->OP = OP_in;
//		OK = OK&PP.init(op);
		infilename = OP.get_infilename();
		{
			outfilename = File_Path(infilename,infilename,"out");
//			size_t found = infilename.find_last_of(".");
//			outfilename = infilename.substr(0,found) + ".out";
//			cout	<<"Plot::init(op)"	<<endl
//					<<"\t"<<"if:"<<infilename.properties()	<<endl
//					<<"\t"<<"of:"<<outfilename.properties()	<<endl;
		}

		OK = NS_Utils::insert_N_values(OP.Plot.qT(),OP.Plot.qT_N(),qT_values);
		e_plot	= E_Plot_qT;

		if(OK && OP.Plot.y.is_set){
			OK = NS_Utils::insert_N_values(OP.Plot.y(),OP.Plot.y_N(),y_values);
			e_plot	= E_Plot_qT_y;
		}else{
			y_values.clear();
		}
		N_min_grid = 0;
		if( OP.Plot.Gridplot.is_set && (OP.Plot.Gridplot()>0) ){
			N_min_grid	= OP.Plot.Gridplot();
			if(e_plot==E_Plot_qT)	e_plot	= E_Grid_Plot_qT;
			if(e_plot==E_Plot_qT_y)	e_plot	= E_Grid_Plot_qT_y;

//			OP.Scales.Scale_Error.value = false;
		}

		if(OK){
			ofprec	= max_of(static_cast<int>(log10(10./OP.CUBA.accuracy())+2),6);
			ofwidth	= ofprec+3;

			CS.init(OP);
//			init_toplot();
			toplot	= CS.to_plot(toplot,y_values.size()>0);
			res = OP.Scheme.resummation();
		}



		return OK;
	}

	bool createPlot(){
//		NS_Utils::DOUT dout("Plot::createPlot");
		bool OK(true);
		cout << "Outfilename: " << outfilename << endl;
		ofstream of(outfilename.c_str());

		of << _FirstLine() << endl;

		int tmp_prec(of.precision());
		int tmp_width(of.width());

		of.precision(ofprec);
		of.width(ofwidth);


		cout	<< std::setw(ofwidth+1+6)<<"duration ";
		CS.FL_into_stream(cout,toplot,ofwidth);
		cout	<<endl;

		Timer timer;
		timer.start();

		switch (e_plot) {
			case E_Plot_qT:
				tplot<E_Plot_qT>(of);
				break;
			case E_Plot_qT_y:
				tplot<E_Plot_qT_y>(of);
				break;
			case E_Grid_Plot_qT:
				tplot<E_Grid_Plot_qT>(of);
				break;
			case E_Grid_Plot_qT_y:
				tplot<E_Grid_Plot_qT_y>(of);
				break;
			default:
				break;
		}

		cout	<<endl <<"Time: "<<timer.stop()<<endl;

		of	<< "####################################################################"<<endl
			<< "## Time: "<<timer<<endl;

		of.precision(tmp_prec);
		of.width(tmp_width);

		of.close();
		return OK;
	}
};

template<  >
inline void Plot::make_cumulant_values<Plot::E_Plot_qT>(const int i_qT, const double qT){
}

template<  >
inline void Plot::make_cumulant_values<Plot::E_Plot_qT_y>(const int i_qT, const double qT){
}

template<  >
inline void Plot::tplot_qT<Plot::E_Grid_Plot_qT>(ofstream &of, Timer &timer, const double qT){
	vdcit itr_qT_cum(qT_cumulant_values.begin());
	const vdcit end_qT_cum(qT_cumulant_values.end());

	while(itr_qT_cum!=end_qT_cum){
		_plot_qT_qTmu(of,timer,*(itr_qT_cum++),qT);
	}
}

template<  >
inline void Plot::tplot_qT<Plot::E_Plot_qT_y>(ofstream &of, Timer &timer, const double qT){
	vdcit itr_y(y_values.begin());
	const vdcit end_y(y_values.end());

	while(itr_y!=end_y){
		tplot_qT_y<E_Plot_qT_y>(of,timer,qT,*(itr_y++));
	}
}

template<  >
inline void Plot::tplot_qT_y<Plot::E_Grid_Plot_qT_y>(ofstream &of, Timer &timer, const double qT, const double y){
	vdcit itr_qT_cum(qT_cumulant_values.begin());
	const vdcit end_qT_cum(qT_cumulant_values.end());

	while(itr_qT_cum!=end_qT_cum){
		_plot_qT_qTmu_y(of,timer,*(itr_qT_cum++),qT,y);
	}
}

template<  >
inline void Plot::tplot_qT<Plot::E_Grid_Plot_qT_y>(ofstream &of, Timer &timer, const double qT){
	vdcit itr_y(y_values.begin());
	const vdcit end_y(y_values.end());

	while(itr_y!=end_y){
		tplot_qT_y<E_Grid_Plot_qT_y>(of,timer,qT,*(itr_y++));
	}
}




#endif
